
import QtQuick 2.0
import QtQuick.Window 2.2

import QmlToolbox.Base 1.0
import QmlToolbox.Controls 1.0

import gloperate.rendering 1.0


ApplicationWindow
{
    id: window

    visible:    true
    width:      1024
    height:     768
    visibility: Window.Windowed // Window.FullScreen

    // Stage
    property string stage: 'AirTrafficPipeline'

    // Pipeline interface
    GlOperatePipeline
    {
        id: gloperatePipeline

        onCanvasChanged:
        {
            propertyEditors.update();

            if (canvas != null)
            {
                canvas.onStageInputChanged(function(slot, status)
                {
                    gloperatePipeline.slotChanged('root', slot, status);
                });

                // Load data set?
                if (commandLineParams.length > 0)
                {
                    gloperatePipeline.setValue('root', 'filename', commandLineParams[0]);
                }
            }
        }

        onSlotChanged: // (string path, string slot, var status)
        {
            if (path == 'root' && slot == 'updateTrigger')
            {
                timer.start();
            }
        }

        Timer
        {
            id: timer

            interval: 10
            repeat:   false

            onTriggered:
            {
                categoryFilter.update();
                timeline.update();
                timeline.emitUpdate();

                gloperate.canvas0.setValue('root.Trackball', 'pitch',   -50.7562);
                gloperate.canvas0.setValue('root.Trackball', 'yaw',      97.0541);
                gloperate.canvas0.setValue('root.Trackball', 'zoom',      0.0379259);
                gloperate.canvas0.setValue('root.Trackball', 'rotateX',   0.0);
                gloperate.canvas0.setValue('root.Trackball', 'rotateY', -45.0);
            }
        }
    }

    // Property editor panel
    Panel
    {
        id: propertyEditors

        anchors.top:     parent.top
        anchors.bottom:  parent.bottom
        anchors.left:    parent.left
        width:           400

        clip: true

        Behavior on width
        {
            PropertyAnimation { }
        }

        ScrollArea
        {
            anchors.fill:         parent
            anchors.bottomMargin: 2

            contentHeight: column.height + 2 * column.anchors.margins

            Column
            {
                id: column

                anchors.top:     parent.top
                anchors.left:    parent.left
                anchors.right:   parent.right
                anchors.margins: 8

                spacing: 8

                Category
                {
                    id: categoryDataSet

                    width: parent.width

                    title:          'Data Set'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'dataset'
                    compact:        false
                }

                Category
                {
                    id: categoryGeneral

                    width: parent.width

                    title:          'Visualization'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'general'
                    compact:        true
                }

                Category
                {
                    id: categoryLines

                    width: parent.width

                    title:          'Lines'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'lines'
                    compact:        true
                }

                Category
                {
                    id: categoryTrajectories

                    width: parent.width

                    title:          'Trajectories'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'trajectories'
                    compact:        true
                }

                Category
                {
                    id: categoryMapping

                    width: parent.width

                    title:          'Mapping'
                    properties:     gloperatePipeline
                    path:           'root.Trajectories.config1'
                    categoryFilter: ''
                    compact:        false
                    lineSpacing:    8
                    contentVisible: false
                }

                Category
                {
                    id: categoryFilter

                    width: parent.width

                    title:          'Filtering'
                    properties:     gloperatePipeline
                    path:           'root.Filter'
                    categoryFilter: ''
                    compact:        false
                    lineSpacing:    16
                    contentVisible: false
                }
            }
        }

        function update()
        {
            categoryDataSet.update();
            categoryGeneral.update();
            categoryLines.update();
            categoryTrajectories.update();
            categoryMapping.update();
            categoryFilter.update();
        }
    }

    // Renderer
    RenderItem
    {
        id: render

        anchors.left:   propertyEditors.right
        anchors.right:  parent.right
        anchors.top:    parent.top
        anchors.bottom: parent.bottom

        stage: window.stage
        focus: true

        onCanvasInitialized:
        {
            gloperatePipeline.canvas = gloperate.canvas0;
        }

        TimeLine
        {
            id: timeline

            anchors.left:    parent.left
            anchors.right:   parent.right
            anchors.bottom:  parent.bottom
            anchors.margins: 16
            height:          32

            properties: gloperatePipeline
            timeSpans: 2
        }
    }

    // Keyboard shortcuts
    Shortcut
    {
        sequence: "ESC"

        onActivated:
        {
            if (propertyEditors.width == 0) {
                propertyEditors.width = 400;
            } else {
                propertyEditors.width = 0;
            }
        }
    }

    Shortcut
    {
        sequence: "F10"

        onActivated:
        {
            window.visibility = (window.visibility == Window.Windowed) ?  Window.FullScreen : Window.Windowed;

            if (window.visibility == Window.FullScreen) {
                propertyEditors.width = 0;
            }
        }
    }

    Shortcut
    {
        sequence: "F5"

        onActivated:
        {
            if (gloperate.system.logLevel() < 5)
            {
                console.log('Entering debug mode');
                gloperate.system.setLogLevel(5);
            }

            else
            {
                console.log('Leaving debug mode');
                gloperate.system.setLogLevel(3);
            }
        }
    }

    Component.onCompleted:
    {
        propertyEditors.update();
    }
}
