
import QtQuick 2.0

import QmlToolbox.Base 1.0
import QmlToolbox.Controls 1.0

import gloperate.rendering 1.0


ApplicationWindow
{
    id: window

    visible: true
    width:   1024
    height:  768

    // Stage
    property string stage: 'ClimatePipeline'

    // Pipeline interface
    GlOperatePipeline
    {
        id: gloperatePipeline

        onCanvasChanged:
        {
            propertyEditors.update();

            if (canvas != null)
            {
                canvas.onStageInputChanged(function(slot, status)
                {
                    gloperatePipeline.slotChanged('root', slot, status);
                });

                // Load script?
                if (commandLineParams.length > 0)
                {
                    var filename = commandLineParams[0];
                    if (filename.slice(-3) === '.js') {
                        gloperate.system.load(commandLineParams[0]);
                    }
                }
            }
        }

        onSlotChanged: // (string path, string slot, var status)
        {
            if (path == 'root' && slot == 'updateTrigger')
            {
                timer.start();
            }
        }

        Timer
        {
            id: timer

            interval: 100
            repeat:   false

            onTriggered:
            {
                categoryFilter.update();
                timeline.update();
                timeline.emitUpdate();
            }
        }
    }

    // Property editor panel
    Panel
    {
        id: propertyEditors

        anchors.top:     parent.top
        anchors.bottom:  parent.bottom
        anchors.left:    parent.left
        width:           400

        clip: true

        Behavior on width
        {
            PropertyAnimation { }
        }

        ScrollArea
        {
            anchors.fill:         parent
            anchors.bottomMargin: 2

            contentHeight: column.height + 2 * column.anchors.margins

            Column
            {
                id: column

                anchors.top:     parent.top
                anchors.left:    parent.left
                anchors.right:   parent.right
                anchors.margins: 8

                spacing: 8

                Category
                {
                    id: categoryDataSet

                    width: parent.width

                    title:          'Data Set'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'dataset'
                    compact:        false
                }

                Category
                {
                    id: categoryGeneral

                    width: parent.width

                    title:          'Visualization'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'general'
                    compact:        true
                }

                Category
                {
                    id: categoryNodes

                    width: parent.width

                    title:          'Nodes'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'nodes'
                    compact:        true
                }

                Category
                {
                    id: categoryEdges

                    width: parent.width

                    title:          'Edges'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'edges'
                    compact:        true
                }

                Category
                {
                    id: categoryPlot

                    width: parent.width

                    title:          'Plot'
                    properties:     gloperatePipeline
                    path:           'root'
                    categoryFilter: 'plot'
                    compact:        true
                }

                Category
                {
                    id: categoryFilter

                    width: parent.width

                    title:          'Filtering'
                    properties:     gloperatePipeline
                    path:           'root.Filter'
                    categoryFilter: ''
                    compact:        false
                    lineSpacing:    12
                }
            }
        }

        function update()
        {
            categoryDataSet.update();
            categoryGeneral.update();
            categoryNodes.update();
            categoryEdges.update();
            categoryPlot.update();
            categoryFilter.update();
        }
    }

    // Renderer
    RenderItem
    {
        id: render

        anchors.left:   propertyEditors.right
        anchors.right:  parent.right
        anchors.top:    parent.top
        anchors.bottom: parent.bottom

        stage: window.stage
        focus: true

        onCanvasInitialized:
        {
            gloperatePipeline.canvas = gloperate.canvas0;
        }

        Rectangle
        {
            id: selection

            visible: selectionArea.active
            x:       selectionArea.areaX1
            y:       selectionArea.areaY1
            width:   selectionArea.areaW
            height:  selectionArea.areaH

            color: '#88888888'
        }

        MouseArea
        {
            id: selectionArea

            anchors.fill: parent

            propagateComposedEvents: true

            onPressed:
            {
                if (mouse.button == Qt.LeftButton && (mouse.modifiers & Qt.ShiftModifier))
                {
                    active = true;
                    x1     = mouse.x;
                    y1     = mouse.y;
                    x2     = mouse.x;
                    y2     = mouse.y;

                    mouse.accepted = true;
                } else {
                    mouse.accepted = false;
                }
            }

            onReleased:
            {
                if (mouse.button == Qt.LeftButton && active)
                {
                    active = false;

                    gloperatePipeline.setValue('root', 'selectionArea', [areaX1, areaY1, areaX2, areaY2]);
                }
            }

            onPositionChanged:
            {
                if (active)
                {
                    x2 = mouse.x;
                    y2 = mouse.y;
                }
            }

            property bool active: false
            property int  x1: 0
            property int  y1: 0
            property int  x2: 0
            property int  y2: 0
            property int  areaX1: Math.min(x1, x2)
            property int  areaY1: Math.min(y1, y2)
            property int  areaX2: Math.max(x1, x2)
            property int  areaY2: Math.max(y1, y2)
            property int  areaW: areaX2 - areaX1 + 1
            property int  areaH: areaY2 - areaY1 + 1
        }

        TimeLine
        {
            id: timeline

            anchors.left:    parent.left
            anchors.right:   parent.right
            anchors.bottom:  parent.bottom
            anchors.margins: 16
            height:          32

            properties: gloperatePipeline
        }
    }

    // Keyboard shortcuts
    Shortcut
    {
        sequence: "ESC"

        onActivated:
        {
            if (propertyEditors.width == 0) {
                propertyEditors.width = 400;
            } else {
                propertyEditors.width = 0;
            }
        }
    }

    Shortcut
    {
        sequence: "F5"

        onActivated:
        {
            if (gloperate.system.logLevel() < 5)
            {
                console.log('Entering debug mode');
                gloperate.system.setLogLevel(5);
            }

            else
            {
                console.log('Leaving debug mode');
                gloperate.system.setLogLevel(3);
            }
        }
    }

    Shortcut
    {
        sequence: "F10"

        onActivated:
        {
            console.log("========================================");
            console.log("gloperate.canvas0.setValue('root.Trackball', 'pitch',   " + gloperate.canvas0.getValue('root.Trackball', 'pitch')   + ");");
            console.log("gloperate.canvas0.setValue('root.Trackball', 'yaw',     " + gloperate.canvas0.getValue('root.Trackball', 'yaw')     + ");");
            console.log("gloperate.canvas0.setValue('root.Trackball', 'zoom',    " + gloperate.canvas0.getValue('root.Trackball', 'zoom')    + ");");
            console.log("gloperate.canvas0.setValue('root.Trackball', 'rotateX', " + gloperate.canvas0.getValue('root.Trackball', 'rotateX') + ");");
            console.log("gloperate.canvas0.setValue('root.Trackball', 'rotateY', " + gloperate.canvas0.getValue('root.Trackball', 'rotateY') + ");");
            console.log("========================================");
        }
    }

    Component.onCompleted:
    {
        propertyEditors.update();

        console.log("Welcome to GTX climate viewer.");
        console.log("Press F10 to output the current camera position.");
    }
}
