
#version 330


#define NUM_ATTRS 32


////////////////////////////////////////////////////////////////////////////////////////////////////
// Uniforms
////////////////////////////////////////////////////////////////////////////////////////////////////

// Attribute data
uniform samplerBuffer staticAttributes;
uniform samplerBuffer nodeAttributes;
uniform samplerBuffer edgeAttributes;
uniform samplerBuffer lineAttributes;
uniform samplerBuffer filters;
uniform int numStaticAttributes = 0;
uniform int numNodeAttributes   = 0;
uniform int numEdgeAttributes   = 0;
uniform int numLineAttributes   = 0;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Functions
////////////////////////////////////////////////////////////////////////////////////////////////////

// Fetch attribute value
float fetchAttribute(int attr, int index)
{
    if (attr >= numEdgeAttributes + numNodeAttributes + numStaticAttributes) {
        // Get line attribute
        return texelFetch(lineAttributes, index * NUM_ATTRS + attr - numEdgeAttributes - numNodeAttributes - numStaticAttributes).r;
    } else if (attr >= numNodeAttributes + numStaticAttributes) {
        // Get edge attribute
        return texelFetch(edgeAttributes, index * NUM_ATTRS + attr - numNodeAttributes - numStaticAttributes).r;
    } else if (attr >= numStaticAttributes) {
        // Get node attribute
        return texelFetch(nodeAttributes, index * NUM_ATTRS + attr - numStaticAttributes).r;
    } else if (attr >= 0) {
        // Get static attribute
        return texelFetch(staticAttributes, min(index, 4) * NUM_ATTRS + attr).r;
    } else {
        // Default value
        return 0.0;
    }
}

// Read minimum value of attribute <attr>
float minValue(int attr)
{
    return fetchAttribute(attr, 0);
}

// Read maximum value of attribute <attr>
float maxValue(int attr)
{
    return fetchAttribute(attr, 1);
}

// Read mean value of attribute <attr>
float meanValue(int attr)
{
    return fetchAttribute(attr, 2);
}

// Read standart deviation of attribute <attr>
float stdDev(int attr)
{
    return fetchAttribute(attr, 3);
}

// Read attribute <attr> of vertex <index> from texture data (unmodified value)
float attributeValue(int nodeIndex, int edgeIndex, int attr)
{
    int index = (attr >= numNodeAttributes + numStaticAttributes ? edgeIndex : nodeIndex);

    return fetchAttribute(attr, index + 4);
}

// Read attribute <attr> of vertex <index> from texture data (normalized between 0.0 and 1.0)
// Linear normalization
float normalizedAttributeLinear(int nodeIndex, int edgeIndex, int attr)
{
    if (attr >= 0) {
        // Get attribute value
        float value = attributeValue(nodeIndex, edgeIndex, attr);

        // Get attribute statistics
        float min = minValue(attr);
        float max = maxValue(attr);

        // Normalize
        return (value - min) / (max - min);
    } else {
        return 0.0;
    }
}

// Read attribute <attr> of vertex <index> from texture data (normalized between 0.0 and 1.0)
// n-sigma normalization
float normalizedAttribute(int nodeIndex, int edgeIndex, int attr)
{
    if (attr >= 0) {
        // Get attribute value
        float value = attributeValue(nodeIndex, edgeIndex, attr);

        // Get attribute statistics
        float min   = minValue(attr);
        float max   = maxValue(attr);
        float mean  = meanValue(attr);
        float sigma = stdDev(attr);

        // Calculate n-sigma range
        float n = 1;
        min = mean - n * sigma;
        max = mean + n * sigma;

        // Calculate value
        if (abs(max - min) < 0.00001) {
            return clamp(value, 0.0, 1.0);
        } else {
            return (value - min) / (max - min);
        }
    } else {
        return 0.0;
    }
}

// Filter by node attribute values
float filterNodeAttrs(int nodeIndex)
{
    float visible = 1.0;

    for (int i=0; i<numNodeAttributes; i++)
    {
        int attrIndex = numStaticAttributes + i;

        float attrValue = attributeValue(nodeIndex, 0, attrIndex);
        float minValue  = texelFetch(filters, attrIndex * 4 + 0).r;
        float maxValue  = texelFetch(filters, attrIndex * 4 + 1).r;

        if (!(attrValue >= minValue && attrValue <= maxValue)) visible = 0.0;
    }

    return visible;
}

// Filter by edge attribute values
float filterEdgeAttrs(int edgeIndex)
{
    float visible = 1.0;

    for (int i=0; i<numEdgeAttributes; i++)
    {
        int attrIndex = numStaticAttributes + numNodeAttributes + i;

        float attrValue = attributeValue(0, edgeIndex, attrIndex);
        float minValue  = texelFetch(filters, attrIndex * 4 + 0).r;
        float maxValue  = texelFetch(filters, attrIndex * 4 + 1).r;

        if (!(attrValue >= minValue && attrValue <= maxValue)) visible = 0.0;
    }

    return visible;
}

// Filter by node and edge attribute values
float filterAttrs(int nodeIndex, int edgeIndex)
{
    return filterEdgeAttrs(edgeIndex) * filterNodeAttrs(nodeIndex);
}
