
#version 330

#define MAX_VERTICES 6
layout(lines_adjacency) in;
layout(triangle_strip, max_vertices = MAX_VERTICES) out;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Uniforms
////////////////////////////////////////////////////////////////////////////////////////////////////

// Options
uniform float densityFactor = 1.0;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Input from vertex shader
////////////////////////////////////////////////////////////////////////////////////////////////////

in vec3  v_position[4]; // Position
in float v_visible1[4]; // Visibility (timespan #1)
in float v_visible2[4]; // Visibility (timespan #2)
in float v_visible3[4]; // Visibility (timespan #3)
in float v_density [4]; // Density value


////////////////////////////////////////////////////////////////////////////////////////////////////
// Output to fragment shader
////////////////////////////////////////////////////////////////////////////////////////////////////

out float g_density1;
out float g_density2;
out float g_density3;
out float g_pos;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Geometry shader
////////////////////////////////////////////////////////////////////////////////////////////////////

vec3 intersect(vec3 a1, vec3 a2, vec3 b1, vec3 b2)
{
    vec3 r = normalize(a2 - a1);
    vec3 n = cross(b2-b1, vec3(0, 0, 1));
    float t = dot(b1 - a1, n) / dot(r, n);
    vec3 pos = a1 + t * r;
    return pos;
}

void main(void)
{
    float width = 0.02;

    // Check if both vertices are visible
    bool visible1 = (v_visible1[1] * v_visible1[2] >= 1.0);
    bool visible2 = (v_visible2[1] * v_visible2[2] >= 1.0);
    bool visible3 = (v_visible3[1] * v_visible3[2] >= 1.0);
    if (visible1 || visible2 || visible3) {
        // Get vertices of all 4 points
        vec3 v0 = v_position[0].xyz;
        vec3 v1 = v_position[1].xyz;
        vec3 v2 = v_position[2].xyz;
        vec3 v3 = v_position[3].xyz;
        if (v0 == v1) v0 = v1 - normalize(v2 - v1) * width;
        if (v3 == v2) v3 = v2 + normalize(v2 - v1) * width;

        // Get normal for each of the 3 line segments
        vec3 n0 = cross(normalize(v1-v0), vec3(0, 0, 1));
        vec3 n1 = cross(normalize(v2-v1), vec3(0, 0, 1));
        vec3 n2 = cross(normalize(v3-v2), vec3(0, 0, 1));

        // Extrude points by the direction of their normal, then find the intersection points
        vec3 p1 = intersect(v0 - n0*width, v1 - n0*width, v1 - n1*width, v2 - n1*width);
        vec3 p2 = intersect(v0 + n0*width, v1 + n0*width, v1 + n1*width, v2 + n1*width);
        vec3 p3 = intersect(v1 - n1*width, v2 - n1*width, v2 - n2*width, v3 - n2*width);
        vec3 p4 = intersect(v1 + n1*width, v2 + n1*width, v2 + n2*width, v3 + n2*width);

        // The density is the same at all points (not counting the gradient, which is applied later)
        g_density1 = v_visible1[1] * v_density[1] * densityFactor;
        g_density2 = v_visible2[1] * v_density[1] * densityFactor;
        g_density3 = v_visible3[1] * v_density[1] * densityFactor;

        // Extrude line segment into a filled rectangle

        // Position #1
        gl_Position =  vec4(p1, 1.0);
        g_pos       =  1.0;
        EmitVertex();

        // Position #2
        gl_Position =  vec4(p2, 1.0);
        g_pos       = -1.0;
        EmitVertex();

        // Position #3
        gl_Position =  vec4(p3, 1.0);
        g_pos       =  1.0;
        EmitVertex();

        // Position #4
        gl_Position =  vec4(p4, 1.0);
        g_pos       = -1.0;
        EmitVertex();

        EndPrimitive();
    }
}
