
#version 330


#define MAX_VERTICES 64
layout(lines) in;
layout(line_strip, max_vertices = MAX_VERTICES) out;

#define M_PI   3.1415926535897932384626433832795
#define RADIUS 0.2  // 400000.0


////////////////////////////////////////////////////////////////////////////////////////////////////
// Uniforms
////////////////////////////////////////////////////////////////////////////////////////////////////

// Matrices
uniform mat4 viewProjectionMatrix;

// Options
uniform float arcScale         = 10.0;
uniform float arcTess          = 64.0;
uniform bool  edgesOnTop       = false;
uniform bool  directConnection = true;
uniform float edgeContextColor = 0.1;

////////////////////////////////////////////////////////////////////////////////////////////////////
// Input from vertex shader
////////////////////////////////////////////////////////////////////////////////////////////////////

flat in int   v_nodeId[2];  // Node ID
flat in int   v_edgeId[2];  // Edge ID
     in vec4  v_pos[2];     // Position (world space)
     in vec3  v_latlon[2];  // Position (lat/lon/height)
     in float v_visible[2]; // Edge visiblity
     in float v_width[2];   // Edge width (0..1)
     in float v_color[2];   // Edge color (0..1, mapped on color map)


////////////////////////////////////////////////////////////////////////////////////////////////////
// Output to fragment shader
////////////////////////////////////////////////////////////////////////////////////////////////////

flat          out int   g_edgeId;    // Edge ID
noperspective out vec2  g_screenpos; // Position (screen space)
              out float g_width;     // Edge width (0..1)
              out vec4  g_color;     // Edge color (0..1, mapped on color map)


////////////////////////////////////////////////////////////////////////////////////////////////////
// Imported functions
////////////////////////////////////////////////////////////////////////////////////////////////////

vec4  mapColor(float v);
float pixelLength(vec4 clipStart, vec4 clipEnd);
vec2  clipToPixelSpace(vec4 clipCoords);
vec4  fromLatLon(vec3 latLonHeight);
vec3  correctHeight(vec3 absolutePos, float heightDelta);
bool  isEdgeSelected(int nodeId1, int nodeId2);


////////////////////////////////////////////////////////////////////////////////////////////////////
// Geometry shader
////////////////////////////////////////////////////////////////////////////////////////////////////

void main(void)
{
    // Ignore edge if it is marked disabled
    if (v_visible[0] * v_visible[1] == 0.0)
        return;

    // Check if edge is selected
    bool selected = isEdgeSelected(v_nodeId[0], v_nodeId[1]);

    // Calculate distance between the two coordinates
    float lat1  = v_latlon[0].x;
    float lat2  = v_latlon[1].x;
    float lon1  = v_latlon[0].y;
    float lon2  = v_latlon[1].y;
    float polar = sin(lat1)*sin(lat2) + cos(lat1)*cos(lat2) * cos(lon2-lon1);
    float dist  = acos(polar);

    // Calculate arc height by edge length
    float arc = dist * 0.1 * arcScale;

    // Compute number of parts - make sure it's even or 1
    vec4 clipPos1 = viewProjectionMatrix * v_pos[0];
    vec4 clipPos2 = viewProjectionMatrix * v_pos[1];
    float lineLength = pixelLength(clipPos1, clipPos2);
    int tess = int(arcTess) - 1;
    int maxParts = tess - tess % 2;
    int numParts = clamp(2 * int(lineLength / 20), 1, maxParts);
    for (int i=0; i<=numParts; i++) {
        float t = float(i) / float(numParts);

        vec4 pos;
        vec4 clipPos;
        float h;
        float width;
        float color;

        if (directConnection) {
            // Interpolate position and attributes
            pos   = mix(v_pos   [0]  , v_pos   [1]  , t);
            h     = mix(v_latlon[0].z, v_latlon[1].z, t);
            width = mix(v_width [0]  , v_width [1]  , t);
            color = mix(v_color [0]  , v_color [1]  , t);

            // Correct height, add arc
            float arcHeight = sin(t*M_PI) * arc;
            pos = vec4(correctHeight(pos.xyz, h+arcHeight * RADIUS), 1.0);

            // Convert to clip position
            clipPos = viewProjectionMatrix * pos;
        } else {
            // Choose shorter path on the globe
            if (lat1 < lat2 && (lat1 + 2*M_PI - lat2) < (lat2 - lat1)) {
                lat1 = lat1 + 2*M_PI;
            } else if (lat2 < lat1 && (lat2 + 2*M_PI - lat1) < (lat1 - lat2)) {
                lat2 = lat2 + 2*M_PI;
            }
            if (lon1 < lon2 && (lon1 + 2*M_PI - lon2) < (lon2 - lon1)) {
                lon1 = lon1 + 2*M_PI;
            } else if (lon2 < lon1 && (lon2 + 2*M_PI - lon1) < (lon1 - lon2)) {
                lon2 = lon2 + 2*M_PI;
            }

            // Interpolate position and attributes
            float lat = mix(lat1, lat2, t);
            float lon = mix(lon1, lon2, t);
            if (abs(lat1 - lat2) < abs(lat2 - lat1)) {
            } else {
            }

            h     = mix(v_latlon[0].z, v_latlon[1].z, t);
            width = mix(v_width [0]  , v_width [1]  , t);
            color = mix(v_color [0]  , v_color [1]  , t);
            pos   = fromLatLon(vec3(lat, lon, h));

            // Correct height, add arc
            float arcHeight = sin(t*M_PI) * arc;
            pos = vec4(correctHeight(pos.xyz, h+arcHeight * RADIUS), 1.0);

            // Convert to clip position
            clipPos = viewProjectionMatrix * pos;
        }

        // Adjust depth
        if (edgesOnTop)
        {
            clipPos.z -= 0.01;
        }

        // Set output variables
        gl_Position  = clipPos;
        g_edgeId     = v_edgeId[0];
        g_screenpos  = clipToPixelSpace(clipPos);
        g_width      = width;
        g_color      = vec4(mapColor(color).rgb, 1.0);
        if (!selected) {
            g_color = vec4(edgeContextColor, edgeContextColor, edgeContextColor, edgeContextColor);
        }
        EmitVertex();
    }

    EndPrimitive();
}
