
#version 330


layout(points) in;
layout(triangle_strip, max_vertices = 4) out;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Uniforms
////////////////////////////////////////////////////////////////////////////////////////////////////

// Matrices
uniform mat4 projectionMatrix;

// Options
uniform float minNodeSize      = 0.01;
uniform float maxNodeSize      = 0.02;
uniform float nodeScale        = 1.0;
uniform float nodeContextColor = 0.4;

////////////////////////////////////////////////////////////////////////////////////////////////////
// Input from vertex shader
////////////////////////////////////////////////////////////////////////////////////////////////////

     in vec4  v_pos[1];     // Position
flat in int   v_id[1];      // Node ID
     in float v_visible[1]; // Node visibility
     in float v_color[1];   // Node color (0..1, mapped on color map)
     in float v_size[1];    // Node size (0..1)


////////////////////////////////////////////////////////////////////////////////////////////////////
// Output to fragment shader
////////////////////////////////////////////////////////////////////////////////////////////////////

flat out int   g_id;      // ID
     out vec3  g_pos;     // Fragment position
     out vec3  g_center;  // Sphere center
     out vec3  g_up;      // Up vector
     out vec4  g_color;   // Sphere color (0..1, mapped on color map)
     out float g_radius;  // Sphere radius
     out float g_hovered; // Is the element currently hovered by the mouse?


////////////////////////////////////////////////////////////////////////////////////////////////////
// Imported functions
////////////////////////////////////////////////////////////////////////////////////////////////////

vec4 mapColor(float v);
bool isSelected(int nodeId);
bool isHovered(int nodeId);


////////////////////////////////////////////////////////////////////////////////////////////////////
// Geometry shader
////////////////////////////////////////////////////////////////////////////////////////////////////

void main(void)
{
    // Ignore vertex if it is marked invisible
    if (v_visible[0] == 0.0) return;

    // Check if node is selected
    bool selected = isSelected(v_id[0]);

    // Get mapping configuration
    float size  = v_size[0];
    vec4  color = mapColor(v_color[0]);
    if (!selected) {
        size  = 0.0;
        color = vec4(nodeContextColor, nodeContextColor, nodeContextColor, nodeContextColor);
    }

    // Define sphere: center and radius
    vec3  M = v_pos[0].xyz;
    float r = (minNodeSize + (maxNodeSize - minNodeSize) * size) * nodeScale;
    vec3  u = -M;

    // Calculate billboard
    vec3 dir = normalize(M);
    vec3 B = M - dir*r;
    vec3 dx = cross(vec3(0.0, 1.0, 0.0), dir);
    vec3 dy = cross(dx, dir);
    dx *= r;
    dy *= r;

    g_pos = B-dx+dy;
    g_center = M;
    g_up = u;
    g_radius = r;
    gl_Position = projectionMatrix * vec4(g_pos, 1.0);
    g_color = color;
    g_id = v_id[0];
    g_hovered = isHovered(v_id[0]) ? 1.0 : 0.0;

    EmitVertex();

    g_pos = B-dx-dy;
    g_center = M;
    g_up = u;
    g_radius = r;
    gl_Position = projectionMatrix * vec4(g_pos, 1.0);
    g_color = color;
    g_id = v_id[0];
    g_hovered = isHovered(v_id[0]) ? 1.0 : 0.0;
    EmitVertex();

    g_pos = B+dx+dy;
    g_center = M;
    g_up = u;
    g_radius = r;
    gl_Position = projectionMatrix * vec4(g_pos, 1.0);
    g_color = color;
    g_id = v_id[0];
    g_hovered = isHovered(v_id[0]) ? 1.0 : 0.0;
    EmitVertex();

    g_pos = B+dx-dy;
    g_center = M;
    g_up = u;
    g_radius = r;
    gl_Position = projectionMatrix * vec4(g_pos, 1.0);
    g_color = color;
    g_id = v_id[0];
    g_hovered = isHovered(v_id[0]) ? 1.0 : 0.0;
    EmitVertex();
}
