
#version 140
#extension GL_ARB_explicit_attrib_location : require

#define KERNEL_SIZE 16


////////////////////////////////////////////////////////////////////////////////////////////////////
// Uniforms
////////////////////////////////////////////////////////////////////////////////////////////////////

uniform sampler2D mapColor;
uniform sampler2D mapDepth;
uniform sampler2D overlayColor;
uniform sampler2D overlayDepth;
uniform vec4      viewport;
uniform bool      useFxaa           = true;
uniform bool      useUnsharpMasking = true;
uniform bool      useHighlighting   = true;
uniform float     hudAlpha          = 0.0;
uniform bool      logoVisible       = true;
uniform float     selectedId        = 0.0;
uniform float     time              = 0.0;
uniform float     unsharpFactor     = 1.0;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Input from vertex shader
////////////////////////////////////////////////////////////////////////////////////////////////////

in vec2 v_uv;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Output to frame buffer
////////////////////////////////////////////////////////////////////////////////////////////////////

layout (location = 0) out vec4 fragColor;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Fragment shader
////////////////////////////////////////////////////////////////////////////////////////////////////

void calculateAO(
        in    vec2  texCoord,
        in    float threshold,
        in    float aoCap,
        in    float aoMultiplier,
        in    float depth,
        inout float ao,
        inout int pixelsCaculated)
{
    // read depth map
    float d = texture(overlayDepth, texCoord).x;

    // if difference of distances below treshold...
    if (abs(d-depth) < threshold) {
        // map to AO amount
        ao += min(aoCap, max(0.0, depth-d) * aoMultiplier);

        // propagate to sum
        pixelsCaculated++;
    }
}

vec4 unsharpMasking(in vec2 texCoord)
{
    vec2 poisson[16];
    poisson[0] = vec2(0.007937789, 0.73124397);
    poisson[1] = vec2(-0.10177308, -0.6509396);
    poisson[2] = vec2(-0.9906806, -0.63400936);
    poisson[3] = vec2(-0.5583586, -0.3614012);
    poisson[4] = vec2(0.7163085, 0.22836149);
    poisson[5] = vec2(-0.65210974, 0.37117887);
    poisson[6] = vec2(-0.12714535, 0.112056136);
    poisson[7] = vec2(0.48898065, -0.66669613);
    poisson[8] = vec2(-0.9744036, 0.9155904);
    poisson[9] = vec2(0.9274436, -0.9896486);
    poisson[10] = vec2(0.9782181, 0.90990245);
    poisson[11] = vec2(0.96427417, -0.25506377);
    poisson[12] = vec2(-0.5021933, -0.9712455);
    poisson[13] = vec2(0.3091557, -0.17652994);
    poisson[14] = vec2(0.4665941, 0.96454906);
    poisson[15] = vec2(-0.461774, 0.9360856);

    float ao = 0.0;
    vec2 t = vec2(1.0) / viewport.zw;
    float threshold = 1.5;

    int pixelsCaculated = 0;
    float aoCap = 0.50;
    float aoMultiplier= 5000.0;
    float depth = texture(overlayDepth, texCoord).x;

    threshold    *= unsharpFactor;
    aoCap        *= unsharpFactor;
    aoMultiplier *= unsharpFactor;

    int iterations = 5;
    int kernelSize = KERNEL_SIZE;
    for(int i = 0; i < iterations; i++) {
        // Iterate over kernel
        for(int j = 0; j < kernelSize; j++) {
            calculateAO(
                texCoord + poisson[j] * t,
                threshold,
                aoCap,
                aoMultiplier,
                depth,
                ao,
                pixelsCaculated);
        }

        t*=2.0;
        aoMultiplier*= 1.5;
        //kernelSize = 4;
    }

    ao /= float(pixelsCaculated);

    float valScaled = 1.0 - ao; //mix(minValue, maxValue, 1.0-ao);

    return texture(overlayColor, texCoord) * vec4(valScaled, valScaled, valScaled, 1.0);
}

// Get mixed color between map and overlay
vec4 mixedColor(in vec2 texCoord)
{
    // Map layer
    vec4  c1 = texture(mapColor, texCoord);

    // Overlays
    vec4  c2;
    if (useUnsharpMasking) c2 = unsharpMasking(texCoord);
    else                   c2 = texture(overlayColor, texCoord);

    // Mix map and overlays, do manual depth-buffer test
    return mix(c1, c2, c2.w);
}

// FXAA shader, GLSL code adapted from:
// http://horde3d.org/wiki/index.php5?title=Shading_Technique_-_FXAA
// Whitepaper describing the technique:
// http://developer.download.nvidia.com/assets/gamedev/files/sdk/11/FXAA_WhitePaper.pdf
vec4 fxaa(in vec2 texCoord) {
    // The inverse of the texture dimensions along X and Y
    vec2 texcoordOffset = vec2(1.0/viewport.z, 1.0/viewport.w);

    // The parameters are hardcoded for now, but could be
    // made into uniforms to control fromt he program.
    float FXAA_SPAN_MAX   = 8.0;
    float FXAA_REDUCE_MUL = 1.0/8.0;
    float FXAA_REDUCE_MIN = (1.0/128.0);

    vec3  rgbNW  = mixedColor(texCoord.xy + (vec2(-1.0, -1.0) * texcoordOffset)).xyz;
    vec3  rgbNE  = mixedColor(texCoord.xy + (vec2(+1.0, -1.0) * texcoordOffset)).xyz;
    vec3  rgbSW  = mixedColor(texCoord.xy + (vec2(-1.0, +1.0) * texcoordOffset)).xyz;
    vec3  rgbSE  = mixedColor(texCoord.xy + (vec2(+1.0, +1.0) * texcoordOffset)).xyz;
    vec3  rgbM   = mixedColor(texCoord.xy).xyz;
    float alphaM = mixedColor(texCoord.xy).a;

    vec3 luma = vec3(0.299, 0.587, 0.114);
    float lumaNW = dot(rgbNW, luma);
    float lumaNE = dot(rgbNE, luma);
    float lumaSW = dot(rgbSW, luma);
    float lumaSE = dot(rgbSE, luma);
    float lumaM = dot( rgbM, luma);

    float lumaMin = min(lumaM, min(min(lumaNW, lumaNE), min(lumaSW, lumaSE)));
    float lumaMax = max(lumaM, max(max(lumaNW, lumaNE), max(lumaSW, lumaSE)));

    vec2 dir;
    dir.x = -((lumaNW + lumaNE) - (lumaSW + lumaSE));
    dir.y = ((lumaNW + lumaSW) - (lumaNE + lumaSE));

    float dirReduce = max((lumaNW + lumaNE + lumaSW + lumaSE) * (0.25 * FXAA_REDUCE_MUL), FXAA_REDUCE_MIN);

    float rcpDirMin = 1.0/(min(abs(dir.x), abs(dir.y)) + dirReduce);

    dir = min(vec2(FXAA_SPAN_MAX, FXAA_SPAN_MAX),
              max(vec2(-FXAA_SPAN_MAX, -FXAA_SPAN_MAX), dir * rcpDirMin)) * texcoordOffset;

    vec3 rgbA = (1.0/2.0) * (
        mixedColor(texCoord.xy + dir * (1.0/3.0 - 0.5)).xyz +
        mixedColor(texCoord.xy + dir * (2.0/3.0 - 0.5)).xyz);
    vec3 rgbB = rgbA * (1.0/2.0) + (1.0/4.0) * (
        mixedColor(texCoord.xy + dir * (0.0/3.0 - 0.5)).xyz +
        mixedColor(texCoord.xy + dir * (3.0/3.0 - 0.5)).xyz);
    float lumaB = dot(rgbB, luma);

    vec4 color;
    if((lumaB < lumaMin) || (lumaB > lumaMax)) {
        color.xyz=rgbA;
    } else {
        color.xyz=rgbB;
    }

    color.a = alphaM;
    return color;
}

vec4 crop(sampler2D tex, vec2 pixelPos)
{
    vec2 tc = vec2(
        clamp((pixelPos.x-18.0) / 100.0, 0.0, 1.0),
        clamp((pixelPos.y-(viewport.w-116.0)) / 100.0, 0.0, 1.0)
    );
    vec4 color = texture(tex, tc);
    color.a *= float(logoVisible && pixelPos.x >= 18.0 && pixelPos.x <= 118.0 &&
                    pixelPos.y >= viewport.w-116.0 && pixelPos.y <= viewport.w-16.0);
    return color;
}

void main()
{
    // Combined color from background map and overlays
    vec4 color;
    if (useFxaa) color = fxaa(v_uv);
    else         color = mixedColor(v_uv);

    // Final color
    fragColor = color;
}
