
#pragma once


#include <QVector>
#include <QStringList>
#include <QColor>
#include <QPixmap>

#include <gtxlib-core/gtxlib-core_api.h>


namespace globjects
{
    class Texture;
}


namespace gtx
{


/**
*  @brief
*    Color map that holds 4 colors
*/
class GTXLIB_CORE_API ColorMap
{
public:
    /**
    *  @brief
    *    List available color maps
    *
    *  @param[in] relDataPath
    *    Path to data directory (usually "", unless loaded from plugins)
    *
    *  @return
    *    List of color maps (filenames)
    */
    static QStringList listColorMaps(const std::string & relDataPath);


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param colors
    *    Color map string, or ""
    */
    ColorMap(const std::string & colors = "");

    /**
    *  @brief
    *    Copy constructor
    *
    *  @param source
    *    Source color map
    */
    ColorMap(const ColorMap & source);

    /**
    *  @brief
    *    Destructor
    */
    ~ColorMap();

    /**
    *  @brief
    *    Copy operator
    *
    *  @param source
    *    Source color map
    */
    const ColorMap &operator =(const ColorMap & source);

    /**
    *  @brief
    *    Clear color map
    */
    void clear();

    /**
    *  @brief
    *    Load color from file
    *
    *  @param[in] filename
    *    Name of color description or image file
    *
    *  @remarks
    *    Supports .color files (content of toString()) and ordinary image files
    */
    void load(const std::string & filename);

    /**
    *  @brief
    *    Save color to file
    *
    *  @param[in] filename
    *    Name of color description file
    *
    *  @remarks
    *    Writes the content of toConfigString() into a file
    */
    void save(const std::string & filename) const;

    /**
    *  @brief
    *    Get color map as string
    *
    *  @return
    *    Color map string (e.g. "#000000 #ff0000 #00ff00 #0000ff")
    */
    std::string toString() const;

    /**
    *  @brief
    *    Set color map from string
    *
    *  @param[in] colorMap
    *    Color map string (e.g. "#000000 #ff0000 #00ff00 #0000ff")
    */
    void fromString(const std::string & colorMap);

    /**
    *  @brief
    *    Get color map as config string
    *
    *  @return
    *    Color map string (e.g. "237,248,251; 178,226,226; 102,194,164; 35,139,69;")
    */
    std::string toConfigString() const;

    /**
    *  @brief
    *    Set color map from config string
    *
    *  @param[in] colorMap
    *    Color map string (e.g. "237,248,251; 178,226,226; 102,194,164; 35,139,69;")
    */
    void fromConfigString(const std::string & colorMap);

    /**
    *  @brief
    *    Get color map as string in HTML format
    *
    *  @return
    *    Color map string (e.g. "#000000 #ff0000 #00ff00 #0000ff")
    */
    std::string toHTMLString() const;

    /**
    *  @brief
    *    Set color map from string in HTML format
    *
    *  @param[in] colorMap
    *    Color map string (e.g. "#000000 #ff0000 #00ff00 #0000ff")
    */
    void fromHTMLString(const std::string & colorMap);

    /**
    *  @brief
    *    Check if color map is discrete
    *
    *  @return
    *    'true' if discrete, 'false' if continuous
    */
    bool discrete() const;

    /**
    *  @brief
    *    Set if color map is discrete
    *
    *  @param[in] discrete
    *    'true' if discrete, 'false' if continuous
    */
    void setDiscrete(bool discrete);

    /**
    *  @brief
    *    Check if color map is reversed
    *
    *  @return
    *    'true' if reversed, else 'false'
    */
    bool reversed() const;

    /**
    *  @brief
    *    Set if color map is reversed
    *
    *  @param[in] reversed
    *    'true' if reversed, else 'false'
    */
    void setReversed(bool reversed);

    /**
    *  @brief
    *    Get number of colors
    *    
    *  @return
    *    Number of colors
    */
    int numColors() const;

    /**
    *  @brief
    *    Get color
    *
    *  @param[in] index
    *    Color index (0..3)
    *    
    *  @return
    *    Color
    */
    QColor color(int index) const;

    /**
    *  @brief
    *    Set color
    *
    *  @param[in] index
    *    Color index (0..3)
    *  @param[in] color
    *    Color
    */
    void setColor(int index, const QColor & color);

    /**
    *  @brief
    *    Get color map as pixmap
    *
    *  @return
    *    Pixmap
    */
    QPixmap pixmap() const;

    /**
    *  @brief
    *    Get color map as texture
    *
    *  @return
    *    Texture
    *
    *  @remarks
    *    The returned texture is always a new instance that
    *    has to be managed (deleted) by the caller.
    */
    globjects::Texture * texture() const;


protected:
    std::string color2HTML(const QColor & color) const;
    QColor  HTML2color(const std::string & code) const;


protected:
    QVector<QColor> m_colors;    ///< List of colors
    std::string         m_image;     ///< Name of image file
    bool            m_discrete;  ///< Is the color map discrete (true) or continuous (false)?
    bool            m_reversed;  ///< Is the color map reversed?
};


} // namespace gtx
