
#pragma once


#include <chrono>

#include <gtxlib-core/gtxlib-core_api.h>


namespace gtx
{


/**
*  @brief
*    Time span
*/
class GTXLIB_CORE_API TimeSpan
{
public:
    /**
    *  @brief
    *    Constructor
    */
    TimeSpan();

    /**
    *  @brief
    *    Constructor
    *
    *  @param minTime
    *    Minimum time
    *  @param maxTime
    *    Maximum time
    *  @param minDailyTime
    *    Minimum daily time
    *  @param maxDailyTime
    *    Maximum daily time
    *  @param dailyTime
    *    Use daily time?
    */
    TimeSpan(const std::chrono::time_point & minTime, const std::chrono::time_point & maxTime, const std::chrono::duration & minDailyTime = std::chrono::duration(), const std::chrono::duration & maxDailyTime = std::chrono::duration(), bool dailyTime = false);

    /**
    *  @brief
    *    Constructor
    *
    *  @param timespan
    *    TimeSpan to copy
    */
    TimeSpan(const TimeSpan & timespan);

    /**
    *  @brief
    *    Destructor
    */
    ~TimeSpan();

    /**
    *  @brief
    *    Comparison operator
    */
    bool operator ==(const TimeSpan & timespan) const;

    /**
    *  @brief
    *    Assignment operator
    */
    TimeSpan & operator =(const TimeSpan & timespan);

    /**
    *  @brief
    *    Check if time span is empty
    *
    *  @return
    *    'true' if time span is empty, else 'false'
    */
    bool empty() const;

    /**
    *  @brief
    *    Clear time span
    */
    void clear();

    /**
    *  @brief
    *    Get minimum time
    *
    *  @return
    *    Minimum time
    */
    std::chrono::time_point minTime() const;

    /**
    *  @brief
    *    Set minimum time
    *
    *  @param time
    *    Minimum time
    */
    void setMinTime(const std::chrono::time_point & time);

    /**
    *  @brief
    *    Get maximum time
    *
    *  @return
    *    Maximum time
    */
    std::chrono::time_point maxTime() const;

    /**
    *  @brief
    *    Set maximum time
    *
    *  @param time
    *    Maximum time
    */
    void setMaxTime(const std::chrono::time_point & time);

    /**
    *  @brief
    *    Expand time span
    *
    *  @param time
    *    Time
    */
    void expand(const std::chrono::time_point & time);

    /**
    *  @brief
    *    Get minimum daily time
    *
    *  @return
    *    Minimum daily time
    */
    std::chrono::duration minDailyTime() const;

    /**
    *  @brief
    *    Set minimum daily time
    *
    *  @param time
    *    Minimum daily time
    */
    void setMinDailyTime(const std::chrono::duration & time);

    /**
    *  @brief
    *    Get maximum daily time
    *
    *  @return
    *    Maximum daily time
    */
    std::chrono::duration maxDailyTime() const;

    /**
    *  @brief
    *    Set maximum daily time
    *
    *  @param time
    *    Maximum daily time
    */
    void setMaxDailyTime(const std::chrono::duration & time);

    /**
    *  @brief
    *    Get daily time state
    *
    *  @return
    *    'true' if daily time is used, else 'false'
    */
    bool dailyTime() const;

    /**
    *  @brief
    *    Check if time is inside the time span
    *
    *  @param time
    *    Time
    *
    *  @return
    *    'true' if time is inside the time span, else 'false'
    */
    bool isInside(const std::chrono::time_point & time) const;

    /**
    *  @brief
    *    Get number of seconds from the beginning of the time span to the given point in time
    *
    *  @param time
    *    Time
    *
    *  @return
    *    Elapsed time (in seconds)
    */
    long secondsSinceStart(const std::chrono::time_point & time) const;


protected:
    bool                    m_empty;        ///< Is the time span still empty?
    bool                    m_dailyTime;    ///< Is a daily time defined?
    std::chrono::time_point m_minTime;      ///< Minimum time
    std::chrono::time_point m_maxTime;      ///< Maximum time
    std::chrono::duration   m_minDailyTime; ///< Minimum daily time
    std::chrono::duration   m_maxDailyTime; ///< Minimum daily time
};


} // namespace gtx
