
#pragma once


#include <memory>
#include <string>
#include <vector>
#include <map>
#include <unordered_map>

#include <gtxlib-core/data/Item.h>
#include <gtxlib-core/data/AttributeInfo.h>


namespace gtx
{


/**
*  @brief
*    Collection of data items that belong to a group (e.g., nodes, or edges)
*/
class GTXLIB_CORE_API AbstractItemGroup
{
public:
    /**
    *  @brief
    *    Constructor
    */
    AbstractItemGroup();

    /**
    *  @brief
    *    Destructor
    */
    virtual ~AbstractItemGroup();

    /**
    *  @brief
    *    Get attributes
    *
    *  @return
    *    List of attributes
    */
    const std::vector<std::string> & attributes() const;

    /**
    *  @brief
    *    Get typed attribute info
    *
    *  @param[in] name
    *    Attribute name
    *
    *  @return
    *    Attribute info, null if it doesn't exist
    *
    *  @remarks
    *    The type is not checked, so the caller needs to make sure that the type is correct.
    */
    const AttributeInfo * attribute(const std::string & name) const;

    /**
    *  @brief
    *    Get or create typed attribute info
    *
    *  @param[in] name
    *    Attribute name
    *
    *  @return
    *    Attribute info (never null)
    *
    *  @remarks
    *    If the attribute already exists, it is returned and cast to the specified type.
    *    The type is not checked, so the caller needs to make sure that the type is correct.
    *    If the attribute does not yet exist, a new attribute is created.
    */
    AttributeInfo * attribute(const std::string & name);

    /**
    *  @brief
    *    Clear group
    *
    *  @remarks
    *    Will destroy all owned items and clear the item list
    */
    virtual void clear() = 0;

    /**
    *  @brief
    *    Check if group is empty
    *
    *  @return
    *    'true' if group is empty, else 'false'
    */
    virtual bool empty() const = 0;

    /**
    *  @brief
    *    Get number of items
    *
    *  @return
    *    Number of items
    */
    virtual size_t size() const = 0;

    /**
    *  @brief
    *    Check if group contains a specific item
    *
    *  @param[in] id
    *    Item ID
    *
    *  @return
    *    'true' if item is in the group, else 'false'
    */
    virtual bool contains(id_t id) const = 0;

    /**
    *  @brief
    *    Get items (untyped)
    *
    *  @return
    *    Items
    */
    virtual const std::unordered_map<id_t, Item *> & items() const = 0;

    /**
    *  @brief
    *    Get item
    *
    *  @param[in] id
    *    Item ID
    *
    *  @return
    *    Item, null if no such item exists
    */
    virtual const Item * item(id_t id) const = 0;

    /**
    *  @brief
    *    Get item
    *
    *  @param[in] id
    *    Item ID
    *
    *  @return
    *    Item, null if no such item exists
    */
    virtual Item * item(id_t id) = 0;


protected:
    std::map<std::string, std::unique_ptr<AttributeInfo>> m_attributes; ///< List of attributes, accessible by name
    std::vector<std::string>                              m_names;      ///< List of attribute names
};


} // namespace gtx
