
#pragma once


#include <string>
#include <map>

#include <gtxlib-core/gtxlib-core_api.h>


namespace gtx
{


/**
*  @brief
*    Aircraft database
*/
class GTXLIB_CORE_API AircraftDB
{
public:
    /**
    *  @brief
    *    ICAO Aircraft definition
    */
    struct ICAOAircraft
    {
        std::string id;                 // ICAO aircraft code
        std::string manufacturer;       // Manufacturer name
        std::string model;              // Model name
        std::string description;        // Description
        std::string type;               // Aircraft type (e.g., "landplane")
        std::string engine;             // Engine type
        int         numEngines;         // Number of engines
        std::string weightClass;        // Weight class (e.g., "Heavy")
        std::string classCode;          // Class code (e.g., "Heavy 4 engine Jet")
        int         numPassengers;      // Number of passengers
        int         numSeats;           // Number of seats
        std::string seatsSpecificity;   // Seats specificity (e.g., "aircraft_class")
        float       m3;                 // m3
        std::string m3Units;            // Units for m3 field (e.g., "kilograms_per_cubic_nautical_mile")
        float       m2;                 // m2
        std::string m2Units;            // Units for m2 field (e.g., "kilograms_per_square_nautical_mile")
        float       m1;                 // m1
        std::string m1Units;            // Units for m1 field (e.g., "kilograms_per_nautical_mile")
        float       b;                  // b
        std::string bUnits;             // Units for b field (e.g., "kilograms")
        std::string fuelUseSpecificity; // Fuel use specificity (e.g., "aircraft")
    };

    /**
    *  @brief
    *    Aircraft definition
    */
    struct Aircraft
    {
        std::string id;                         // ACFT ID
        std::string description;                // Description
        std::string source;                     // Source Of data
        std::string engineType;                 // Engine type
        int         numEngines;                 // Number of engines
        std::string weightClass;                // Weight class
        std::string ownerCategory;              // Owner category
        int         maxTakeoffWeight;           // Max gross takeoff weight
        int         maxLandingWeight;           // Max gross landing weight
        int         maxLandingDistance;         // Max landing distance
        int         maxSeaLevelStaticThrust;    // Max sea level static thrust
        int         noiseChapter;               // Noise chaper
        std::string npdID;                      // NPD ID
        std::string powerParameter;             // Power parameter
        int         approachSpectralClassID;    // Approach spectral class ID
        int         departureSpectralClassID;   // Departure spectral class ID
        std::string lateralDirectivityID;       // Lateral directivity identifier
    };


public:
    /**
    *  @brief
    *    Constructor
    */
    AircraftDB();

    /**
    *  @brief
    *    Destructor
    */
    ~AircraftDB();

    /**
    *  @brief
    *    Get aircraft from aircraft DB
    *
    *  @param[in] planeType
    *    Plane type
    *
    *  @return
    *    Aircraft description
    */
    const Aircraft & aircraft(const std::string & planeType) const;

    /**
    *  @brief
    *    Get aircraft from ICAO aircraft DB
    *
    *  @param[in] planeType
    *    Plane type
    *
    *  @return
    *    ICAO Aircraft description
    */
    const ICAOAircraft & icaoAircraft(const std::string & planeType) const;


protected:
    /**
    *  @brief
    *    Load aircraft DB
    */
    void loadAircraftDB();

    /**
    *  @brief
    *    Load ICAO aircraft DB
    */
    void loadICAOAircraftDB();


protected:
    std::map<std::string, ICAOAircraft> m_icao;      ///< List of ICAO aircraft definitions
    std::map<std::string, Aircraft>     m_aircrafts; ///< List of aircraft definitions
};


} // namespace gtx
