
#pragma once


#include <string>

#include <gtxlib-core/gtxlib-core.h>


namespace gtx
{


/**
*  @brief
*    Information about an attribute
*/
class GTXLIB_CORE_API AttributeInfo
{
public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] name
    *    Attribute name
    */
    AttributeInfo(const std::string & name);

    /**
    *  @brief
    *    Destructor
    */
    ~AttributeInfo();

    /**
    *  @brief
    *    Get name
    *
    *  @return
    *    Attribute name
    */
    const std::string & name() const;

    /**
    *  @brief
    *    Check if attribute statistics shall be calculated (default: true)
    *
    *  @return
    *    'true' if statistics shall be calculated, else 'false'
    */
    bool autoCalculateStatistics() const;

    /**
    *  @brief
    *    Set if attribute statistics shall be calculated (default: true)
    *
    *  @param[in] statistics
    *    'true' if statistics shall be calculated, else 'false'
    */
    void setAutoCalculateStatistics(bool statistics);

    /**
    *  @brief
    *    Get minimum value
    *
    *  @return
    *    Minimum value
    */
    float minimumValue() const;

    /**
    *  @brief
    *    Set minimum value
    *
    *  @param[in] value
    *    Minimum value
    */
    void setMinimumValue(float value);

    /**
    *  @brief
    *    Get maximum value
    *
    *  @return
    *    Maximum value
    */
    float maximumValue() const;

    /**
    *  @brief
    *    Set maximum value
    *
    *  @param[in] value
    *    Maximum value
    */
    void setMaximumValue(float value);

    /**
    *  @brief
    *    Get sum of all values
    *
    *  @return
    *    Sum
    */
    float sum() const;

    /**
    *  @brief
    *    Set sum of all values
    *
    *  @param[in] value
    *    Sum
    */
    void setSum(float value);

    /**
    *  @brief
    *    Get number of values
    *
    *  @return
    *    Number of values
    */
    unsigned long numValues() const;

    /**
    *  @brief
    *    Set number of values
    *
    *  @param[in] numValues
    *    Number of values
    */
    void setNumValues(unsigned long numValues);

    /**
    *  @brief
    *    Get mean of all values
    *
    *  @return
    *    Mean value
    */
    float meanValue() const;

    /**
    *  @brief
    *    Set mean of all values
    *
    *  @param[in] value
    *    Mean value
    */
    void setMeanValue(float value);

    /**
    *  @brief
    *    Get standard deviation
    *
    *  @return
    *    Standard deviation
    */
    float standardDeviation() const;

    /**
    *  @brief
    *    Set standard deviation
    *
    *  @param[in] value
    *    Standard deviation
    */
    void setStandardDeviation(float value);

    /**
    *  @brief
    *    Update value statistics by adding a new value
    *
    *  @param[in] value
    *    Attribute value
    */
    void addValue(float value);


protected:
    void calculateMean();


protected:
    std::string   m_name;        ///< Attribute name
    bool          m_statistics;  ///< 'true' if statistics shall be calculated, else 'false'
    unsigned long m_numValues;   ///< Number of values
    float         m_min;         ///< Minimum value
    float         m_max;         ///< Maximum value
    float         m_sum;         ///< Sum of all values
    float         m_mean;        ///< Mean of all values
    float         m_stdDev;      ///< Standard deviation
};


} // namespace gtx
