
#pragma once


#include <string>
#include <vector>
#include <unordered_map>

#include <gtxlib-core/base/Area.h>
#include <gtxlib-core/data/ItemGroup.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>
#include <gtxlib-core/data/Line.h>


namespace gtx
{


/**
*  @brief
*    Container for a data set
*/
class GTXLIB_CORE_API DataSet
{
public:
    /**
    *  @brief
    *    Constructor
    */
    DataSet();

    /**
    *  @brief
    *    Destructor
    */
    virtual ~DataSet();

    /**
    *  @brief
    *    Clear data set
    */
    void clear();

    /**
    *  @brief
    *    Check if data set is empty
    *
    *  @return
    *    'true' if data set is empty, else 'false'
    */
    bool empty() const;

    /**
    *  @brief
    *    Print out statistics
    */
    void printStatistics();

    /**
    *  @brief
    *    Get filename
    *
    *  @return
    *    Filename
    */
    std::string filename() const;

    /**
    *  @brief
    *    Set filename
    *
    *  @param[in] filename
    *    Filename
    */
    void setFilename(const std::string & filename);

    /**
    *  @brief
    *    Get area of whole data set
    *
    *  @return
    *    Area
    */
    const Area & area() const;

    /**
    *  @brief
    *    Get static items
    *
    *  @return
    *    Item list
    */
    const ItemGroup<Item> & statics() const;

    /**
    *  @brief
    *    Get nodes
    *
    *  @return
    *    Item list
    */
    const ItemGroup<Node> & nodes() const;

    /**
    *  @brief
    *    Get nodes
    *
    *  @return
    *    Item list
    */
    ItemGroup<Node> & nodes();

    /**
    *  @brief
    *    Get edges
    *
    *  @return
    *    Item list
    */
    const ItemGroup<Edge> & edges() const;

    /**
    *  @brief
    *    Get edges
    *
    *  @return
    *    Item list
    */
    ItemGroup<Edge> & edges();

    /**
    *  @brief
    *    Get lines
    *
    *  @return
    *    Item list
    */
    const ItemGroup<Line> & lines() const;

    /**
    *  @brief
    *    Get lines
    *
    *  @return
    *    Item list
    */
    ItemGroup<Line> & lines();

    /**
    *  @brief
    *    Get minimum time
    *
    *  @return
    *    Minimum time stamp
    */
    float minimumTime() const;

    /**
    *  @brief
    *    Get maximum time
    *
    *  @return
    *    Maximum time stamp
    */
    float maximumTime() const;

    /**
    *  @brief
    *    Fix item IDs (begin with 0 and remove gaps)
    */
    void fixIds();

    /**
    *  @brief
    *    Calculate the geographical length of edges
    */
    void calculateEdgeLength();

    /**
    *  @brief
    *    Add standard node attributes ('Zero', 'One', 'Lat', 'Lon')
    */
    void addStandardAttributes();

    /**
    *  @brief
    *    Analyze trajectories and add standard attributes
    */
    void analyzeTrajectories();

    /**
    *  @brief
    *    Calculate attribute statistics (min, max, mean, etc.)
    *
    *  @remarks
    *    This function has to be called once after loading a data set.
    *    It for example calculates ranges for the data attributes,
    *    extracts attribute information, and prepares the data set
    *    for GPU-based rendering.
    */
    void calculateAttributeStatistics();

    /**
    *  @brief
    *    Get map layers
    *
    *  @return
    *    Additional map layers that are displayed on top of the map
    */
    const std::vector<std::string> & mapLayers() const;

    /**
    *  @brief
    *    Set map layers
    *
    *  @param[in] mapLayers
    *    Additional map layers that are displayed on top of the map
    */
    void setMapLayers(std::vector<std::string> & mapLayers);


protected:
    // Meta info
    std::string              m_filename;  ///< File from which the data set has been loaded
    float                    m_minTime;   ///< Minimum time stamp
    float                    m_maxTime;   ///< Maximum time stamp

    // Data items
    ItemGroup<Item>          m_statics;   ///< List of static items
    ItemGroup<Node>          m_nodes;     ///< List of nodes
    ItemGroup<Edge>          m_edges;     ///< List of edges
    ItemGroup<Line>          m_lines;     ///< List of lines
    Area                     m_area;      ///< Area of all elements

    // Additional data
    std::vector<std::string> m_mapLayers; ///< Additional map layers that are displayed on top of the map
};


} // namespace gtx
