
#pragma once


#include <memory>
#include <vector>
#include <map>
#include <string>

#include <gtxlib-core/gtxlib-core.h>
#include <gtxlib-core/base/DataTypes.h>
#include <gtxlib-core/data/Attribute.h>


namespace gtx
{


/**
*  @brief
*    Base class for data item
*/
class GTXLIB_CORE_API Item
{
public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] id
    *    ID
    */
    Item(id_t id = null_id);

    /**
    *  @brief
    *    Destructor
    */
    virtual ~Item();

    /**
    *  @brief
    *    Get ID
    *
    *  @return
    *    ID
    */
    id_t id() const;

    /**
    *  @brief
    *    Set ID
    *
    *  @param id
    *    ID
    */
    void setId(id_t id);

    /**
    *  @brief
    *    Get name
    *
    *  @return
    *    Name
    */
    const std::string & name() const;

    /**
    *  @brief
    *    Set name
    *
    *  @param name
    *    Name
    */
    void setName(const std::string & name);

    /**
    *  @brief
    *    Get attributes
    *
    *  @return
    *    List of attributes
    */
    const std::vector<std::string> & attributes() const;

    /**
    *  @brief
    *    Check if attribute exists
    *
    *  @param[in] name
    *    Attribute name
    *
    *  @return
    *    'true' if such attribute is set, else 'false'
    */
    bool hasAttribute(const std::string & name) const;

    /**
    *  @brief
    *    Get attribute
    *
    *  @param[in] name
    *    Attribute name
    *
    *  @return
    *    Attribute, or null if attribute does not exist
    */
    const Attribute * attribute(const std::string & name) const;

    /**
    *  @brief
    *    Get or create attribute
    *
    *  @param[in] name
    *    Attribute name
    *
    *  @return
    *    Attribute (never null)
    *
    *  @remarks
    *    If the attribute already exists, it is returned.
    *    Otherwise, a new attribute is created.
    */
    Attribute * attribute(const std::string & name);


protected:
    id_t                                              m_id;         ///< Internal ID
    std::string                                       m_name;       ///< Internal name
    std::map<std::string, std::unique_ptr<Attribute>> m_attributes; ///< List of attributes, accessible by name
    std::vector<std::string>                          m_names;      ///< List of attribute names
};


} // namespace gtx
