
#pragma once


#include <memory>
#include <vector>
#include <map>
#include <unordered_map>
#include <algorithm>
#include <string>

#include <gtxlib-core/base/DataTypes.h>

#include <gtxlib-core/data/AttributeInfo.h>
#include <gtxlib-core/data/AbstractItemGroup.h>


namespace gtx
{


/**
*  @brief
*    Collection of data items that belong to a group (e.g., nodes, or edges)
*/
template <typename T>
class GTXLIB_CORE_API ItemGroup : public AbstractItemGroup
{
public:
    /**
    *  @brief
    *    Constructor
    */
    ItemGroup();

    /**
    *  @brief
    *    Destructor
    */
    virtual ~ItemGroup();

    // Virtual AbstractItemGroup interface
    virtual void clear() override;
    virtual bool empty() const override;
    virtual size_t size() const override;
    virtual bool contains(id_t id) const override;
    virtual const std::unordered_map<id_t, Item *> & items() const override;
    virtual const Item * item(id_t id) const override;
    virtual Item * item(id_t id) override;

    /**
    *  @brief
    *    Get item map
    *
    *  @return
    *    Item map
    */
    const std::unordered_map<id_t, T *> & map() const;

    /**
    *  @brief
    *    Get item
    *
    *  @param[in] id
    *    Item ID
    *
    *  @return
    *    Item, null if no such item exists
    */
    const T * get(id_t id) const;

    /**
    *  @brief
    *    Get item
    *
    *  @param[in] id
    *    Item ID
    *
    *  @return
    *    Item, null if no such item exists
    */
    T * get(id_t id);

    /**
    *  @brief
    *    Create new item
    *
    *  @param[in] id
    *    ID
    *
    *  @return
    *    Item (never null)
    */
    T * create(id_t id = null_id);


protected:
    std::unordered_map<id_t, T *>   m_items;    ///< List of items, accessable by ID
    std::vector<std::unique_ptr<T>> m_ownItems; ///< List of items that are owned by the group
    id_t                            m_firstId;  ///< Lowest ID in this group
    id_t                            m_lastId;   ///< Highest ID in this group
    id_t                            m_nextId;   ///< Next automatic ID
};


} // namespace gtx


#include <gtxlib-core/data/ItemGroup.inl>
