
#pragma once


#include <vector>
#include <string>
#include <memory>

#include <globjects/Texture.h>

#include <gtxlib-core/gtxlib-core_api.h>


namespace gtx
{


class DataSet;
class AbstractItemGroup;


/**
*  @brief
*    Storage to store/access attributes for each item (e.g., nodes or edges) on the GPU
*/
class GTXLIB_CORE_API AttributeStorage
{
public:
    /**
    *  @brief
    *    Constructor
    */
    AttributeStorage();

    /**
    *  @brief
    *    Destructor
    */
    ~AttributeStorage();

    /**
    *  @brief
    *    Create attribute storage for data set
    *
    *  @param[in] dataset
    *    Data set
    */
    void setData(const DataSet & dataset);

    /**
    *  @brief
    *    Clear attribute storage
    */
    void clearData();

    /**
    *  @brief
    *    Get available attributes
    *
    *  @return
    *    List of attributes
    */
    const std::vector<std::string> & attributes() const;

    /**
    *  @brief
    *    Get number of static attributes
    *
    *  @return
    *    Number of static attributes
    */
    int numStaticAttributes() const;

    /**
    *  @brief
    *    Get number of node attributes
    *
    *  @return
    *    Number of node attributes
    *
    *  @remarks
    *    Depending on this value, a shader program can decide whether to fetch
    *    a node or an edge/line attribute, based on the attribute index
    */
    int numNodeAttributes() const;

    /**
    *  @brief
    *    Get number of edge attributes
    *
    *  @return
    *    Number of edge attributes
    */
    int numEdgeAttributes() const;

    /**
    *  @brief
    *    Get number of line attributes
    *
    *  @return
    *    Number of line attributes
    */
    int numLineAttributes() const;

    /**
    *  @brief
    *    Get static attributes texture
    *
    *  @return
    *    Texture
    */
    globjects::Texture * staticAttributesTexture();

    /**
    *  @brief
    *    Get node attributes texture
    *
    *  @return
    *    Texture
    */
    globjects::Texture * nodeAttributesTexture();

    /**
    *  @brief
    *    Get edge attributes texture
    *
    *  @return
    *    Texture
    */
    globjects::Texture * edgeAttributesTexture();

    /**
    *  @brief
    *    Get line attributes texture
    *
    *  @return
    *    Texture
    */
    globjects::Texture * lineAttributesTexture();


protected:
    std::unique_ptr<globjects::Texture> createAttributesTexture(const AbstractItemGroup & group, const std::string & prefix, std::vector<std::string> & attrs, unsigned int & numAttrs);


protected:
    std::vector<std::string>            m_attrs;                   ///< List of attributes
    unsigned int                        m_numStaticAttributes;     ///< Number of static attributes
    unsigned int                        m_numNodeAttributes;       ///< Number of node attributes
    unsigned int                        m_numEdgeAttributes;       ///< Number of edge attributes
    unsigned int                        m_numLineAttributes;       ///< Number of line attributes
    std::unique_ptr<globjects::Texture> m_staticAttributesTexture; ///< Texture containing static attributes
    std::unique_ptr<globjects::Texture> m_nodeAttributesTexture;   ///< Texture containing the node attributes
    std::unique_ptr<globjects::Texture> m_edgeAttributesTexture;   ///< Texture containing the edge attributes
    std::unique_ptr<globjects::Texture> m_lineAttributesTexture;   ///< Texture containing the line attributes
};


} // namespace gtx
