
#include <gtxlib-core/base/ColorMap.h>

#include <QFile>
#include <QFileInfo>
#include <QTextStream>
#include <QtGui/QImage>

#include <glbinding/gl/gl.h>

#include <globjects/Texture.h>

#include <gloperate-qt/viewer/Converter.h>

#include "tools/Tools.h"


namespace gtx
{


QStringList ColorMap::listColorMaps(const std::string & relDataPath)
{
    return Tools::listFiles(std::string::fromStdString(relDataPath) + "data/colors/", "*.color");
}

ColorMap::ColorMap(const std::string & colors)
: m_colors(4)
, m_discrete(false)
, m_reversed(false)
{
    // Initialize colors
    for (int i=0; i<m_colors.size(); i++) {
        m_colors[i] = QColor(0, 0, 0, 255);
    }

    // Set colors
    if (colors.length() > 0) {
        fromString(colors);
    }
}

ColorMap::ColorMap(const ColorMap & source)
{
    // Copy options
    m_discrete = source.m_discrete;
    m_reversed = source.m_reversed;

    // Copy image
    m_image = source.m_image;

    // Copy colors
    for (int i=0; i<source.numColors(); i++) {
        m_colors.append(source.color(i));
    }
}

ColorMap::~ColorMap()
{
    clear();
}

const ColorMap &ColorMap::operator =(const ColorMap & source)
{
    // Clear old color
    clear();

    // Copy options
    m_discrete = source.m_discrete;
    m_reversed = source.m_reversed;

    // Copy image
    m_image = source.m_image;

    // Copy colors
    for (int i=0; i<source.numColors(); i++) {
        m_colors.append(source.color(i));
    }

    // Return reference
    return *this;
}

void ColorMap::clear()
{
    // Remove image filename
    m_image = "";

    // Remove colors
    m_colors.clear();

    // Reset options
    m_discrete = false;
    m_reversed = false;
}

void ColorMap::load(const std::string & filename)
{
    // Check if file exists
    QFileInfo fi(filename);
    if (fi.exists()) {
        // Color description file
        if (fi.suffix() == "color") {
            // Read color description from file
            QFile f(filename);
            if (f.open(QIODevice::ReadOnly | QIODevice::Text)) {
                 QTextStream in(&f);
                 std::string line = in.readLine();
                 if (!line.isNull())
                    fromConfigString(line);
            }
        }

        // Image file
        else {
            // Set color description using the image
            fromConfigString(filename);
        }
    }
}

void ColorMap::save(const std::string & filename) const
{
    // Open file
    QFile f(filename);
    if (f.open(QIODevice::WriteOnly | QIODevice::Text)) {
         // Write color description to file
         QTextStream out(&f);
         out << toConfigString();
    }
}

std::string ColorMap::toString() const
{
    return toConfigString();
}

void ColorMap::fromString(const std::string & colorMap)
{
    if (colorMap.startsWith("#")) fromHTMLString  (colorMap);
    else                          fromConfigString(colorMap);
}

std::string ColorMap::toConfigString() const
{
    std::string str = "";
    if (m_image != "") {
        str = m_image + ";";
    } else {
        for (int i=0; i<m_colors.size(); i++) {
            if (i > 0) str += " ";
            str += std::string::number(m_colors[i].red())   + "," +
                   std::string::number(m_colors[i].green()) + "," +
                   std::string::number(m_colors[i].blue())  + "," +
                   std::string::number(m_colors[i].alpha()) + ";";
        }
    }
    if (m_discrete) str += " discrete;";
    if (m_reversed) str += " reversed;";
    return str;
}

void ColorMap::fromConfigString(const std::string & colorMap)
{
    // Clear old color
    clear();

    // Get colors
    QStringList colors = colorMap.split(";");
    for (int i=0; i<colors.size(); i++) {
        std::string str = colors[i].trimmed();
             if (str == "discrete")       m_discrete = true;
        else if (str == "reversed")       m_reversed = true;
        else if (QFileInfo(str).exists()) m_image = str;
        else if (str.length() > 0) {
            QColor color(255, 255, 255, 255);
            QStringList components = str.split(",");
            for (int j=0; j<components.size(); j++) {
                std::string value = components[j].trimmed();
                bool ok = false;
                int v = value.toInt(&ok);
                if (ok) {
                    if (j == 0) color.setRed  (v);
                    if (j == 1) color.setGreen(v);
                    if (j == 2) color.setBlue (v);
                    if (j == 3) color.setAlpha(v);
                }
            }
            m_colors.append(color);
        }
    }
}

std::string ColorMap::toHTMLString() const
{
    // Put all colors into one string
    std::string color = "";
    for (int i=0; i<m_colors.size(); i++) {
        if (i > 0) color += " ";
        color += color2HTML(m_reversed ? m_colors[m_colors.size()-1-i] : m_colors[i]);
    }
    return color;
}

void ColorMap::fromHTMLString(const std::string & colorMap)
{
    // Clear old color
    clear();

    // Get colors
    QStringList colors = colorMap.split(" ");
    for (int i=0; i<colors.size(); i++) {
        m_colors.append(HTML2color(colors[i]));
    }
}

bool ColorMap::discrete() const
{
    return m_discrete;
}

void ColorMap::setDiscrete(bool discrete)
{
    m_discrete = discrete;
}

bool ColorMap::reversed() const
{
    return m_reversed;
}

void ColorMap::setReversed(bool reversed)
{
    m_reversed = reversed;
}

int ColorMap::numColors() const
{
    return m_colors.size();
}

QColor ColorMap::color(int index) const
{
    // Return color
    if (index >= 0 && index < m_colors.size())
        return m_colors[index];
    else
        return QColor(255, 255, 255, 255);
}

void ColorMap::setColor(int index, const QColor & color)
{
    // Check if index is valid
    if (index >= 0) {
        // Resize container, if necessary
        if (index >= m_colors.size())
            m_colors.resize(index+1);

        // Set color
        m_colors[index] = color;
    }
}

QPixmap ColorMap::pixmap() const
{
    QImage image;

    // Load image from disk or create it from a list of colors
    if (m_image != "" && QFileInfo(m_image).exists()) {
        // Read file from disk
        image = QImage(m_image);
    } else {
        // Get number of colors
        int size = m_colors.size();
        if (size <= 0) size = 1;

        // Create image
        image = QImage(size, 1, QImage::Format_RGB32);
        for (int i=0; i<size; i++) {
            image.setPixel(i, 0, color(i).rgba());
        }
    }

    // Reverse color map?
    if (m_reversed) {
        image = image.mirrored(true, false);
    }

    // Convert to pixmap
    QPixmap pixmap;
    pixmap.convertFromImage(image);
    return pixmap;
}

globjects::Texture * ColorMap::texture() const
{
    // Load image from disk or create it from a list of colors
    QImage image;
    if (m_image != "" && QFileInfo(m_image).exists()) {
        // Read file from disk
        image.load(m_image);
    } else {
        // Get number of colors
        int size = m_colors.size();
        if (size <= 0) size = 1;

        // Create texture data
        image = QImage(size, 1, QImage::Format_ARGB32);
        for (int i=0; i<size; i++) {
            image.setPixel(i, 0, this->color(i).rgb());
        }
    }

    // Reverse color map?
    if (m_reversed) {
        image = image.mirrored(true, false);
    }

    // Convert image into RGBA format
    QImage converted = gloperate_qt::Converter::convert(image);

    // Create texture
    globjects::Texture * texture = globjects::Texture::createDefault(gl::GL_TEXTURE_1D);
    texture->image1D(
        0,
        gl::GL_RGBA8,
        converted.width(),
        0,
        gl::GL_RGBA,
        gl::GL_UNSIGNED_BYTE,
        converted.constBits()
    );
    return texture;
}

std::string ColorMap::color2HTML(const QColor & color) const
{
    int r = color.red();
    int g = color.green();
    int b = color.blue();
    int a = color.alpha();
    std::string sr; sr.setNum(r, 16); if (sr.size() < 2) sr = "0" + sr;
    std::string sg; sg.setNum(g, 16); if (sg.size() < 2) sg = "0" + sg;
    std::string sb; sb.setNum(b, 16); if (sb.size() < 2) sb = "0" + sb;
    std::string sa; sa.setNum(a, 16); if (sa.size() < 2) sa = "0" + sa;
    std::string str = "#" + sr + sg + sb + sa;
    return str;
}

QColor ColorMap::HTML2color(const std::string & code) const
{
    QColor color(255, 255, 255, 255);
    if (code[0] == '#') {
        std::string number = code.mid(1);
        if (number.length() == 8 || number.length() == 6) {
            std::string sr = number.mid(0, 2);
            std::string sg = number.mid(2, 2);
            std::string sb = number.mid(4, 2);
            std::string sa = (number.length() == 8 ? number.mid(6, 2) : "ff");
            bool ok;
            int r = sr.toInt(&ok, 16); if (ok) color.setRed(r);
            int g = sg.toInt(&ok, 16); if (ok) color.setGreen(g);
            int b = sb.toInt(&ok, 16); if (ok) color.setBlue(b);
            int a = sa.toInt(&ok, 16); if (ok) color.setAlpha(a);
        } else if (number.length() == 4 || number.length() == 3) {
            std::string sr = number.mid(0, 1);
            std::string sg = number.mid(1, 1);
            std::string sb = number.mid(2, 1);
            std::string sa = (number.length() == 4 ? number.mid(3, 1) : "f");
            sr = sr + sr;
            sg = sg + sg;
            sb = sb + sb;
            sa = sa + sa;
            bool ok;
            int r = sr.toInt(&ok, 16); if (ok) color.setRed(r);
            int g = sg.toInt(&ok, 16); if (ok) color.setGreen(g);
            int b = sb.toInt(&ok, 16); if (ok) color.setBlue(b);
            int a = sa.toInt(&ok, 16); if (ok) color.setAlpha(a);
        }
    }
    return color;
}


} // namespace gtx
