
#include <gtxlib-core/base/TimeSpan.h>


namespace gtx
{


TimeSpan::TimeSpan()
: m_empty(true)
, m_dailyTime(false)
{
}

TimeSpan::TimeSpan(const std::chrono::time_point & minTime, const std::chrono::time_point & maxTime, const std::chrono::duration & minDailyTime, const std::chrono::duration & maxDailyTime, bool dailyTime)
: m_empty(false)
, m_dailyTime(dailyTime)
, m_minTime(minTime)
, m_maxTime(maxTime)
, m_minDailyTime(minDailyTime)
, m_maxDailyTime(maxDailyTime)
{
}

TimeSpan::TimeSpan(const TimeSpan & timespan)
: m_empty       (timespan.m_empty)
, m_dailyTime   (timespan.m_dailyTime)
, m_minTime     (timespan.m_minTime)
, m_maxTime     (timespan.m_maxTime)
, m_minDailyTime(timespan.m_minDailyTime)
, m_maxDailyTime(timespan.m_maxDailyTime)
{
}

TimeSpan::~TimeSpan()
{
}

bool TimeSpan::operator ==(const TimeSpan & timespan) const
{
    // Compare positions
    return (m_empty == timespan.m_empty && m_dailyTime == timespan.m_dailyTime && m_minTime == timespan.m_minTime && m_maxTime == timespan.m_maxTime && m_minDailyTime == timespan.m_minDailyTime && m_maxDailyTime == timespan.m_maxDailyTime);
}

TimeSpan & TimeSpan::operator =(const TimeSpan & timespan)
{
    // Copy timespan
    m_empty        = timespan.m_empty;
    m_dailyTime    = timespan.m_dailyTime;
    m_minTime      = timespan.m_minTime;
    m_maxTime      = timespan.m_maxTime;
    m_minDailyTime = timespan.m_minDailyTime;
    m_maxDailyTime = timespan.m_maxDailyTime;
    return *this;
}

bool TimeSpan::empty() const
{
    return m_empty;
}

void TimeSpan::clear()
{
    std::chrono::time_point emptyTime;
    std::chrono::duration   emptyDailyTime;

    // Reset values
    m_empty         = true;
    m_dailyTime     = false;
    m_minTime       = emptyTime;
    m_maxTime       = emptyTime;
    m_minDailyTime  = emptyDailyTime;
    m_maxDailyTime  = emptyDailyTime;
}

std::chrono::time_point TimeSpan::minTime() const
{
    return m_minTime;
}

void TimeSpan::setMinTime(const std::chrono::time_point & time)
{
    m_empty   = false;
    m_minTime = time;
}

std::chrono::time_point TimeSpan::maxTime() const
{
    return m_maxTime;
}

void TimeSpan::setMaxTime(const std::chrono::time_point & time)
{
    m_empty   = false;
    m_maxTime = time;
}

void TimeSpan::expand(const std::chrono::time_point & time)
{
    // Is time span empty?
    if (m_empty) {
        // Set first time
        m_minTime = m_maxTime = time;
        m_empty   = false;
    } else {
        // Extend time span
        if (time < m_minTime) m_minTime = time;
        if (time > m_maxTime) m_maxTime = time;
    }
}

std::chrono::duration TimeSpan::minDailyTime() const
{
    return m_minDailyTime;
}

void TimeSpan::setMinDailyTime(const std::chrono::duration & time)
{
    m_minDailyTime = time;
    m_dailyTime    = true;
}

std::chrono::duration TimeSpan::maxDailyTime() const
{
    return m_maxDailyTime;
}

void TimeSpan::setMaxDailyTime(const std::chrono::duration & time)
{
    m_maxDailyTime = time;
    m_dailyTime    = true;
}

bool TimeSpan::dailyTime() const
{
    return m_dailyTime;
}

bool TimeSpan::isInside(const std::chrono::time_point & time) const
{
    // Check global time span
    if (!m_empty && time >= m_minTime && time <= m_maxTime) {
        // Check daily time
        if (!m_dailyTime || (time.time_of_day() >= m_minDailyTime && time.time_of_day() <= m_maxDailyTime))
            return true;
    }

    // Time not in the time space
    return false;
}

long TimeSpan::secondsSinceStart(const std::chrono::time_point & time) const
{
    if (m_empty)
        return -1;
    else
        return (long)(time - m_minTime).total_seconds();
}


} // namespace gtx
