
#include <gtxlib-core/data/AircraftDB.h>

#include <fstream>
#include <vector>

#include <cppassist/string/manipulation.h>
#include <cppassist/string/conversion.h>

#include <gtxlib-core/gtxlib-core.h>


using namespace cppassist;


namespace
{


std::istream & getline(std::istream & is, std::string & str)
{
    std::istream & res = std::getline(is, str);

    if (!str.empty() && *str.rbegin() == '\r')
    {
        str.erase(str.length() - 1, 1);
    }

    return res;
}


}


namespace gtx
{


AircraftDB::AircraftDB()
{
    loadAircraftDB();
    loadICAOAircraftDB();
}

AircraftDB::~AircraftDB()
{
}

const AircraftDB::Aircraft & AircraftDB::aircraft(const std::string & planeType) const
{
    static AircraftDB::Aircraft nullAircraft;

    auto it = m_aircrafts.find(planeType);
    if (it != m_aircrafts.end()) {
        return m_aircrafts.at(planeType);
    }

    return nullAircraft;
}

const AircraftDB::ICAOAircraft & AircraftDB::icaoAircraft(const std::string & planeType) const
{
    static AircraftDB::ICAOAircraft nullAircraft;

    auto it = m_icao.find(planeType);
    if (it != m_icao.end()) {
        return m_icao.at(planeType);
    }

    return nullAircraft;
}

void AircraftDB::loadAircraftDB()
{
    // Load aircraft definitions
    std::ifstream f(gtx::dataPath() + "/datasets/aircraft.csv");
    std::string line;

    // Skip header
    getline(f, line);

    // Read file
    while (getline(f, line)) {
        // Read line
        std::vector<std::string> values = string::split(line, ';');
        if (values.size() == 17) {
            // Save aircraft description
            Aircraft aircraft;
            aircraft.id                         = values[0];
            aircraft.description                = values[1];
            aircraft.source                     = values[2];
            aircraft.engineType                 = values[3];
            aircraft.numEngines                 = string::fromString<int>(values[4]);
            aircraft.weightClass                = values[5];
            aircraft.ownerCategory              = values[6];
            aircraft.maxTakeoffWeight           = string::fromString<int>(values[7]);
            aircraft.maxLandingWeight           = string::fromString<int>(values[8]);
            aircraft.maxLandingDistance         = string::fromString<int>(values[9]);
            aircraft.maxSeaLevelStaticThrust    = string::fromString<int>(values[10]);
            aircraft.noiseChapter               = string::fromString<int>(values[11]);
            aircraft.npdID                      = values[12];
            aircraft.powerParameter             = values[13];
            aircraft.approachSpectralClassID    = string::fromString<int>(values[14]);
            aircraft.departureSpectralClassID   = string::fromString<int>(values[15]);
            aircraft.lateralDirectivityID       = values[16];
            m_aircrafts[aircraft.id] = aircraft;
        }
    }
}

void AircraftDB::loadICAOAircraftDB()
{
    // Load aircraft definitions
    std::ifstream f(gtx::dataPath() + "/datasets/aircraft2.csv");
    std::string line;

    // Skip header
    getline(f, line);

    // Read file
    while (getline(f, line)) {
        // Read line
        std::vector<std::string> values = string::split(line, ',');
        if (values.size() == 21) {
            // Save aircraft description
            ICAOAircraft aircraft;
            aircraft.id                 = values[0];
            aircraft.manufacturer       = values[1];
            aircraft.model              = values[2];
            aircraft.description        = values[3];
            aircraft.type               = values[4];
            aircraft.engine             = values[5];
            aircraft.numEngines         = string::fromString<int>(values[6]);
            aircraft.weightClass        = values[7];
            aircraft.classCode          = values[8];
            aircraft.numPassengers      = string::fromString<int>(values[9]);
            aircraft.numSeats           = string::fromString<int>(values[10]);
            aircraft.seatsSpecificity   = values[11];
            aircraft.m3                 = string::fromString<float>(values[12]);
            aircraft.m3Units            = values[13];
            aircraft.m2                 = string::fromString<float>(values[14]);
            aircraft.m2Units            = values[15];
            aircraft.m1                 = string::fromString<float>(values[16]);
            aircraft.m1Units            = values[17];
            aircraft.b                  = string::fromString<float>(values[18]);
            aircraft.bUnits             = values[19];
            aircraft.fuelUseSpecificity = values[20];
            m_icao[aircraft.id] = aircraft;
        }
    }
}


} // namespace gtx
