
#include <gtxlib-core/data/DataSet.h>

#include <glm/glm.hpp>

#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>
#include <gtxlib-core/data/Line.h>


namespace gtx
{


DataSet::DataSet()
: m_minTime(0.0f)
, m_maxTime(0.0f)
{
}

DataSet::~DataSet()
{
    // Clear data
    clear();
}

void DataSet::clear()
{
    // Reset other data
    m_area.clear();
}

bool DataSet::empty() const
{
    // Check if there are any elements
    return m_nodes.empty() && m_edges.empty() && m_lines.empty();
}

void DataSet::printStatistics()
{
    printf("Filename: %s\n", m_filename.c_str());
    printf("Nodes:    %u\n", (unsigned int) m_nodes.size());
    printf("Edges:    %u\n", (unsigned int) m_edges.size());
    printf("Lines:    %u\n", (unsigned int) m_lines.size());
    printf("Area:     [%.4f, %.4f, %.4f, %.4f]\n", m_area.minPos().getLat(), m_area.minPos().getLong(), m_area.maxPos().getLat(), m_area.maxPos().getLong());
    printf("\n");
}

std::string DataSet::filename() const
{
    return m_filename;
}

void DataSet::setFilename(const std::string & filename)
{
    m_filename = filename;
}

const Area & DataSet::area() const
{
    return m_area;
}

const ItemGroup<Item> & DataSet::statics() const
{
    return m_statics;
}

const ItemGroup<Node> & DataSet::nodes() const
{
    return m_nodes;
}

ItemGroup<Node> & DataSet::nodes()
{
    return m_nodes;
}

const ItemGroup<Edge> & DataSet::edges() const
{
    return m_edges;
}

ItemGroup<Edge> & DataSet::edges()
{
    return m_edges;
}

const ItemGroup<Line> & DataSet::lines() const
{
    return m_lines;
}

ItemGroup<Line> & DataSet::lines()
{
    return m_lines;
}

float DataSet::minimumTime() const
{
    return m_minTime;
}

float DataSet::maximumTime() const
{
    return m_maxTime;
}

void DataSet::fixIds()
{
    // Fix edge ids
    for (auto pair : m_edges.map())
    {
        auto id = pair.second->id();
        pair.second->node1()->setEdgeId(id);
        pair.second->node2()->setEdgeId(id);
    }

    // Fix line ids
    for (auto pair : m_lines.map())
    {
        auto id = pair.second->id();

        for (auto * node : pair.second->nodes())
        {
            node->setEdgeId(id);
        }
    }
}

void DataSet::calculateEdgeLength()
{
    // Get edges
    for (auto it = m_edges.map().begin(); it != m_edges.map().end(); ++it)
    {
        // Get edge and nodes
        auto * edge = it->second;
        auto * node1 = edge->node1();
        auto * node2 = edge->node2();

        // Get positions
        float lat1 = glm::radians(node1->position().getLat());
        float lat2 = glm::radians(node2->position().getLat());
        float lon1 = glm::radians(node1->position().getLong());
        float lon2 = glm::radians(node2->position().getLong());

        // Calculate geographical distance
        float polar  = sin(lat1)*sin(lat2) + cos(lat1)*cos(lat2) * cos(lon2-lon1);
        float length = glm::degrees(acos(polar) * 10.0f);

        // Set distance as edge attribute
        edge->attribute("length")->setValue(length);
    }
}

void DataSet::addStandardAttributes()
{
    // Create static attributes "Zero" and "One"
    m_statics.clear();
    auto * item = m_statics.create();
    item->attribute("Zero")->setValue(0.0f);
    item->attribute("One") ->setValue(1.0f);

    // Put node positions in attributes "Lat" and "Lon"
    // Get nodes
    for (auto it = m_nodes.map().begin(); it != m_nodes.map().end(); ++it)
    {
        // Get node
        auto * node = it->second;

        // Set attributes
        node->attribute("Lat")->setValue(glm::radians(node->position().getLat()));
        node->attribute("Lon")->setValue(glm::radians(node->position().getLong()));
    }

    // Set statistice for attribute 'Zero'
    auto * attributeInfo = m_statics.attribute("Zero");
    attributeInfo->setAutoCalculateStatistics(false);
    attributeInfo->setMinimumValue(0.0f);
    attributeInfo->setMaximumValue(1.0f);
    attributeInfo->setMeanValue(0.5f);
    attributeInfo->setStandardDeviation(0.0f);

    // Set statistice for attribute 'One'
    attributeInfo = m_statics.attribute("One");
    attributeInfo->setAutoCalculateStatistics(false);
    attributeInfo->setMinimumValue(0.0f);
    attributeInfo->setMaximumValue(1.0f);
    attributeInfo->setMeanValue(0.5f);
    attributeInfo->setStandardDeviation(0.0f);

    // Set statistice for attribute 'Lat'
    attributeInfo = m_nodes.attribute("Lat");
    attributeInfo->setAutoCalculateStatistics(false);
    attributeInfo->setMinimumValue(-90.0f);
    attributeInfo->setMaximumValue( 90.0f);
    attributeInfo->setMeanValue(0.0f);
    attributeInfo->setStandardDeviation(90.0f);

    // Set statistice for attribute 'Lon'
    attributeInfo = m_nodes.attribute("Lon");
    attributeInfo->setAutoCalculateStatistics(false);
    attributeInfo->setMinimumValue(-180.0f);
    attributeInfo->setMaximumValue( 180.0f);
    attributeInfo->setMeanValue(0.0f);
    attributeInfo->setStandardDeviation(180.0f);
}

void DataSet::analyzeTrajectories()
{
    float minLat = 0.0f;
    float maxLat = 0.0f;
    float minLon = 0.0f;
    float maxLon = 0.0f;

    // Process all lines
    bool first = true;
    for (auto it = m_lines.map().begin(); it != m_lines.map().end(); ++it)
    {
        // Get line
        auto * line = it->second;

        // Get all nodes on the line
        float length = 0.0f;
        Node * last = nullptr;
        for (auto node : line->nodes())
        {
            // Add line id
            node->attribute("lineId")->setValue(line->id());

            // Add attribute for the length of the line
            node->attribute("length")->setValue(length);
            length += 0.1f;

            // Calculate acceleration
            float acc = 0.0f;
            if (last) {
                auto speed1 = last->attribute("velocity")->value();
                auto time1  = last->timeInt();
                auto speed2 = node->attribute("velocity")->value();
                auto time2  = node->timeInt();
                acc = fabs(speed2 - speed1) / (time2 - time1);
            }
            node->attribute("acceleration")->setValue(acc);

            // Calculate area
            float lat = node->position().getLat();
            float lon = node->position().getLong();
            if (first) {
                minLat = maxLat = lat;
                minLon = maxLon = lon;
                first = false;
            } else {
                if (lat > maxLat) maxLat = lat;
                if (lat < minLat) minLat = lat;
                if (lon > maxLon) maxLon = lon;
                if (lon < minLon) minLon = lon;
            }

            // Next node
            last = node;
        }
    }

    // Save area
    m_area.setMinPos(Position(minLat, minLon, 0.0f));
    m_area.setMaxPos(Position(maxLat, maxLon, 0.0f));
}

void DataSet::calculateAttributeStatistics()
{
    m_minTime = 0.0f;
    m_maxTime = 0.0f;

    // Iterate over nodes, edges, and lines
    for (int i : {0, 1, 2})
    {
        // Get item group
        auto & group = (i == 2) ? static_cast<AbstractItemGroup &>(m_lines) :
                       (i == 1) ? static_cast<AbstractItemGroup &>(m_edges) :
                                  static_cast<AbstractItemGroup &>(m_nodes);

        // Collect attributes and calculate minimum, maximum, and mean values
        for (auto pair : group.items())
        {
            // Get item
            auto * item = pair.second;

            // Get attributes
            auto attributes = item->attributes();
            for (auto attr : attributes)
            {
                // Get attribute
                auto * attributeInfo = group.attribute(attr);
                if (!attributeInfo->autoCalculateStatistics())
                    continue;

                // Update attribute
                auto * attribute = item->attribute(attr);
                attributeInfo->addValue(attribute->value());
            }

            // Calculate minimum and maximum time
            if (i == 0) {
                auto * node = static_cast<Node *>(item);
                float time = (float)node->timeInt();

                if (m_minTime == m_maxTime && m_minTime == 0.0f) {
                    m_minTime = m_maxTime = time;
                } else {
                    if (time < m_minTime) m_minTime = time;
                    if (time > m_maxTime) m_maxTime = time;
                }
            }
        }

        // Calculate additional attribute metrics
        for (auto attr : group.attributes())
        {
            // Get attribute info
            auto * attributeInfo = group.attribute(attr);
            if (!attributeInfo->autoCalculateStatistics())
                continue;

            // Calculate standard deviation
            attributeInfo->setMeanValue(attributeInfo->sum() / attributeInfo->numValues());
            float meanValue = attributeInfo->meanValue();
            float stdDeviation = 0.0f;

            for (auto pair : group.items())
            {
                // Get item
                auto * item = pair.second;

                // Get value
                float value = item->attribute(attr)->value();
                float difference = value - meanValue;
                stdDeviation += difference * difference;
            }

            if (attributeInfo->numValues() > 0.0f) {
                attributeInfo->setStandardDeviation(sqrt(stdDeviation / attributeInfo->numValues()));
            }
        }
    }
}

const std::vector<std::string> & DataSet::mapLayers() const
{
    return m_mapLayers;
}

void DataSet::setMapLayers(std::vector<std::string> & mapLayers)
{
    m_mapLayers = mapLayers;
}


} // namespace gtx
