
#include <gtxlib-core/geometry/AttributeStorage.h>

#include <iostream>
#include <vector>

#include <glm/glm.hpp>

#include <cppassist/memory/make_unique.h>

#include <glbinding/gl/gl.h>

#include <globjects/Buffer.h>

#include <gtxlib-core/data/DataSet.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>
#include <gtxlib-core/data/Line.h>


using namespace globjects;


namespace gtx
{


AttributeStorage::AttributeStorage()
: m_numStaticAttributes(0)
, m_numNodeAttributes(0)
, m_numEdgeAttributes(0)
, m_numLineAttributes(0)
{
}

AttributeStorage::~AttributeStorage()
{
}

void AttributeStorage::setData(const DataSet & dataset)
{
    // Clear attributes
    m_attrs.clear();

    // Create attributes textures
    m_staticAttributesTexture = createAttributesTexture(dataset.statics(), "",        m_attrs, m_numStaticAttributes);
    m_nodeAttributesTexture   = createAttributesTexture(dataset.nodes(),   "Node - ", m_attrs, m_numNodeAttributes);
    m_edgeAttributesTexture   = createAttributesTexture(dataset.edges(),   "Edge - ", m_attrs, m_numEdgeAttributes);
    m_lineAttributesTexture   = createAttributesTexture(dataset.lines(),   "Line - ", m_attrs, m_numLineAttributes);
}

void AttributeStorage::clearData()
{
    m_staticAttributesTexture = cppassist::make_unique<globjects::Texture>(gl::GL_TEXTURE_BUFFER);
    m_nodeAttributesTexture   = cppassist::make_unique<globjects::Texture>(gl::GL_TEXTURE_BUFFER);
    m_edgeAttributesTexture   = cppassist::make_unique<globjects::Texture>(gl::GL_TEXTURE_BUFFER);
    m_lineAttributesTexture   = cppassist::make_unique<globjects::Texture>(gl::GL_TEXTURE_BUFFER);
}

const std::vector<std::string> & AttributeStorage::attributes() const
{
    return m_attrs;
}

int AttributeStorage::numStaticAttributes() const
{
    return m_numStaticAttributes;
}

int AttributeStorage::numNodeAttributes() const
{
    return m_numNodeAttributes;
}

int AttributeStorage::numEdgeAttributes() const
{
    return m_numEdgeAttributes;
}

int AttributeStorage::numLineAttributes() const
{
    return m_numLineAttributes;
}

globjects::Texture * AttributeStorage::staticAttributesTexture()
{
    return m_staticAttributesTexture.get();
}

globjects::Texture * AttributeStorage::nodeAttributesTexture()
{
    return m_nodeAttributesTexture.get();
}

globjects::Texture * AttributeStorage::edgeAttributesTexture()
{
    return m_edgeAttributesTexture.get();
}

globjects::Texture * AttributeStorage::lineAttributesTexture()
{
    return m_lineAttributesTexture.get();
}

std::unique_ptr<globjects::Texture> AttributeStorage::createAttributesTexture(const AbstractItemGroup & group, const std::string & prefix, std::vector<std::string> & attrs, unsigned int & numAttrs)
{
    static const int attrMin  = 0;  // Minimum value
    static const int attrMax  = 1;  // Maximum value
    static const int attrMean = 2;  // Mean value
    static const int attrDev  = 3;  // Standard deviation
    static const int maxAttrs = 32; // Maximum number of attributes per item

    // Get number of items
    // This assumes that the item IDs have been linearized and start with 0 (DataSet::fixIds() has been called)
    unsigned int numItems  = group.size();
    unsigned int numChunks = numItems + 4;   // 0-4: minValues, maxValues, mean values, deviation

    // Determine texture size
    unsigned int numFloats = numChunks * maxAttrs;
    numFloats = ((numFloats / 1024) + 1) * 1024;

    // Create texture data
    float * data = (float*)malloc(numFloats * sizeof(float));
    for (unsigned int i=0; i<numFloats; i++) {
        data[i] = 0.0f;
    }

    // Add attributes
    auto & attributes = group.attributes();
    for (unsigned int attrIndex = 0; attrIndex < attributes.size(); attrIndex++)
    {
        // Get attribute
        auto   attrName  = attributes[attrIndex];
        auto * attribute = group.attribute(attrName);

        // Add attribute to list
        attrs.push_back(prefix + attrName);

        // Fill in statistics
        data[attrMin  * maxAttrs + attrIndex] = attribute->minimumValue();
        data[attrMax  * maxAttrs + attrIndex] = attribute->maximumValue();
        data[attrMean * maxAttrs + attrIndex] = attribute->meanValue();
        data[attrDev  * maxAttrs + attrIndex] = attribute->standardDeviation();

        // Fill in attribute values
        for (auto it = group.items().begin(); it != group.items().end(); ++it)
        {
            // Get item
            auto * item = it->second;
            int itemIndex = item->id();

            // Get attribute value
            float value = item->attribute(attrName)->value();
            data[(itemIndex + 4) * maxAttrs + attrIndex] = value;
        }
    }

    // Check maximum texture size
    gl::GLint maxSize;
    gl::glGetIntegerv(gl::GL_MAX_TEXTURE_BUFFER_SIZE, &maxSize);
    unsigned  maxNumFloats = maxSize / sizeof(float);
    if (numFloats > maxNumFloats) {
        std::cout << "ERROR: Attribute texture too large (" << numFloats << " floats). Truncating to " << maxNumFloats << " floats." << std::endl;
        numFloats = maxNumFloats;
    }

    // Save number of attributes
    numAttrs = attributes.size();

    // Create texture
    auto texture = cppassist::make_unique<globjects::Texture>(gl::GL_TEXTURE_BUFFER);
    // [TODO] Memory leak?
    globjects::Buffer * buffer = new globjects::Buffer();
    buffer->setData(numFloats * sizeof(float), data, gl::GL_STATIC_DRAW);
    texture->texBuffer(gl::GL_R32F, buffer);

    // Release texture data
    free(data);

    // Return texture
    return texture;
}


} // namespace gtx
