
#include <gtxlib-core/geometry/EdgeGeometry.h>

#include <vector>

#include <glm/glm.hpp>

#include <cppassist/memory/make_unique.h>

#include <glbinding/gl/gl.h>

#include <globjects/Buffer.h>
#include <globjects/VertexAttributeBinding.h>

#include <gtxlib-core/data/ItemGroup.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>


using namespace globjects;


namespace gtx
{


EdgeGeometry::EdgeGeometry()
: m_numElements(0)
, m_size(0)
{
}

EdgeGeometry::~EdgeGeometry()
{
}

void EdgeGeometry::setData(const DataSet & dataset)
{
    // Create vertex array
    std::vector<glm::vec4> vertices;
    std::vector<glm::vec4> parameters;

    // Reset data
    m_numElements = 0;

    // Iterate over edges
    const ItemGroup<Edge> & edges = dataset.edges();
    for (auto it = edges.map().begin(); it != edges.map().end(); ++it)
    {
        auto * edge = it->second;

        // Add nodes to vertex array
        for (int i=0; i<2; i++) {
            // Get node
            auto * node = (i==0) ? edge->node1() : edge->node2();

            // [DEBUG]
            // std::cout << "(" << node->position().getLat() << ", " << node->position().getLong() << ", " << node->position().getHeight() << ")\n";

            // Position (lat / lon / height) in radians
            vertices.push_back(
                glm::vec4(
                    glm::radians(node->position().getLat()),
                    glm::radians(node->position().getLong()),
                    node->position().getHeight(),
                    0.0f
                )
            );

            // Node-ID / Edge-ID / Time / -
            parameters.push_back(
                glm::vec4(
                    (float)node->id(),
                    (float)edge->id(),
                    node->hasAttribute("time") ? node->attribute("time")->value() : 0.0f,
                    0.0f
                )
            );

            m_numElements++;
        }
    }

    // Calculate size
    m_size = vertices.size() * sizeof(glm::vec4) + parameters.size() * sizeof(glm::vec4);

    // Create VAO
    m_vao = cppassist::make_unique<VertexArray>();

    // Create binding for vertices
    {
        // Create buffer
        m_bufferVertices = cppassist::make_unique<Buffer>();
        m_bufferVertices->setData(vertices, gl::GL_STATIC_DRAW);

        // Create binding
        auto binding = m_vao->binding(0);
        binding->setAttribute(0);
        binding->setBuffer(m_bufferVertices.get(), 0, sizeof(glm::vec4));
        binding->setFormat(4, gl::GL_FLOAT, gl::GL_FALSE, 0);
        m_vao->enable(0);
    }

    // Create binding for parameters
    {
        // Create buffer
        m_bufferParameters = cppassist::make_unique<Buffer>();
        m_bufferParameters->setData(parameters, gl::GL_STATIC_DRAW);

        // Create binding
        auto binding = m_vao->binding(1);
        binding->setAttribute(1);
        binding->setBuffer(m_bufferParameters.get(), 0, sizeof(glm::vec4));
        binding->setFormat(4, gl::GL_FLOAT, gl::GL_FALSE, 0);
        m_vao->enable(1);
    }

    // Unbind VAO
    m_vao->unbind();
}

unsigned int EdgeGeometry::numElements() const
{
    return m_numElements;
}

unsigned int EdgeGeometry::size() const
{
    return m_size;
}

globjects::VertexArray * EdgeGeometry::vao()
{
    return m_vao.get();
}

const globjects::VertexArray * EdgeGeometry::vao() const
{
    return m_vao.get();
}

void EdgeGeometry::draw() const
{
    if (m_vao)
    {
        // Set line width to maximum, shader will discard texels to make smaller lines
        gl::GLfloat range[2];
        gl::glGetFloatv(gl::GL_ALIASED_LINE_WIDTH_RANGE, range);
        gl::glLineWidth(std::min(range[1], 10.0f));

        // Draw lines
        m_vao->drawArrays(gl::GL_LINES, 0, m_numElements);
        m_vao->unbind();

        // Reset line width
        gl::glLineWidth(1.0f);
    }
}


} // namespace gtx
