
#include <gtxlib-core/geometry/LineGeometry.h>

#include <vector>

#include <glm/glm.hpp>

#include <cppassist/memory/make_unique.h>

#include <glbinding/gl/gl.h>

#include <globjects/Buffer.h>
#include <globjects/VertexAttributeBinding.h>

#include <gtxlib-core/data/ItemGroup.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Line.h>


using namespace globjects;


namespace gtx
{


LineGeometry::LineGeometry()
: m_size(0)
{
}

LineGeometry::~LineGeometry()
{
}

void LineGeometry::setData(const DataSet & dataset)
{
    const auto & lines = dataset.lines();

    // Create vertex array
    std::vector<glm::vec4> vertices;
    std::vector<glm::vec4> parameters;

    // Reset data
    m_size = 0;

    // Iterate over lines
    for (auto it = lines.map().begin(); it != lines.map().end(); ++it)
    {
        auto * line = it->second;

        // Add nodes to vertex array
        bool first = true;
        for (auto it2 = line->nodes().begin(); it2 != line->nodes().end(); ++it2)
        {
            Node * node = *it2;

            // [DEBUG]
            // std::cout << "(" << node->position().getLat() << ", " << node->position().getLong() << ", " << node->position().getHeight() << ")\n";

            // Duplicate first and last node
            int count = (first || it2 == line->nodes().end()) ? 2 : 1;
            for (int i=0; i<count; i++) {
                // Position (lat / lon / height) in radians
                vertices.push_back(
                    glm::vec4(
                        glm::radians(node->position().getLat()),
                        glm::radians(node->position().getLong()),
                        node->position().getHeight(),
                        0.0f
                    )
                );

                // Node-ID / Line-ID / Time / -
                parameters.push_back(
                    glm::vec4(
                        (float)node->id(),
                        (float)line->id(),
                        node->hasAttribute("time") ? node->attribute("time")->value() : -1.0f,
                        0.0f
                    )
                );

                m_size++;
            }

            first = false;
        }
    }

    // Create VAO
    m_vao = cppassist::make_unique<VertexArray>();

    // Create binding for vertices
    {
        // Create buffer
        m_bufferVertices = cppassist::make_unique<Buffer>();
        m_bufferVertices->setData(vertices, gl::GL_STATIC_DRAW);

        // Create binding
        auto binding = m_vao->binding(0);
        binding->setAttribute(0);
        binding->setBuffer(m_bufferVertices.get(), 0, sizeof(glm::vec4));
        binding->setFormat(4, gl::GL_FLOAT, gl::GL_FALSE, 0);
        m_vao->enable(0);
    }

    // Create binding for parameters
    {
        // Create buffer
        m_bufferParameters = cppassist::make_unique<Buffer>();
        m_bufferParameters->setData(parameters, gl::GL_STATIC_DRAW);

        // Create binding
        auto binding = m_vao->binding(1);
        binding->setAttribute(1);
        binding->setBuffer(m_bufferParameters.get(), 0, sizeof(glm::vec4));
        binding->setFormat(4, gl::GL_FLOAT, gl::GL_FALSE, 0);
        m_vao->enable(1);
    }

    // Unbind VAO
    m_vao->unbind();
}

unsigned int LineGeometry::size() const
{
    return m_size;
}

globjects::VertexArray * LineGeometry::vao()
{
    return m_vao.get();
}

const globjects::VertexArray * LineGeometry::vao() const
{
    return m_vao.get();
}

void LineGeometry::draw() const
{
    if (m_vao)
    {
        // Set line width to maximum, shader will discard texels to make smaller lines
        gl::GLfloat range[2];
        gl::glGetFloatv(gl::GL_ALIASED_LINE_WIDTH_RANGE, range);
        gl::glLineWidth(std::min(range[1], 10.0f));

        // Set blend mode (needed for density maps)
        gl::glBlendFunc(gl::GL_SRC_ALPHA, gl::GL_ONE);

        // Draw lines
        m_vao->drawArrays(gl::GL_LINE_STRIP_ADJACENCY, 0, m_size);
        m_vao->unbind();

        // Reset blend mode
        gl::glBlendFunc(gl::GL_SRC_ALPHA, gl::GL_ONE_MINUS_SRC_ALPHA);

        // Reset line width
        gl::glLineWidth(1.0f);
    }
}


} // namespace gtx
