
#include <gtxlib-core/geometry/NodeGeometry.h>

#include <vector>

#include <glm/glm.hpp>

#include <cppassist/memory/make_unique.h>

#include <glbinding/gl/gl.h>

#include <globjects/Buffer.h>
#include <globjects/VertexAttributeBinding.h>

#include <gtxlib-core/data/ItemGroup.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Line.h>


using namespace globjects;


namespace gtx
{


NodeGeometry::NodeGeometry()
: m_numElements(0)
, m_size(0)
{
}

NodeGeometry::~NodeGeometry()
{
}

void NodeGeometry::setData(const DataSet & dataset, unsigned int maxNodes)
{
    const ItemGroup<Node> & nodes = dataset.nodes();

    // Create vertex array
    std::vector<glm::vec4> vertices;
    std::vector<glm::vec4> parameters;

    // Add nodes to vertex array
    m_numElements = 0;
    for (auto it = nodes.map().begin(); it != nodes.map().end(); ++it)
    {
        auto * node = it->second;

        // Count nodes
        if (maxNodes > 0 && m_numElements >= maxNodes) {
            break;
        }
        m_numElements++;

        // Position (lat / lon / height) in radians
        vertices.push_back(
            glm::vec4(
                glm::radians(node->position().getLat()),
                glm::radians(node->position().getLong()),
                node->position().getHeight(),
                0.0f
            )
        );

        // Node-ID / Edge-ID / Time / -
        parameters.push_back(
            glm::vec4(
                (float)node->id(),
                (float)node->edgeId(),
                (float)node->timeInt(),
                0.0f
            )
        );
    }

    // Calculate size
    m_size = vertices.size() * sizeof(glm::vec4) + parameters.size() * sizeof(glm::vec4);

    // Create VAO
    m_vao = cppassist::make_unique<VertexArray>();

    // Create binding for vertices
    {
        // Create buffer
        m_bufferVertices = cppassist::make_unique<Buffer>();
        m_bufferVertices->setData(vertices, gl::GL_STATIC_DRAW);

        // Create binding
        auto binding = m_vao->binding(0);
        binding->setAttribute(0);
        binding->setBuffer(m_bufferVertices.get(), 0, sizeof(glm::vec4));
        binding->setFormat(4, gl::GL_FLOAT, gl::GL_FALSE, 0);
        m_vao->enable(0);
    }

    // Create binding for parameters
    {
        // Create buffer
        m_bufferParameters = cppassist::make_unique<Buffer>();
        m_bufferParameters->setData(parameters, gl::GL_STATIC_DRAW);

        // Create binding
        auto binding = m_vao->binding(1);
        binding->setAttribute(1);
        binding->setBuffer(m_bufferParameters.get(), 0, sizeof(glm::vec4));
        binding->setFormat(4, gl::GL_FLOAT, gl::GL_FALSE, 0);
        m_vao->enable(1);
    }

    // Unbind VAO
    m_vao->unbind();
}

unsigned int NodeGeometry::numElements() const
{
    return m_numElements;
}

unsigned int NodeGeometry::size() const
{
    return m_size;
}

globjects::VertexArray * NodeGeometry::vao()
{
    return m_vao.get();
}

const globjects::VertexArray * NodeGeometry::vao() const
{
    return m_vao.get();
}

void NodeGeometry::drawPatches() const
{
    if (m_vao)
    {
        gl::glPatchParameteri(gl::GL_PATCH_VERTICES, 1);
        m_vao->drawArrays(gl::GL_PATCHES, 0, m_numElements);
        m_vao->unbind();
    }
}

void NodeGeometry::draw() const
{
    if (m_vao)
    {
        m_vao->drawArrays(gl::GL_POINTS, 0, m_numElements);
        m_vao->unbind();
    }
}


} // namespace gtx
