
#include <gtxlib-core/geometry/PlotGeometry.h>

#include <vector>

#include <glm/glm.hpp>

#include <cppassist/memory/make_unique.h>

#include <glbinding/gl/gl.h>

#include <globjects/Buffer.h>
#include <globjects/VertexAttributeBinding.h>

#include <gtxlib-core/data/ItemGroup.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Line.h>


using namespace globjects;


namespace gtx
{


PlotGeometry::PlotGeometry()
: m_numElements(0)
, m_size(0)
{
}

PlotGeometry::~PlotGeometry()
{
}

void PlotGeometry::setData(const DataSet & dataset, int nodeId)
{
    const ItemGroup<Node> & nodes = dataset.nodes();

    // Check dataset properties
    {
        auto it = nodes.map().begin();
        auto * node = it->second;
        if (!node) return;

        auto attrs = node->attributes();
        if (!node->hasAttribute("id") || !node->hasAttribute("time") || !node->hasAttribute("betweenness_centrality")) return;
    }

    // Get selected ID
    int selectedId = -1;
    {
        if (nodeId < 0) return;

        auto * node = nodes.get(nodeId);
        if (!node) return;

        selectedId = (int)node->attribute("id")->value();

        if (selectedId < 0) return;
    }

    // Create vertex array
    std::vector<glm::vec4> vertices;

    // Add nodes to vertex array
    m_numElements = 0;
    for (auto it = nodes.map().begin(); it != nodes.map().end(); ++it)
    {
        auto * node = it->second;

        // Get Id
        int id = (int)node->attribute("id")->value();
        if (id != selectedId) {
            continue;
        }

        // Get attribute for X-axis
        float time = node->attribute("time")->value();
        float x = (time - nodes.attribute("time")->minimumValue()) / (nodes.attribute("time")->maximumValue() - nodes.attribute("time")->minimumValue());

        // Get attribute for Y-axis
        float temp = node->attribute("betweenness_centrality")->value();
        float y = (temp - nodes.attribute("betweenness_centrality")->minimumValue()) / (nodes.attribute("betweenness_centrality")->maximumValue() - nodes.attribute("betweenness_centrality")->minimumValue());

        // Add vertex
        vertices.push_back(
            glm::vec4(x, y, id, 0.0f)
        );

        // Count nodes
        m_numElements++;
    }

    // Sort vertices first by id, then by x-coordinate
    std::sort(vertices.begin(), vertices.end(), [] (const glm::vec4 & v1, const glm::vec4 & v2) -> bool {
        if (v1.z < v2.z) {
            return true;
        } else if (v1.z > v2.z) {
            return false;
        } else {
            return (v1.x < v2.x);
        }
    });

    // Calculate size
    m_size = vertices.size() * sizeof(glm::vec4);

    // Create VAO
    m_vao = cppassist::make_unique<VertexArray>();

    // Create binding for vertices
    {
        // Create buffer
        m_bufferVertices = cppassist::make_unique<Buffer>();
        m_bufferVertices->setData(vertices, gl::GL_STATIC_DRAW);

        // Create binding
        auto binding = m_vao->binding(0);
        binding->setAttribute(0);
        binding->setBuffer(m_bufferVertices.get(), 0, sizeof(glm::vec4));
        binding->setFormat(4, gl::GL_FLOAT, gl::GL_FALSE, 0);
        m_vao->enable(0);
    }

    // Unbind VAO
    m_vao->unbind();
}

unsigned int PlotGeometry::numElements() const
{
    return m_numElements;
}

unsigned int PlotGeometry::size() const
{
    return m_size;
}

globjects::VertexArray * PlotGeometry::vao()
{
    return m_vao.get();
}

const globjects::VertexArray * PlotGeometry::vao() const
{
    return m_vao.get();
}

void PlotGeometry::draw() const
{
    if (m_vao)
    {
        m_vao->drawArrays(gl::GL_LINE_STRIP, 0, m_numElements);
        m_vao->unbind();
    }
}


} // namespace gtx
