
#include <gtxlib-core/loader/LoaderAjdMat.h>

#include <iostream>
#include <fstream>

#include <cppassist/string/conversion.h>
#include <cppassist/string/manipulation.h>

#include <cppfs/FilePath.h>

#include <gtxlib-core/base/Time.h>
#include <gtxlib-core/data/DataSet.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>


using namespace cppassist;
using namespace cppfs;


namespace
{


std::istream & getline(std::istream & is, std::string & str)
{
    std::istream & res = std::getline(is, str);

    if (!str.empty() && *str.rbegin() == '\r')
    {
        str.erase(str.length() - 1, 1);
    }

    return res;
}


}


namespace gtx
{


CPPEXPOSE_COMPONENT(LoaderAjdMat, gloperate::AbstractLoader)


LoaderAjdMat::LoaderAjdMat(gloperate::Environment * environment)
: gloperate::Loader<DataSet>(environment)
{
}

LoaderAjdMat::~LoaderAjdMat()
{
}

bool LoaderAjdMat::canLoad(const std::string & ext) const
{
    return (string::toLower(ext) == "adjmat");
}

std::vector<std::string> LoaderAjdMat::loadingTypes() const
{
    std::vector<std::string> types;
    types.push_back("Adjacency matrix files (*.adjmat)");
    return types;
}

std::string LoaderAjdMat::allLoadingTypes() const
{
    // Return supported types
    return "*.adjmat";
}

DataSet * LoaderAjdMat::load(const std::string & filename, const cppexpose::Variant &, std::function<void(int, int)> progress) const
{
    // Check filename
    if (filename.empty()) {
        return nullptr;
    }

    // Get path to containing directory
    std::string dir = FilePath(FilePath(filename).directoryPath()).path();

    // Open file
    std::ifstream f(filename);
    std::string line;

     // Create data set
    NodeAttributesList nodeAttributes;
    DataSet * dataset = new DataSet();

    // Read data
    while (getline(f, line))
    {
        // Split command and parameters
        auto items = string::split(line, ' ');
        std::string command = items.size() >= 1 ? items[0] : "";

        // Read node data
        if (command == "NodeData") {
            std::string filename = items.size() >= 2 ? items[1] : "";
            readNodeAttributes(dir + "/" + filename, nodeAttributes);
        }

        // Read adjacency matrix
        else if (command == "Layer") {
            // Get layer
            std::string layerId  = items.size() >= 2 ? items[1] : "";
            int layer = (layerId != "") ? string::fromString<int>(layerId) : 0;

            // Get filename
            std::string filename = items.size() >= 3 ? items[2] : "";
            if (filename == "") continue;

            // Get attribute mapping
            AttributeMapping mapping;
            for (size_t i=3; i<items.size(); i++) {
                // Read mapping string in form 'from:to'
                auto fromTo = string::split(items[i], ':');
                if (fromTo.size() < 1) continue;

                // Get from and to values
                std::string to   = fromTo[0];
                std::string from = fromTo.size() >= 2 ? fromTo[1] : fromTo[0];

                // Add mapping to list
                mapping[from] = to;
            }

            // Read adjacency matrix
            readAdjacencyMatrix(dir + "/" + filename, *dataset, layer, nodeAttributes, mapping);
        }
    }

    // Finalize dataset
    dataset->fixIds();
    dataset->calculateEdgeLength();
    dataset->addStandardAttributes();

    // Done
    if (progress) progress(-1, 0);
    return dataset;
}

void LoaderAjdMat::readNodeAttributes(const std::string & filename, NodeAttributesList & nodeAttributes) const
{
    // Open file
    std::ifstream f(filename);
    std::string line;

    // Read keys
    if (!getline(f, line)) return;
    auto keys = string::split(line, ',');

    // Read values
    while (getline(f, line))
    {
        // Read values for current node
        auto items = string::split(line, ',');

        // Get attribute values for current node
        NodeAttributes values;
        for (size_t i=0; i<items.size(); i++)
        {
            values[keys[i]] = items[i];
        }

        // Add node values to list
        nodeAttributes.push_back(values);
    }
}

void LoaderAjdMat::readAdjacencyMatrix(const std::string & filename, DataSet & dataset, int layer, const NodeAttributesList & nodeAttributes, const AttributeMapping & mapping) const
{
    // Open file
    std::ifstream f(filename);
    std::string line;

    // Get node offset
    size_t nodeOfs = dataset.nodes().size();

    unsigned int numEdges = 0;

    // Read adjacency matrix
    int lineCount = 0;
    while (getline(f, line))
    {
        // Get line of matrix
        auto items = string::split(line, ',');
        size_t numNodes = items.size();

        // Process each node
        for (size_t i=0; i<numNodes; i++) {
            // On first line, add the nodes to the network
            if (lineCount == 0) {
                // Create node
                auto * node = dataset.nodes().create();

                // Add node attributes
                for (auto it : mapping)
                {
                    // Get attribute name and value
                    std::string from  = it.first;
                    std::string to    = it.second;
                    std::string value = nodeAttributes[i].at(from);

                    // Add attribute
                    if (to == "lat") {
                        Position pos = node->position();
                        pos.setLat(string::fromString<float>(value));
                        node->setPosition(pos);
                    } else if (to == "lon") {
                        Position pos = node->position();
                        pos.setLong(string::fromString<float>(value));
                        node->setPosition(pos);
                    } else {
                        node->attribute(to)->setValue(string::fromString<float>(value));
                    }
                }

                // Add layer as attribute
                node->attribute("layer")->setValue((float)layer);
            }

            // Add edge
            if (items[i] == "1") {
                // Get nodes
                auto * node1 = dataset.nodes().get(nodeOfs + lineCount);
                auto * node2 = dataset.nodes().get(nodeOfs + i);

                // Create edge
                auto * edge = dataset.edges().create();
                edge->setNode1(node1);
                edge->setNode2(node2);

                // Count edges
                numEdges++;
            }
        }

        // Next line
        lineCount++;

        // Abort after too many edges
        if (numEdges > 1000000) {
            std::cout << "ERROR: Data set contains more than 1000000 edges. Truncating." << std::endl;
            break;
        }
    }
}


} // namespace gtx
