
#include <gtxlib-core/loader/LoaderDot.h>

#include <map>
#include <fstream>

#include <cppassist/string/conversion.h>
#include <cppassist/string/manipulation.h>

#include <gtxlib-core/data/DataSet.h>
#include <gtxlib-core/data/ItemGroup.h>
#include <gtxlib-core/data/Item.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>


using namespace cppassist;


namespace
{


std::string left(const std::string & string, std::string::size_type length)
{
    return string.substr(0, length);
}

std::string right(const std::string & string, std::string::size_type length)
{
    return string.substr(string.size() - length, length);
}


}


namespace gtx
{


CPPEXPOSE_COMPONENT(LoaderDot2, gloperate::AbstractLoader)


LoaderDot2::LoaderDot2(gloperate::Environment * environment)
: gloperate::Loader<DataSet>(environment)
{
}

LoaderDot2::~LoaderDot2()
{
}

bool LoaderDot2::canLoad(const std::string & ext) const
{
    return (string::toLower(ext) == "dot");
}

std::vector<std::string> LoaderDot2::loadingTypes() const
{
    std::vector<std::string> types;
    types.push_back("Graph files (*.dot)");
    return types;
}

std::string LoaderDot2::allLoadingTypes() const
{
    // Return supported types
    return "*.dot";
}

DataSet * LoaderDot2::load(const std::string & filename, const cppexpose::Variant &, std::function<void(int, int)> progress) const
{
    // Open file
    std::ifstream file(filename);
    if (!file.good())
        return nullptr;

    // Get file size
    file.seekg(0, file.end);
    int fileSize = file.tellg();
    file.seekg(0, file.beg);

    // Create data set
    DataSet * dataset = new DataSet();
    auto & nodes = dataset->nodes();
    auto & edges = dataset->edges();
    Node * node = nullptr; // Current node
    Edge * edge = nullptr; // Current edge

    // Read file line by line
    std::string line;
    while (std::getline(file, line))
    {
        // Update progress
        if (progress) progress(file.tellg(), fileSize);

        // Read line
        line = string::trim(line);
        std::string::size_type isEdge = line.find("--");
        if (isEdge == std::string::npos) isEdge = line.find("->");

        // Parse line
        if (left(line, 2) == "//" || left(line, 2) == "/*" || line.size() == 0) {
            // Comment or empty line
        } else if (right(line, 1) == "[" && isEdge == std::string::npos) {
            // Beginning of a node
            std::string sID = string::trim(line.substr(0, line.size() - 1));
            id_t id = string::fromString<int>(sID);

            // Create node
            node = nodes.create(id);
        } else if ((right(line, 1) == "[" || right(line, 1) == ";") && isEdge != std::string::npos) {
            // Beginning of an edge
            bool noAttributes = (right(line, 1) == ";");
            line = line.substr(0, line.size() - 1);
            int i = isEdge;
            std::string leftString  = string::trim(left(line, (std::string::size_type)i));
            std::string rightString = string::trim(line.substr(i+2));
            id_t id1 = string::fromString<int>(leftString);
            id_t id2 = string::fromString<int>(rightString);

            // Create edge
            edge = edges.create();
            edge->setNode1(nodes.get(id1));
            edge->setNode2(nodes.get(id2));
            edge->attribute("length")->setValue(0.0f);

            // If there are no edge attributes, end edge immediately
            if (noAttributes) {
                edge = nullptr;
            }
        } else if (line == "];") {
            // End node or edge
            node = nullptr;
            edge = nullptr;
        } else if (line.find("=") != std::string::npos) {
            // Attribute of a node or edge
            Item * item = node ? static_cast<Item *>(node) : static_cast<Item *>(edge);
            if (item)
            {
                // Get attribute name and value
                int i = line.find("=");
                std::string name = string::trim(left(line, i));    if (left(name, 1) == "\"") name = name.substr(1, name.size() - 2);
                std::string val  = string::trim(line.substr(i+1)); if (left(val,  1) == "\"") val  = val .substr(1, val. size() - 2);
                float value = string::fromString<float>(val);

                // Position
                if (name == "lat" && node) {
                    // Check value range
                    if (value > 90.0f) value -= 180.0f;

                    // Set position
                    Position pos = node->position();
                    pos.setLat(value);
                    node->setPosition(pos);
                } else if (name == "lon" && node) {
                    // Check value range
                    if (value > 180.0f) value -= 360.0f;

                    // Set position
                    Position pos = node->position();
                    pos.setLong(value);
                    node->setPosition(pos);
                }

                // Attribute
                else {
                    item->attribute(name)->setValue(value);
                }
            }
        }
    }

    // Finalize dataset
    dataset->fixIds();
    dataset->calculateEdgeLength();
    dataset->addStandardAttributes();

    // Done
    if (progress) progress(-1, 0);
    return dataset;
}


} // namespace gtx
