
#include <gtxlib-core/loader/LoaderFanomos.h>

#include <cstring>
#include <fstream>

#include <proj_api.h>

#include <cppassist/string/conversion.h>
#include <cppassist/string/manipulation.h>

#include <gtxlib-core/base/Time.h>
#include <gtxlib-core/data/DataSet.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Line.h>
#include <gtxlib-core/data/AircraftDB.h>


using namespace cppassist;


namespace gtx
{


CPPEXPOSE_COMPONENT(LoaderFanomos, gloperate::AbstractLoader)


/**
*  @brief
*    FANOMOS header
*/
struct FANOMOSHeader
{
    int flightID;               // Flight ID
    std::string time;           // Actual time of arrival/departure (depending on flight type)
    std::string origin;         // Origin air port (ICAO)
    std::string destination;    // Destination air port (ICAO)
    std::string runway;         // Runway designation
    std::string route;          // Route
    std::string callsign;       // Callsign (ICAO)
    std::string planeType;      // Airplane type (ICAO)
    std::string flightType;     // Flight type (arrival/departure)
    int numTracks;              // Number of tracks belonging to the flight
    int trackID;                // Track ID
    std::string begin;          // Time of first plot
    std::string end;            // Time of last plot
    int numPoints;              // Number of plots
};

/**
*  @brief
*    FANOMOS plot
*/
struct FANOMOSPlot {
    int t; // Time in seconds from first plot
    int x; // X (UTM32)
    int y; // Y (UTM32) - 5000000
    int h; // Height in meters above MSL (mean sea level)
    int v; // Velocity on ground in m/s
    int d; // Distance to first plot in m
};

/**
*  @brief
*    Read FANOMOS header
*/
static FANOMOSHeader ReadHeader(const char * line)
{
    // Prepare header
    FANOMOSHeader header;
    char token[256];

    // Parse string
    int          pos = 0;
    const char * cur = line;
    while (cur)
    {
        // Get next delimiter (',' or '\0')
        const char * stop = strchr(cur, ',');
        if (!stop)   stop = strchr(cur, '\0');
        int len = stop ? (stop - cur) : 0;

        // Get token
        strncpy(token, cur, len);
        token[len] = '\0';

        // Save token in structure
        if (pos == 0) {
            // Flight ID
            header.flightID = atoi(token);
        } else if (pos == 1) {
            // Time of arrival/departure [DATE]
            header.time = token;
        } else if (pos == 2) {
            // Time of arrival/departure [TIME]
            header.time = header.time + " " + token;
        } else if (pos == 3) {
            // Origin
            header.origin = token;
        } else if (pos == 4) {
            // Destination
            header.destination = token;
        } else if (pos == 5) {
            // Runway
            header.runway = token;
        } else if (pos == 6) {
            // Route
            header.route = token;
        } else if (pos == 7) {
            // Callsign
            header.callsign = token;
        } else if (pos == 8) {
            // Airplane type
            header.planeType = token;
        } else if (pos == 9) {
            // Airplane type (? there are two values, may belong together to planeType?)
            //header.planeType = header.planeType + "," + token;
            header.planeType = token;
        } else if (pos == 10) {
            // Flight type
            header.flightType = token;
        } else if (pos == 11) {
            // Unknown
        } else if (pos == 12) {
            // Number of tracks
            header.numTracks = atoi(token);
        } else if (pos == 13) {
            // Track ID
            header.trackID = atoi(token);
        } else if (pos == 14) {
            // Time of first plot [DATE]
            header.begin = token;
        } else if (pos == 15) {
            // Time of first plot [TIME]
            header.begin = header.begin + " " + token;
        } else if (pos == 16) {
            // Time of last plot [DATE]
            header.end = token;
        } else if (pos == 17) {
            // Time of last plot [TIME]
            header.end = header.end + " " + token;
        } else if (pos == 18) {
            // Number of plots
            header.numPoints = atoi(token);
        }

        // Next
        if (*stop == ',')
            cur = stop + 1;
        else
            cur = nullptr;
        pos++;
    }

    // Return header
    return header;
}

/**
*  @brief
*    Read FANOMOS plot
*/
static FANOMOSPlot ReadPlot(const char * line)
{
    // Prepare plot
    FANOMOSPlot plot;
    char token[256];

    // Parse string
    int          pos = 0;
    const char * cur = line;
    while (cur)
    {
        // Get next delimiter (',' or '\0')
        const char * stop = strchr(cur, ',');
        if (!stop)   stop = strchr(cur, '\0');
        int len = stop ? (stop - cur) : 0;
        if (len > 0)
        {
            // Get token
            strncpy(token, cur, len);
            token[len] = '\0';

            // Save token in structure
            if (pos == 0) {
                // t
                plot.t = atoi(token);
            } else if (pos == 1) {
                // x
                plot.x = atoi(token);
            } else if (pos == 2) {
                // y
                plot.y = atoi(token);
            } else if (pos == 3) {
                // h
                plot.h = atoi(token);
            } else if (pos == 4) {
                // v
                plot.v = atoi(token);
            } else if (pos == 5) {
                // d
                plot.d = atoi(token);
            }

            // Next
            cur = stop + 1;
            pos++;
        } else cur = nullptr;
    }

    // Return plot
    return plot;
}

/**
*  @brief
*    Convert coordinates from UTM32 to lat/long
*/
static bool convertUTM32ToLatLong(float x, float y, float & latitude, float & longitude)
{
    // Initialize 'UTM32' coordinate system
    projPJ pj_utm32 = pj_init_plus("+proj=utm +zone=32 +ellps=WGS84 +datum=WGS84 +units=m");
    if (!pj_utm32) {
        return false;
    }

    // Initialize 'LatLong' coordinate system
    projPJ pj_latlong = pj_init_plus("+proj=latlong +ellps=WGS84 +datum=WGS84");
    if (!pj_latlong) {
        return false;
    }

    // Convert from 'UTM32' to 'LatLong'
    double newX = x;
    double newY = y;
    pj_transform(pj_utm32, pj_latlong, 1, 1, &newX, &newY, nullptr);

    // Return coordinates
    latitude  = newY * RAD_TO_DEG;
    longitude = newX * RAD_TO_DEG;

    // Deinitialize
    pj_free(pj_utm32);
    pj_free(pj_latlong);
    return false;
}

LoaderFanomos::LoaderFanomos(gloperate::Environment * environment)
: gloperate::Loader<DataSet>(environment)
{
}

LoaderFanomos::~LoaderFanomos()
{
}

bool LoaderFanomos::canLoad(const std::string & ext) const
{
    return (string::toLower(ext) == "exp");
}

std::vector<std::string> LoaderFanomos::loadingTypes() const
{
    std::vector<std::string> types;
    types.push_back("FANOMOS export (*.exp)");
    return types;
}

std::string LoaderFanomos::allLoadingTypes() const
{
    // Return supported types
    return "*.exp";
}

DataSet * LoaderFanomos::load(const std::string & filename, const cppexpose::Variant &, std::function<void(int, int)> progress) const
{
    // Check filename
    if (filename.empty()) {
        return nullptr;
    }

    // Load aircraft DB
    AircraftDB aircraftDB;

     // Create data set
    DataSet * dataset = new DataSet();

    /*
    // Create node attributes
    AttributesDesc * nodeAttrs = dataset->attributesDesc("node");
    nodeAttrs->clear();
    nodeAttrs->addAttribute<ptime>("time");
    nodeAttrs->addAttribute<float>("height");
    nodeAttrs->addAttribute<float>("velocity");
    nodeAttrs->addAttribute<float>("acceleration");
    nodeAttrs->addAttribute<float>("angle");
    nodeAttrs->addAttribute<float>("aggrSpeed");
    nodeAttrs->addAttribute<int>  ("heightStage");
    nodeAttrs->addAttribute<int>  ("accPhase");
    nodeAttrs->addAttribute<int>  ("curve");

    // Create line attributes
    AttributesDesc * lineAttrs = dataset->attributesDesc("line");
    lineAttrs->clear();
    lineAttrs->addAttribute<int>("flightid");
    lineAttrs->addAttribute<QString>("ata");
    lineAttrs->addAttribute<QString>("origin");
    lineAttrs->addAttribute<QString>("destination");
    lineAttrs->addAttribute<QString>("runway");
    lineAttrs->addAttribute<QString>("route");
    lineAttrs->addAttribute<QString>("callsign");
    lineAttrs->addAttribute<QString>("planeType");
    lineAttrs->addAttribute<int>("flighttype"); // 0 == arrival, 1 == departure
    lineAttrs->addAttribute<int>("weightClass");
    lineAttrs->addAttribute<int>("numTracks");
    lineAttrs->addAttribute<int>("trackID");
    lineAttrs->addAttribute<QString>("begin");
    lineAttrs->addAttribute<QString>("end");
    lineAttrs->addAttribute<int>("numPoints");
    */

    // Header:
    //   FlightID, ATA/ATD, Origin, Destination, Runway, Route, Callsing, PlaneType, FlightType, NumberOfTracks, TrackID, Begin, End, NumberOfPlots

    // Plot:
    //   t, x, y, h, v, d

    // Open file
    FILE * f = fopen(filename.c_str(), "r");
    if (!f) {
        return nullptr;
    }

    // Get file size
    fseek(f, 0, SEEK_END);
    int fileSize = ftell(f);
    fseek(f, 0, SEEK_SET);

    // Current header and trajectory
    FANOMOSHeader header;
    Line * trajectory = nullptr;

    // Read file line by line
    char line[256];
    int  pos = 0;
    while (fgets(line, 256, f)) {
        // Update progress
        if (progress) progress(ftell(f), fileSize);

        // Kill LF
        int len = strlen(line);
        if (line[len-1] == '\n') line[len-1] = '\0';

        // Check if line is a header or a plot line
        if (strchr(line, ':')) {
            // Read header
            header = ReadHeader(line);

            // End trajectory
            if (trajectory) {
                trajectory = nullptr;
            }

            // Get weight class
            int weight = 0;
            std::string planeType = header.planeType;
            std::string weightClass = aircraftDB.icaoAircraft(planeType).weightClass;
            if (weightClass.empty()) {
                weightClass = aircraftDB.aircraft(planeType).weightClass;
            }
                 if (weightClass == "Small")  weight = 0;
            else if (weightClass == "Medium") weight = 1;
            else if (weightClass == "Large")  weight = 2;
            else if (weightClass == "Heavy")  weight = 3;

            // Create a new trajectory
            trajectory = dataset->lines().create();
            trajectory->attribute("flightid")->setValue(header.flightID);
//          trajectory->attributes().setValue<QString>("ata",         header.time);
//          trajectory->attributes().setValue<QString>("origin",      header.origin);
//          trajectory->attributes().setValue<QString>("destination", header.destination);
//          trajectory->attributes().setValue<QString>("runway",      header.runway);
//          trajectory->attributes().setValue<QString>("route",       header.route);
//          trajectory->attributes().setValue<QString>("callsign",    header.callsign);
//          trajectory->attributes().setValue<QString>("planeType",   header.planeType);
            trajectory->attribute("flighttype")->setValue(header.flightType == "Arrival" ? 0.0f : 1.0f);
            trajectory->attribute("weightClass")->setValue(weight);
            trajectory->attribute("numTracks")->setValue(header.numTracks);
            trajectory->attribute("trackID")->setValue(header.trackID);
//          trajectory->attributes().setValue<QString>("begin",       header.begin);
//          trajectory->attributes().setValue<QString>("end",         header.end);
            trajectory->attribute("numPoints")->setValue(header.numPoints);
            pos = 0;
        } else {
            // Read plot
            FANOMOSPlot plot = ReadPlot(line);

            // Convert to lat/long
            float latitude  = 0.0f;
            float longitude = 0.0f;
            convertUTM32ToLatLong(plot.x, plot.y+5000000, latitude, longitude);

            // Get height
            float _h = plot.h;

            // Get speed
            float _v = plot.v;

            // Get time
            auto _time = Time::fromTimeString(header.time) + std::chrono::seconds(plot.t);

            // Create point
            Node * node = dataset->nodes().create();
            node->setPosition(Position::LatLong(latitude, longitude, _h));
            node->setTime(_time);
            node->attribute("time")->setValue((float)node->timeInt());
            node->attribute("height")->setValue(_h);
            node->attribute("velocity")->setValue(_v);

            // Add point to trajectory
            trajectory->add(node);

            // Next
            pos++;
        }
    }

    // Close file
    fclose(f);

    // Analyze trajectories and compute additional attributes
    dataset->analyzeTrajectories();

    // Finalize dataset
    dataset->fixIds();
    dataset->addStandardAttributes();

    // Done
    if (progress) progress(-1, 0);
    return dataset;
}


} // namespace gtx
