
#include <gtxlib-core/loader/LoaderGraphML.h>

#include <iostream>
#include <fstream>

#include <QDebug>
#include <QFile>
#include <QXmlStreamReader>

#include <gtxlib-core/base/Time.h>
#include <gtxlib-core/data/DataSet.h>
#include <gtxlib-core/data/ItemGroup.h>
#include <gtxlib-core/data/Item.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>


namespace gtx
{


CPPEXPOSE_COMPONENT(LoaderGraphML2, gloperate::AbstractLoader)


LoaderGraphML2::LoaderGraphML2(gloperate::Environment * environment)
: gloperate::Loader<DataSet>(environment)
{
}

LoaderGraphML2::~LoaderGraphML2()
{
}

bool LoaderGraphML2::canLoad(const std::string & ext) const
{
    return (QString::fromStdString(ext).toLower() == "gml" || QString::fromStdString(ext).toLower() == "graphml");
}

std::vector<std::string> LoaderGraphML2::loadingTypes() const
{
    std::vector<std::string> types;
    types.push_back("GraphML files (*.gml *.graphml)");
    return types;
}

std::string LoaderGraphML2::allLoadingTypes() const
{
    // Return supported types
    return "*.gml *.graphml";
}

DataSet * LoaderGraphML2::load(const std::string & filename, const cppexpose::Variant &, std::function<void(int, int)> progress) const
{
    // Open file
    QFile file(QString::fromStdString(filename));
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << QString("Failed to open file for reading: %1").arg(QString::fromStdString(filename));
        return nullptr;
    }

    // Create data set
    DataSet * dataset = new DataSet();

    // Initialize
    Node * node = nullptr;           // Current node
    Edge * edge = nullptr;           // Current edge
    QHash<QString, id_t> mapNameID;  // Mapping from node-name to node-ID
    QHash<QString, bool> duplicates; // Track duplicate edges (e.g., "01_02" == "02_01")

    auto standardName = [] (const std::string & name) -> std::string
    {
        if (name == "x" || name == "X" || name == "Longitude" || name == "longitude" || name == "long" || name == "Long" || name == "LONG" || name == "lon" || name == "Lon" || name == "LON") {
            return "lon";
        }

        if (name == "y" || name == "Y" || name == "Latitude" || name == "latitude" || name == "lat" || name == "Lat" || name == "LAT") {
            return "lat";
        }

        return name;
    };

    // Keep track of attributes
    std::map<std::string, std::string> attributeNames;     // id -> name
    std::map<std::string, std::string> nodeAttributeTypes; // id -> type
    std::map<std::string, std::string> edgeAttributeTypes; // id -> type

    // Additional options
    std::vector<std::string> mapLayers;

    // Start XML parser
    QXmlStreamReader xmlParser(&file);
    while (!xmlParser.atEnd()) {
        // Update progress
        if (progress) progress(file.pos(), file.size());

        // Read next XML node
        xmlParser.readNext();

        // Get element
        QString element = "";
        bool startElement = xmlParser.isStartElement();
        if (xmlParser.isStartElement() || xmlParser.isEndElement()) {
            element = xmlParser.name().toString();
        }

        // <maplayer filename="layer1.png" />
        if (startElement && element == "maplayer") {
            std::string filename = xmlParser.attributes().value("filename").toString().toStdString();
            mapLayers.push_back(filename);
        }

        // <key id="lon" for="node" attr.name="Longitude" attr.type="double" />
        else if (startElement && element == "key") {
            // Get attribute description
            std::string id   = xmlParser.attributes().value("id").toString().toStdString();
            std::string name = standardName(xmlParser.attributes().value("attr.name").toString().toStdString());
            std::string dest = xmlParser.attributes().value("for").toString().toStdString();
            std::string type = xmlParser.attributes().value("attr.type").toString().toStdString();

            // Save attribute
            attributeNames[id] = name;
            if (dest == "node") {
                nodeAttributeTypes[id] = type;
            } else if (dest == "edge") {
                edgeAttributeTypes[id] = type;
            }
        }

        // <node id="NodeID">
        else if (startElement && element == "node") {
            // Get node ID
            QString name = xmlParser.attributes().value("id").toString();

            // Create node
            node = dataset->nodes().create();
            node->setName(name.toStdString());

            // Map node-name to node-id
            mapNameID.insert(name, node->id());
        }

        // <edge source="NodeID" target="NodeID">
        else if (startElement && element == "edge") {
            // Get node IDs
            QString sourceName = xmlParser.attributes().value("source").toString();
            QString targetName = xmlParser.attributes().value("target").toString();
            id_t    sourceID   = mapNameID.value(sourceName);
            id_t    targetID   = mapNameID.value(targetName);

            // Identify duplicate edges (directed graphs)
            QString dupID1 = sourceName + "_" + targetName;
            QString dupID2 = targetName + "_" + sourceName;
            if (!duplicates.contains(dupID1) && !duplicates.contains(dupID2)) {
                // Remember this edge in both directions
                duplicates.insert(dupID1, true);
                duplicates.insert(dupID2, true);

                // Create edge
                auto * node1 = dataset->nodes().get(sourceID);
                auto * node2 = dataset->nodes().get(targetID);
                if (node1 && node2) {
                    edge = dataset->edges().create();
                    edge->setNode1(node1);
                    edge->setNode2(node2);
                }
            }
        }

        // <data key="keyName"> value </data>
        else if (startElement && element == "data") {
            // Get attribute
            std::string id    = xmlParser.attributes().value("key").toString().toStdString();
            std::string name  = standardName(attributeNames[id]);
            QString     value = xmlParser.readElementText();

            // Position
            if (node && name == "lat") {
                Position pos = node->position(); pos.setLat (value.toFloat()); node->setPosition(pos);
            } else if (node && name == "lon") {
                Position pos = node->position(); pos.setLong(value.toFloat()); node->setPosition(pos);
            } else if (node && name == "time") {
                node->setTime(Time::fromTimeString(value.toStdString()));
                node->attribute("time")->setValue((float)node->timeInt());
            }

            // Attribute
            else {
                // Set attribute value on node/edge
                Item * item = node ? static_cast<Item *>(node) : static_cast<Item *>(edge);
                auto   type = node ? nodeAttributeTypes[id] : edgeAttributeTypes[id];
                if (item && (type == "double" || type == "float" || type == "real")) {
                    item->attribute(name)->setValue(value.toFloat());
                }
            }
        }

        // </node>
        else if (!startElement && element == "node") {
            node = nullptr;
        }

        // </edge>
        else if (!startElement && element == "edge") {
            edge = nullptr;
        }
    }

    // Close file
    file.close();

    // Add additional data
    dataset->setMapLayers(mapLayers);

    /*
    // Convert data set
    DataSet * dataset2 = convertDataSet(dataset);

    // Release data set
    delete dataset;
    dataset = dataset2;

    // Export data
    exportDataSet(dataset, "out.graphml");
    */

    // Finalize dataset
    dataset->fixIds();
    dataset->calculateEdgeLength();
    dataset->addStandardAttributes();

    // Done
    if (progress) progress(-1, 0);
    return dataset;
}

DataSet * LoaderGraphML2::convertDataSet(DataSet * dataset) const
{
    // Get nodes and edges
    auto & nodes = dataset->nodes();
    auto & edges = dataset->edges();

    // Create new data set
    DataSet * dataset2 = new DataSet;

    // Get edge attributes (which equal the dates of the temporal data set)
    Edge * edge = edges.get(0);
    auto attributes = edge->attributes();

    // Create a copy of the network for each day
    for (auto date : attributes) {
        // Get time stamp
        auto time = Time::fromDateString(date);

        // Get beginning of nodes for the day
        id_t nodeOfs = dataset2->nodes().size();

        // Copy all nodes
        for (auto it : nodes.map()) {
            // Get node
            Node * srcNode = it.second;

            // Create new node
            Node * node = dataset2->nodes().create();
            node->setName(srcNode->name() + "-" + date);
            node->setPosition(srcNode->position());
            node->setTime(time);
            node->attribute("id")->setValue(srcNode->id());
            node->attribute("population")->setValue( srcNode->attribute("Population")->value() );
            node->attribute("temperature")->setValue( srcNode->attribute(date)->value() );
        }

        // Copy all edges
        for (auto it : edges.map()) {
            // Get edge
            Edge * srcEdge = it.second;

            // Create new edge
            Edge * edge = dataset2->edges().create();
            edge->setNode1( dataset2->nodes().get(nodeOfs + srcEdge->node1()->id()) );
            edge->setNode2( dataset2->nodes().get(nodeOfs + srcEdge->node2()->id()) );
            edge->attribute("id")->setValue(srcEdge->id());
            edge->attribute("weight")->setValue( srcEdge->attribute(date)->value() );
        }
    }

    // Done
    return dataset2;
}

void LoaderGraphML2::exportDataSet(DataSet * dataset, const std::string & filename) const
{
    // Open file
    std::ofstream out;
    out.open(filename);

    // Get nodes and edges
    auto & nodes = dataset->nodes();
    auto & edges = dataset->edges();

    // Export graphml
    out << "<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd\">" << std::endl;

    out << "  <key attr.name=\"lat\" attr.type=\"double\" for=\"node\" id=\"lat\" />" << std::endl;
    out << "  <key attr.name=\"lon\" attr.type=\"double\" for=\"node\" id=\"lon\" />" << std::endl;
    out << "  <key attr.name=\"time\" attr.type=\"double\" for=\"node\" id=\"time\" />" << std::endl;

    auto nodeAttributes = nodes.get(0)->attributes();
    for (auto attr : nodeAttributes) {
        out << "  <key attr.name=\"" << attr << "\" attr.type=\"double\" for=\"node\" id=\"" << attr << "\" />" << std::endl;
    }

    auto edgeAttributes = edges.get(0)->attributes();
    for (auto attr : edgeAttributes) {
        out << "  <key attr.name=\"" << attr << "\" attr.type=\"double\" for=\"edge\" id=\"" << attr << "\" />" << std::endl;
    }

    out << "  <graph edgedefault=\"directed\">" << std::endl;

    for (auto it : nodes.map()) {
        auto * node = it.second;

        out << "    <node id=\"" << node->name() << "\">" << std::endl;
        out << "      <data key=\"lat\">" << node->position().getLat() << "</data>" << std::endl;
        out << "      <data key=\"lon\">" << node->position().getLong() << "</data>" << std::endl;
        out << "      <data key=\"time\">" << Time::toString(node->time()) << "</data>" << std::endl;

        auto attributes = node->attributes();
        for (auto attr : attributes) {
            out << "      <data key=\"" << attr << "\">" << node->attribute(attr)->value() << "</data>" << std::endl;
        }

        out << "    </node>" << std::endl;
    }

    for (auto it : edges.map()) {
        auto * edge = it.second;

        out << "    <edge source=\"" << edge->node1()->name() << "\" target=\"" << edge->node2()->name() << "\">" << std::endl;

        auto attributes = edge->attributes();
        for (auto attr : attributes) {
            out << "      <data key=\"" << attr << "\">" << edge->attribute(attr)->value() << "</data>" << std::endl;
        }

        out << "    </edge>" << std::endl;
    }

    out << "  </graph>" << std::endl;
    out << "</graphml>" << std::endl;
}


} // namespace gtx
