
#include <gtxlib-core/loader/LoaderGtx.h>

#include <cppassist/logging/logging.h>
#include <cppassist/fs/RawFile.h>

#include <gtxlib-core/base/DataStream.h>
#include <gtxlib-core/data/Dataset.h>
#include <gtxlib-core/data/AttributesDesc.h>
#include <gtxlib-core/data/ItemList.h>
#include <gtxlib-core/data/Node.h>
#include <gtxlib-core/data/Edge.h>
#include <gtxlib-core/data/EdgeBundle.h>
#include <gtxlib-core/data/Line.h>


namespace
{
    // Header
    const uint32_t VERSION_MAJOR = 1;
    const uint32_t VERSION_MINOR = 0;

    // Bundle type
    const uint8_t TYPE_RECURSIVE_BUNDLE = 1;
    const uint8_t TYPE_EDGE_BUNDLE = 2;
}


namespace gtx
{


CPPEXPOSE_COMPONENT(LoaderGtx, gloperate::AbstractLoader)


LoaderGtx::LoaderGtx(gloperate::Environment * environment)
: gloperate::Loader<Dataset>(environment)
{
}

LoaderGtx::~LoaderGtx()
{
}

bool LoaderGtx::canLoad(const std::string & ext) const
{
    return ext == "gtx";
}

std::vector<std::string> LoaderGtx::loadingTypes() const
{
    std::vector<std::string> types;
    types.push_back("Native binary format (*.gtx)");
    return types;
}

std::string LoaderGtx::allLoadingTypes() const
{
    return "*.gtx";
}

uint32_t LoaderGtx::magicNumber() const
{
    // 'CLIM'
    return ('C' << 24) + ('L' << 16) + ('I' << 8) + 'M';
}

Dataset * LoaderGtx::load(const std::string & filename, const cppexpose::Variant &, std::function<void(int, int)> progress) const
{
    // Open file
    cppassist::RawFile file;
    if (!file.load(filename)) {
        cppassist::debug() << "Failed to open file for reading: " << filename;
        return nullptr;
    }

    // Create data set
    Dataset * dataset = new Dataset();

    std::vector<uint8_t> dataVector(file.data(), file.data() + file.size());

    // Read from file through stream
    DataStream stream(&dataVector);

    // Verify header and version
    uint32_t magic, major, minor;
    stream >> magic
           >> major
           >> minor;
    if (magic != magicNumber() || major != VERSION_MAJOR || minor != VERSION_MINOR) {
        cppassist::debug() << std::string("File format mismatch.");
        delete dataset;
        return nullptr;
    }

    // Read data types
    uint32_t numTypes;
    stream >> numTypes;
    for (size_t i=0; i < numTypes; i++) {
        std::string name;
        stream >> name;
        readAttributesDesc(stream, dataset->attributesDesc(name));
    }

    // Read nodes
    uint32_t numNodes;
    id_t     nodeId;
    stream >> numNodes;
    for (size_t i=0; i<numNodes; i++) {
        stream >> nodeId;
        Position pos = readVec3(stream);
        Node * node = new Node(nodeId, dataset->attributesDesc("node"));
        node->setPosition(pos);
        if (dataset->attributesDesc("node")->attributeCount() > 0)
            readAttributes(stream, &node->attributes());
        dataset->add(node);
        if (progress) progress(stream.position(), file.size());
    }

    // Read edges
    uint32_t numEdges;
    id_t     edgeId, nodeId1, nodeId2;
    stream >> numEdges;
    for (size_t i=0; i<numEdges; i++) {
        stream >> edgeId >> nodeId1 >> nodeId2;
        Node * node1 = (Node*)dataset->item(nodeId1);
        Node * node2 = (Node*)dataset->item(nodeId2);
        Edge * edge = new Edge(edgeId, dataset->attributesDesc("edge"), node1, node2);
        if (dataset->attributesDesc("edge")->attributeCount() > 0)
            readAttributes(stream, &edge->attributes());
        dataset->add(edge);
        if (i % 10 == 0 && progress) progress(stream.position(), file.size());
    }

    // Read bundles
    uint32_t numBundles;
    stream >> numBundles;
    for (size_t i=0; i<numBundles; i++) {
        EdgeBundle * bundle = readBundle(stream, dataset);
        dataset->add(bundle);
        if (dataset->attributesDesc("bundle")->attributeCount() > 0)
            readAttributes(stream, &bundle->attributes());
        if (i % 10 == 0 && progress) progress(stream.position(), file.size());
    }

    // Read lines
    uint32_t numLines;
    id_t     lineId;
    stream >> numLines;
    for (size_t i=0; i<numLines; i++) {
        stream >> lineId;
        Line * line = new Line(lineId, dataset->attributesDesc("line"));
        if (dataset->attributesDesc("line")->attributeCount() > 0)
            readAttributes(stream, &line->attributes());

        int32_t firstId = null_id, lastId = null_id;
        stream >> firstId;
        while (firstId != null_id) {
            stream >> lastId;
            if (firstId == null_id || lastId == null_id || firstId > lastId) {
                cppassist::debug() << std::string("Error reading line nodes.");
                break;
            }
            for (int32_t id = firstId; id <= lastId; id++) {
                Node *node = (Node*)dataset->item(id);
                if (node) line->add(node);
            }
            stream >> firstId;
        }
        dataset->add(line);
        if (i % 10 == 0 && progress) progress(stream.position(), file.size());
    }

    // Read number of traffic flow data sets (empty for all other data sets)
    if (!stream.atEnd()) {
        uint32_t numTrafficFlowDatasets;
        int32_t trafficFlowTimeInterval;
        stream >> numTrafficFlowDatasets;
        stream >> trafficFlowTimeInterval;
        dataset->setNumTrafficFlowDatasets(numTrafficFlowDatasets);
        dataset->setTrafficFlowTimeInterval(trafficFlowTimeInterval);
    }

    // Close file
    cppassist::debug() << "Loaded file from " << filename << ".";

    // Done
    if (progress) progress(-1, 0);
    return dataset;
}

void LoaderGtx::readAttributesDesc(DataStream & stream, AttributesDesc * desc) const
{
    uint32_t attrCount, type;
    std::string name;

    desc->clear();
    stream >> attrCount;

    for (size_t i = 0; i < attrCount; i++) {
        stream >> type >> name;

        if (type == (int) AttributeValueBase::TypeFloat)
            desc->addAttribute<float>(name);
        else if (type == (int) AttributeValueBase::TypeInt)
            desc->addAttribute<int>(name);
        else if (type == (int) AttributeValueBase::TypeTime)
            desc->addAttribute<int>(name);
        else if (type == (int) AttributeValueBase::TypeString)
            desc->addAttribute<std::string>(name);
        else if (type == (int) AttributeValueBase::TypeUnknown)
            cppassist::debug() << "Warning: Reading unknown Attribute type.";
    }
}

void LoaderGtx::readAttributes(DataStream & stream, Attributes * attrs) const
{
    const AttributesDesc * desc = attrs->attributesDesc();

    for (int i = 0; i < desc->attributeCount(); i++)
    {
        AttributeValueBase * value = attrs->attribute(i);
        AttributeDescriptorBase * descriptor = (AttributeDescriptorBase*) desc->descriptor(i);

        switch (descriptor->type()) {
            case AttributeValueBase::TypeFloat: {
                double v;
                stream >> v;
                ((AttributeValue<float>*) value)->setValue(v);
                ((AttributeDescriptor<float>*) descriptor)->update(v);
                break;
            }
            case AttributeValueBase::TypeInt: {
                int32_t n;
                stream >> n;
                ((AttributeValue<int>*) value)->setValue(n);
                ((AttributeDescriptor<int>*) descriptor)->update(n);
                break;
            }
            case AttributeValueBase::TypeTime: {
                uint16_t year, month, day;
                int32_t hours, minutes, seconds, fractional_seconds;

                stream >> year
                       >> month
                       >> day
                       >> hours
                       >> minutes
                       >> seconds
                       >> fractional_seconds;

                // dummy values
                ((AttributeValue<int>*) value)->setValue(0);
                ((AttributeDescriptor<int>*) descriptor)->update(0);
                break;
            }
            case AttributeValueBase::TypeString: {
                std::string s;
                stream >> s;
                ((AttributeValue<std::string>*) value)->setValue(s);
                ((AttributeDescriptor<std::string>*) descriptor)->update(s);
                break;
            }
            case AttributeValueBase::TypeUnknown: {
                cppassist::debug() << std::string("Warning: Unknown Attribute type.");
                break;
            }
            default:
                break;
        }
    }
}

Position LoaderGtx::readVec3(DataStream & stream) const
{
    double x, y, z;
    stream >> x >> y >> z;
    return Position(x, y, z);
}

EdgeBundle * LoaderGtx::readBundle(DataStream & stream, Dataset * dataset) const
{
    int32_t id;
    uint8_t bundleType;

    // Read bundle
    stream >> id;
    Position start = readVec3(stream);
    Position end   = readVec3(stream);
    stream >> bundleType;
    EdgeBundle * bundle = new EdgeBundle(start, end, id);

    // Check if bundle contains sub-bundles or edges
    if (bundleType == TYPE_RECURSIVE_BUNDLE) {
        uint32_t numBundles;
        stream >> numBundles;
        for (size_t i=0; i<numBundles; i++) {
            bundle->addChildBundle(readBundle(stream, dataset));
        }
    } else if (bundleType == TYPE_EDGE_BUNDLE) {
        uint32_t numEdges;
        stream >> numEdges;
        for (size_t i=0; i<numEdges; i++) {
            int32_t edgeId;
            stream >> edgeId;
            bundle->addEdge(dataset->edges().get(edgeId));
        }
    }

    // Return bundle
    return bundle;
}


} // namespace gtx
