
#pragma once


#include <cppexpose/plugin/plugin_api.h>

#include <vector>
#include <string>

#include <glm/fwd.hpp>

#include <gloperate/base/ExtendedProperties.h>
#include <gloperate/pipeline/Pipeline.h>
#include <gloperate/pipeline/Input.h>
#include <gloperate/stages/interfaces/CanvasInterface.h>
#include <gloperate/rendering/Camera.h>
#include <gloperate/input/AbstractEventConsumer.h>

#include <gtx/gtx-version.h>

#include <gtxlib-stages/gtxlib-stages_api.h>


namespace gloperate
{
    class TimerStage;
    class TextureStage;
    class BasicFramebufferStage;
    class TextureRenderTargetStage;
    class RasterizationStage;
    class ClearStage;
    class ColorRenderTarget;
    class DepthRenderTarget;
    class TextureFromRenderTargetExtractionStage;
    class BlitStage;
}


namespace gtx
{


class TrackballStage;
class WorldStage;
class LineStage;
class TrajectoryStage;
class DensityMapStage;
class MapOverlayStage;
class DataLoaderStage;
class GpuDataStage;
class MappingStage;
class FilterStage;
class TemporalFilterStage;
class ColorMapStage;
class TextureArrayStage;
class OutputStage;
class MeasurementStage;
class PickingStage;
class ScreenshotStage;


/**
*  @brief
*    GTX pipeline for air traffic visualization
*/
class GTXLIB_STAGES_API AirTrafficPipeline : public gloperate::Pipeline, gloperate::AbstractEventConsumer
{
public:
    CPPEXPOSE_DECLARE_COMPONENT(
        AirTrafficPipeline, gloperate::Pipeline
      , "RenderStage Gtx" // Tags
      , ""                // Icon
      , ""                // Annotations
      , "Main GTX pipeline"
      , GTX_AUTHOR_ORGANIZATION
      , "v1.0.0"
    )


public:
    // Interfaces
    gloperate::CanvasInterface canvasInterface;            ///< Interface for rendering into a viewer
    Input<gloperate::ColorRenderTarget *>  colorTarget;    ///< Color render target
    Input<gloperate::DepthRenderTarget *>  depthTarget;    ///< Depth render target
    Output<gloperate::ColorRenderTarget *> colorTargetOut; ///< Color render target (passthrough)

    // Inputs
    Input<cppfs::FilePath>     filename;            ///< Path to the data set

    // Geographic projection
    Input<std::string>         projection;          ///< The chosen map projection

    // General options
    Input<gloperate::Color>    backgroundColor;     ///< Background color
    Input<float>               widthScale;          ///< Width scale factor

    // Line rendering
    Input<bool>                linesVisible;        ///< Render lines?
    Input<std::string>         nodeHeight;          ///< Node height mapping
    Input<std::string>         lineColorMap;        ///< Line color map
    Input<std::string>         lineColor;           ///< Line color mapping
    Input<std::string>         lineWidth;           ///< Line width mapping

    // Trajectory rendering
    Input<bool>                trajectoriesVisible; ///< Render trajectories?
    Input<int>                 drawMode;            ///< 1: cpu-triangles-indexed, 2: cpu-strips-multiple-calls, 3: cpu-strips-primitive-restart, 5: gpu
    Input<unsigned int>        maxNodes;            ///< Maximum number of nodes to render
    Input<std::string>         primaryAttribute;    ///< Primary attribute for dynamic mapping

    // Density maps
    Input<bool>                densityMapVisible;   ///< Render density map?

    // Time measurement
    Input<bool>                measureRendering;    ///< 'true' to measure rendering time, else 'false'
    Input<bool>                measureUpdate;       ///< 'true' to measure update time, else 'false'

    // Interaction status
    Input<glm::vec2>           mousePos;            ///< The current mouse position on the screen


public:
    static void initialize();


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] environment
    *    Environment to which the stage belongs (must NOT be null!)
    *  @param[in] name
    *    Stage name
    */
    AirTrafficPipeline(gloperate::Environment * environment, const std::string & name = "");

    /**
    *  @brief
    *    Destructor
    */
    virtual ~AirTrafficPipeline();


protected:
    void updateMappingValues(const std::vector<std::string> & values);

    // Virtual AbstractEventConsumer interface
    virtual void onEvent(gloperate::InputEvent * event) override;


protected:
    // Stages
    std::unique_ptr<ColorMapStage>                                     m_colorMap;                ///< Load the color maps
    std::unique_ptr<TextureArrayStage>                                 m_textures;                ///< Load the textures
    std::unique_ptr<TrackballStage>                                    m_trackball;               ///< Camera navigation
    std::unique_ptr<gloperate::TimerStage>                             m_timer;                   ///< Maintains a virtual time

    std::unique_ptr<DataLoaderStage>                                   m_dataLoader;              ///< Loads the data set
    std::unique_ptr<GpuDataStage>                                      m_gpuData;                 ///< Creates GPU representations of the data
    std::unique_ptr<MappingStage>                                      m_mapping;                 ///< Stage that selects an attribute and determines its index
    std::unique_ptr<FilterStage>                                       m_filter;                  ///< Stage that enables filtering the data set
    std::unique_ptr<TemporalFilterStage>                               m_temporalFilter;          ///< Stage that enables temporal filtering

    std::unique_ptr<WorldStage>                                        m_world;                   ///< Creates the world map
    std::unique_ptr<DensityMapStage>                                   m_densityMap;              ///< Renders a density map
    std::unique_ptr<MapOverlayStage>                                   m_mapOverlay;              ///< Displays an overlay on the map
    std::unique_ptr<LineStage>                                         m_lines;                   ///< Renders lines
    std::unique_ptr<TrajectoryStage>                                   m_trajectories;            ///< Renders trajectories

    std::unique_ptr<gloperate::BasicFramebufferStage>                  m_worldFramebuffer;        ///< Framebuffer for the world
    std::unique_ptr<gloperate::ClearStage>                             m_worldClear;              ///< Clears the world framebuffer
    std::unique_ptr<gloperate::RasterizationStage>                     m_worldRasterization;      ///< Executes rendering of the world sphere
    std::unique_ptr<gloperate::RasterizationStage>                     m_overlayRasterization;    ///< Executes map overlay rendering
    std::unique_ptr<gloperate::TextureFromRenderTargetExtractionStage> m_worldOutput;             ///< Extracts the world output texture

    std::unique_ptr<gloperate::BasicFramebufferStage>                  m_overlayFramebuffer;      ///< Framebuffer for the overlays
    std::unique_ptr<gloperate::TextureRenderTargetStage>               m_idTarget;                ///< Create the render target for the ID texture
    std::unique_ptr<gloperate::ClearStage>                             m_overlayClear;            ///< Clears the overlay framebuffer
    std::unique_ptr<gloperate::RasterizationStage>                     m_lineRasterization;       ///< Executes line rendering
    std::unique_ptr<gloperate::RasterizationStage>                     m_trajectoryRasterization; ///< Executes trajectory rendering
    std::unique_ptr<gloperate::RasterizationStage>                     m_helpersRasterization;    ///< Executes helpers rendering
    std::unique_ptr<gloperate::TextureFromRenderTargetExtractionStage> m_overlayOutput;           ///< Extracts the overlay output texture

    std::unique_ptr<OutputStage>                                       m_output;                  ///< Outputs the final image
    std::unique_ptr<gloperate::BasicFramebufferStage>                  m_outputFramebuffer;       ///< Output framebuffer
    std::unique_ptr<gloperate::ClearStage>                             m_outputClear;             ///< Clears the output framebuffer
    std::unique_ptr<gloperate::RasterizationStage>                     m_outputRasterization;     ///< Renders the final image
    std::unique_ptr<PickingStage>                                      m_picking;                 ///< Stage that performs the picking (selection of objects using an ID buffer)
    std::unique_ptr<MeasurementStage>                                  m_measurement;             ///< Prints the performance measurements
    std::unique_ptr<ScreenshotStage>                                   m_screenshot;              ///< Stage that takes a screenshot

    std::unique_ptr<gloperate::BlitStage>                              m_blit;                    ///< Blits the final image

    // Interaction state
    glm::ivec2 m_mousePos; ///< Current mouse position
};


} // namespace gtx
