
#pragma once


#include <cppexpose/plugin/plugin_api.h>

#include <vector>
#include <set>
#include <string>

#include <glm/fwd.hpp>

#include <gloperate/base/ExtendedProperties.h>
#include <gloperate/pipeline/Pipeline.h>
#include <gloperate/pipeline/Input.h>
#include <gloperate/stages/interfaces/CanvasInterface.h>
#include <gloperate/rendering/Camera.h>
#include <gloperate/input/AbstractEventConsumer.h>

#include <gtx/gtx-version.h>

#include <gtxlib-stages/gtxlib-stages_api.h>


namespace gloperate
{
    class TextureStage;
    class BasicFramebufferStage;
    class TextureRenderTargetStage;
    class RasterizationStage;
    class ClearStage;
    class ColorRenderTarget;
    class DepthRenderTarget;
    class TextureFromRenderTargetExtractionStage;
}


namespace gtx
{


class TrackballStage;
class SelectionStage;
class WorldStage;
class NodeStage;
class EdgeStage;
class PlotStage;
class DataLoaderStage;
class GpuDataStage;
class MappingStage;
class FilterStage;
class TemporalFilterStage;
class ColorMapStage;
class PickingStage;
class OutputStage;


/**
*  @brief
*    GTX pipeline for climate network visualization
*/
class GTXLIB_STAGES_API ClimatePipeline : public gloperate::Pipeline, gloperate::AbstractEventConsumer
{
public:
    CPPEXPOSE_DECLARE_COMPONENT(
        ClimatePipeline, gloperate::Pipeline
      , "RenderStage Gtx" // Tags
      , ""                // Icon
      , ""                // Annotations
      , "Main GTX pipeline"
      , GTX_AUTHOR_ORGANIZATION
      , "v1.0.0"
    )


public:
    // Interfaces
    gloperate::CanvasInterface canvasInterface;            ///< Interface for rendering into a viewer
    Input<gloperate::ColorRenderTarget *>  colorTarget;    ///< Color render target
    Input<gloperate::DepthRenderTarget *>  depthTarget;    ///< Depth render target
    Output<gloperate::ColorRenderTarget *> colorTargetOut; ///< Color render target (passthrough)

    // Inputs
    Input<cppfs::FilePath>     filename;         ///< Path to the data set

    // Geographic projection
    Input<std::string>         projection;       ///< The chosen map projection

    // General options
    Input<gloperate::Color>    backgroundColor;  ///< Background color
    Input<float>               nodeScale;        ///< Node scale factor
    Input<float>               widthScale;       ///< Width scale factor
    Input<float>               heightScale;      ///< Node height scale factor
    Input<float>               arcScale;         ///< Arc scale factor
    Input<float>               arcTess;          ///< Arc tessellation factor
    Input<bool>                edgesOnTop;       ///< 'true' to render edges on top of nodes
    Input<float>               neighborhood;     ///< k-neighborhood for selection
    Input<bool>                unsharpMasking;   ///< Enable unsharp masking?
    Input<float>               unsharpFactor;    ///< Factor by which the effect is intensified

    // Node rendering
    Input<bool>                nodesVisible;     ///< Render nodes?
    Input<std::string>         nodeColorMap;     ///< Node color map
    Input<std::string>         nodeColor;        ///< Node color mapping
    Input<std::string>         nodeSize;         ///< Node size mapping
    Input<std::string>         nodeHeight;       ///< Node height mapping
    Input<float>               nodeContextColor; ///< Opaqueness of unselected nodes

    // Edge rendering
    Input<bool>                edgesVisible;     ///< Render edges?
    Input<bool>                directConnection; ///< 'true' to connect edges by the shortest path
    Input<std::string>         edgeColorMap;     ///< Edge color map
    Input<std::string>         edgeColor;        ///< Edge color mapping
    Input<std::string>         edgeWidth;        ///< Edge width mapping
    Input<float>               edgeContextColor; ///< Opaqueness of unselected edges

    // Plot rendering
    Input<bool>                plotVisible;      ///< Render plot?
    Input<gloperate::Color>    plotColor;        ///< Line color of plot
    Input<gloperate::Color>    plotBackground;   ///< Background color of plot

    // Interaction status
    Input<glm::vec2>           mousePos;         ///< The current mouse position on the screen
    Input<int>                 selectedId;       ///< Currently selected node
    Input< std::set<int> >     selectedIds;      ///< List of selected nodes
    Input<cppexpose::Variant>  selectionArea;    ///< Area on screen to select


public:
    static void initialize();


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] environment
    *    Environment to which the stage belongs (must NOT be null!)
    *  @param[in] name
    *    Stage name
    */
    ClimatePipeline(gloperate::Environment * environment, const std::string & name = "");

    /**
    *  @brief
    *    Destructor
    */
    virtual ~ClimatePipeline();


protected:
    void updateMappingValues(const std::vector<std::string> & values);

    // Virtual Stage interface
    virtual void onInputValueChanged(gloperate::AbstractSlot * slot) override;

    // Virtual AbstractEventConsumer interface
    virtual void onEvent(gloperate::InputEvent * event) override;


protected:
    bool isNodeSelected(int nodeId) const;
    void setNodeSelected(int nodeId, bool selected);
    void resetSelection();


protected:
    // Stages
    std::unique_ptr<gloperate::BasicFramebufferStage>                  m_framebuffer;        ///< Creates a framebuffer
    std::unique_ptr<gloperate::TextureRenderTargetStage>               m_idTarget;           ///< Create the render target for the ID texture

    std::unique_ptr<ColorMapStage>                                     m_colorMap;           ///< Load the color maps
    std::unique_ptr<TrackballStage>                                    m_trackball;          ///< Camera navigation

    std::unique_ptr<DataLoaderStage>                                   m_dataLoader;         ///< Loads the data set
    std::unique_ptr<GpuDataStage>                                      m_gpuData;            ///< Creates GPU representations of the data
    std::unique_ptr<MappingStage>                                      m_mapping;            ///< Stage that selects an attribute and determines its index
    std::unique_ptr<FilterStage>                                       m_filter;             ///< Stage that enables filtering the data set
    std::unique_ptr<TemporalFilterStage>                               m_temporalFilter;     ///< Stage that enables temporal filtering

    std::unique_ptr<SelectionStage>                                    m_selection;          ///< Maintains and update the selection buffer

    std::unique_ptr<WorldStage>                                        m_world;              ///< Creates the world map
    std::unique_ptr<NodeStage>                                         m_nodes;              ///< Renders nodes
    std::unique_ptr<EdgeStage>                                         m_edges;              ///< Renders edges
    std::unique_ptr<PlotStage>                                         m_plot;               ///< Plot attribute values

    std::unique_ptr<gloperate::ClearStage>                             m_clear;              ///< Clears the screen
    std::unique_ptr<gloperate::RasterizationStage>                     m_worldRasterization; ///< Executes rendering of the world sphere
    std::unique_ptr<gloperate::RasterizationStage>                     m_nodeRasterization;  ///< Executes node rendering
    std::unique_ptr<gloperate::RasterizationStage>                     m_edgeRasterization;  ///< Executes edge rendering
    std::unique_ptr<gloperate::RasterizationStage>                     m_bgRasterization;    ///< Executes background rendering
    std::unique_ptr<gloperate::RasterizationStage>                     m_plotRasterization;  ///< Executes plot rendering
    std::unique_ptr<gloperate::TextureFromRenderTargetExtractionStage> m_extractColor;       ///< Extracts color output texture
    std::unique_ptr<gloperate::TextureFromRenderTargetExtractionStage> m_extractDepth;       ///< Extracts depth output texture

    std::unique_ptr<OutputStage>                                       m_output;             ///< Outputs the final image
    std::unique_ptr<gloperate::RasterizationStage>                     m_outputRasterization;///< Executes output rendering

    std::unique_ptr<PickingStage>                                      m_picking;            ///< Stage that performs the picking (selection of objects using an ID buffer)

    // Interaction state
    glm::ivec2    m_mousePos;    ///< Current mouse position
    std::set<int> m_selectedIds; ///< List of currently selected nodes
    int           m_selectedId;  ///< First selected node
};


} // namespace gtx
