
#pragma once


#include <vector>
#include <string>

#include <cppexpose/plugin/plugin_api.h>

#include <gloperate/pipeline/Stage.h>
#include <gloperate/pipeline/Input.h>
#include <gloperate/pipeline/Output.h>

#include <gtx/gtx-version.h>

#include <gtxlib-core/base/Configuration.h>

#include <gtxlib-stages/gtxlib-stages_api.h>


namespace cppexpose {
    class Variant;
}


namespace gtx
{


/**
*  @brief
*    Stage that configures a mapping configuration
*/
class GTXLIB_STAGES_API ConfigurationStage : public gloperate::Stage
{
public:
    CPPEXPOSE_DECLARE_COMPONENT(
        ConfigurationStage, gloperate::Stage
      , ""   // Tags
      , ""   // Icon
      , ""   // Annotations
      , "Stage that configures a mapping configuration"
      , GTX_AUTHOR_ORGANIZATION
      , "v1.0.0"
    )


public:
    // Inputs
    Input<std::vector<std::string>> attributeNames; ///< Available attributes
    Input<std::vector<std::string>> textureNames;   ///< List of textures
    Input<std::vector<std::string>> colorMapNames;  ///< Available color maps

    Input<std::string> geometryType;  ///< Tubes or spheres
    Input<float> tessellation;        ///< Tessellation factor (default: 16)
    Input<std::string> texture;       ///< Name of texture to use
    Input<std::string> colorMap;      ///< Name of color map
    Input<std::string> mapHeight;     ///< Attribute that is mapped on height
    Input<float> height;              ///< Static value for height
    Input<std::string> mapRadius;     ///< Attribute that is mapped on radius
    Input<float> radius;              ///< Static value for tube radius
    Input<std::string> mapColor;      ///< Attribute that is mapped on color
    Input<float> color;               ///< Static value for color
    Input<std::string> mapAlpha;      ///< Attribute that is mapped on alpha
    Input<float> alpha;               ///< Static value for alpha
    Input<std::string> mapTextureX;   ///< Attribute that is mapped on texture.x
    Input<float> textureX;            ///< Static value for texture stretch
    Input<std::string> mapTextureY;   ///< Attribute that is mapped on texture.y
    Input<float> textureY;            ///< Static value for texture torsion
    Input<float> minRadius;           ///< Minimum radius (used when radius == 0.0)
    Input<float> maxRadius;           ///< Maximum radius (used when radius == 1.0)
    Input<float> sizeOfSingleStretch; ///< Length of one texture before it repeats
    Input<float> sizeOfSingleTorsion; ///< Length of one texture torsion
    Input<float> animationSpeed;      ///< Speed of animation (in seconds)

    // Outputs
    Output<Configuration *> configuration; ///< Configuration object


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] environment
    *    Environment to which the stage belongs (must NOT be null!)
    *  @param[in] name
    *    Stage name
    */
    ConfigurationStage(gloperate::Environment * environment, const std::string & name = "ConfigurationStage");

    /**
    *  @brief
    *    Destructor
    */
    virtual ~ConfigurationStage();


protected:
    // Virtual Stage interface
    virtual void onProcess() override;


protected:
    void updateAttributes(const std::vector<std::string> & attributes);
    void updateTextures(const std::vector<std::string> & filenames);
    void updateColorMaps(const std::vector<std::string> & filenames);
    cppexpose::Variant toVariant(const std::vector<std::string> & values) const;
    float mappingValue(const std::string & mapping, float staticValue) const;
    int indexOf(const std::vector<std::string> & values, const std::string & value) const;


protected:
    std::vector<std::string> m_geometries;    ///< List of geometry types
    std::vector<std::string> m_attributes;    ///< List of attributes
    std::vector<std::string> m_textures;      ///< List of textures
    std::vector<std::string> m_colorMaps;     ///< List of color maps
    Configuration            m_configuration; ///< Configuration object
};


} // namespace gtx
