
#pragma once


#include <glm/glm.hpp>

#include <cppexpose/plugin/plugin_api.h>

#include <gloperate/pipeline/Pipeline.h>
#include <gloperate/pipeline/Input.h>
#include <gloperate/pipeline/Output.h>
#include <gloperate/base/ExtendedProperties.h>

#include <gtx/gtx-version.h>

#include <gtxlib-stages/gtxlib-stages_api.h>


namespace globjects
{
    class Texture;
    class Shader;
}

namespace gloperate
{
    class AbstractDrawable;
    class Camera;
    class ProgramStage;
    class RenderPassStage;
}


namespace gtx
{


/**
*  @brief
*    Stage that renders the edges of a data set
*/
class GTXLIB_STAGES_API EdgeStage : public gloperate::Pipeline
{
public:
    CPPEXPOSE_DECLARE_COMPONENT(
        EdgeStage, gloperate::Pipeline
      , ""   // Tags
      , ""   // Icon
      , ""   // Annotations
      , "Stage that renders the edges of a data set"
      , GTX_AUTHOR_ORGANIZATION
      , "v0.1.0"
    )


public:
    // Inputs
    Input<glm::vec4> viewport;                           ///< Viewport
    Input<gloperate::Camera *> camera;                   ///< Camera configuration
    Input<globjects::Shader *> projectionVertexShader;   ///< Vertex shader for transforming geographic coordinates into world coordinates
    Input<globjects::Shader *> projectionGeometryShader; ///< Geometry shader for transforming geographic coordinates into world coordinates
    Input<gloperate::AbstractDrawable *> edgeGeometry;   ///< Geometry containing the edges of the data set
    Input<globjects::Texture *> edgeColorMapTexture;     ///< Color map for the edges
    Input<globjects::Texture *> staticAttributesTexture; ///< Texture buffer containing the static attributes
    Input<globjects::Texture *> nodeAttributesTexture;   ///< Texture buffer containing the node attributes
    Input<globjects::Texture *> edgeAttributesTexture;   ///< Texture buffer containing the edge attributes
    Input<globjects::Texture *> filtersTexture;          ///< Texture buffer containing the current filter values
    Input<globjects::Texture *> selectionTexture;        ///< Texture that contains the current selection of nodes
    Input<int> numStaticAttributes;                      ///< Number of static attributes
    Input<int> numNodeAttributes;                        ///< Number of node attributes
    Input<int> numEdgeAttributes;                        ///< Number of edge attributes
    Input<int> edgeColorIndex;                           ///< Index of attribute which is mapped to edge color
    Input<int> edgeWidthIndex;                           ///< Index of attribute which is mapped to edge width
    Input<int> nodeHeightIndex;                          ///< Index of attribute which is mapped to node height
    Input<float> edgeContextColor;                       ///< Opaqueness of unselected edges
    Input<float> widthScale;                             ///< Edge width scale factor
    Input<float> arcScale;                               ///< Arc height scale factor
    Input<float> arcTess;                                ///< Arc tessellation factor
    Input<bool> edgesOnTop;                              ///< 'true' to render lines above nodes
    Input<bool> selection;                               ///< 'true' to enable interactive selection
    Input<bool> directConnection;                       ///< 'true' to connect edges by the shortest path
    Input<float> neighborhood;                           ///< k-neighborhood for selection
    Input<float> heightScale;                            ///< Node height scale factor
    Input<glm::vec2> timeSpan;                           ///< Entire time span
    Input<glm::vec2> timeSpan1;                          ///< Temporal filter span #1
    Input<glm::vec2> timeSpan2;                          ///< Temporal filter span #2
    Input<glm::vec2> timeSpan3;                          ///< Temporal filter span #3
    Input<glm::vec2> timeSpan4;                          ///< Temporal filter span #4

    // Outputs
    Output<gloperate::AbstractDrawable *> renderPass; ///< Render pass that renders the edges


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] environment
    *    Environment to which the stage belongs (must NOT be null!)
    *  @param[in] name
    *    Stage name
    */
    EdgeStage(gloperate::Environment * environment, const std::string & name = "EdgeStage");

    /**
    * @brief
    *   Destructor
    */
    ~EdgeStage();


protected:
    std::unique_ptr<gloperate::ProgramStage>    m_program;    ///< Builds the program for edge rendering
    std::unique_ptr<gloperate::RenderPassStage> m_renderPass; ///< Builds the render pass for edge rendering
};


} // namespace gtx
