
#pragma once


#include <vector>
#include <string>

#include <glm/glm.hpp>

#include <cppexpose/plugin/plugin_api.h>

#include <gloperate/pipeline/Stage.h>
#include <gloperate/pipeline/Input.h>
#include <gloperate/pipeline/Output.h>
#include <gloperate/base/ExtendedProperties.h>

#include <gtx/gtx-version.h>

#include <gtxlib-stages/gtxlib-stages_api.h>


namespace globjects
{
    class Texture;
}

namespace gloperate
{
    class AbstractDrawable;
}


namespace gtx
{


class DataSet;
class NodeGeometry;
class PatchGeometry;
class EdgeGeometry;
class LineGeometry;
class AttributeStorage;


/**
*  @brief
*    Stage that creates the GPU representations of the input data
*/
class GTXLIB_STAGES_API GpuDataStage : public gloperate::Stage
{
public:
    CPPEXPOSE_DECLARE_COMPONENT(
        GpuDataStage, gloperate::Stage
      , ""   // Tags
      , ""   // Icon
      , ""   // Annotations
      , "Stage that creates the GPU representations of the input data"
      , GTX_AUTHOR_ORGANIZATION
      , "v1.0.0"
    )


public:
    // Inputs
    Input<gtx::DataSet *>                 dataset;                 ///< Dataset object
    Input<unsigned int>                   maxNodes;                ///< Maximum number of nodes to render

    // Outputs
    Output<gloperate::AbstractDrawable *> nodeGeometry;            ///< Node geometry drawable (as points)
    Output<gloperate::AbstractDrawable *> patchGeometry;           ///< Node geometry drawable (as patches)
    Output<gloperate::AbstractDrawable *> edgeGeometry;            ///< Edge geometry drawable
    Output<gloperate::AbstractDrawable *> lineGeometry;            ///< Line geometry drawable
    Output<unsigned int>                  nodeGeometrySize;        ///< Size of node geometry (in bytes)
    Output<globjects::Texture *>          mapLayersTexture;        ///< Texture containing additional map layers
    Output<globjects::Texture *>          staticAttributesTexture; ///< Static attributes texture
    Output<globjects::Texture *>          nodeAttributesTexture;   ///< Node attributes texture
    Output<globjects::Texture *>          edgeAttributesTexture;   ///< Edge attributes texture
    Output<globjects::Texture *>          lineAttributesTexture;   ///< Line attributes texture
    Output<std::vector<std::string>>      attributeNames;          ///< List of attributes
    Output<int>                           numStaticAttributes;     ///< Number of static attributes
    Output<int>                           numNodeAttributes;       ///< Number of node attributes
    Output<int>                           numEdgeAttributes;       ///< Number of edge attributes
    Output<int>                           numLineAttributes;       ///< Number of line attributes
    Output<glm::vec4>                     area;                    ///< Area in which the data is defined


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] environment
    *    Environment to which the stage belongs (must NOT be null!)
    *  @param[in] name
    *    Stage name
    */
    GpuDataStage(gloperate::Environment * environment, const std::string & name = "GpuDataStage");

    /**
    *  @brief
    *    Destructor
    */
    virtual ~GpuDataStage();


protected:
    // Virtual Stage interface
    virtual void onContextDeinit(gloperate::AbstractGLContext * context) override;
    virtual void onProcess() override;

    // Helper functions
    void createMapLayersTexture();


protected:
    std::unique_ptr<NodeGeometry>       m_nodeGeometry;     ///< Node geometry
    std::unique_ptr<PatchGeometry>      m_patchGeometry;    ///< Node geometry
    std::unique_ptr<EdgeGeometry>       m_edgeGeometry;     ///< Edge geometry
    std::unique_ptr<LineGeometry>       m_lineGeometry;     ///< Line geometry
    std::unique_ptr<AttributeStorage>   m_attributeStorage; ///< Attribute storage
    std::unique_ptr<globjects::Texture> m_mapLayersTexture; ///< Map layers
};


} // namespace gtx
