
#pragma once


#include <vector>

#include <cppexpose/plugin/plugin_api.h>

#include <gloperate/pipeline/Stage.h>
#include <gloperate/pipeline/Input.h>
#include <gloperate/pipeline/Output.h>
#include <gloperate/base/ExtendedProperties.h>

#include <gtx/gtx-version.h>

#include <gtxlib-stages/gtxlib-stages_api.h>


namespace globjects
{
    class Texture;
    class Buffer;
}

namespace gloperate
{
    class ColorRenderTarget;
}


namespace gtx
{


/**
*  @brief
*    Stage that performs the picking (selection of objects using an ID buffer)
*/
class GTXLIB_STAGES_API PickingStage : public gloperate::Stage
{
public:
    CPPEXPOSE_DECLARE_COMPONENT(
        PickingStage, gloperate::Stage
      , ""   // Tags
      , ""   // Icon
      , ""   // Annotations
      , "Stage that performs the picking (selection of objects using an ID buffer)"
      , GTX_AUTHOR_ORGANIZATION
      , "v1.0.0"
    )


public:
    // Inputs
    Input<gloperate::ColorRenderTarget *> idTarget; ///< ID render target
    Input<glm::vec2>                      mousePos; ///< The current mouse position on the screen
    Input<bool>                           fixId;    ///< If 'true', id errors through blending are fixed by adding +1 (-1 = default value in buffer)
    Input<int>                            radius;   ///< If radius > 1, selectedIDs will be filled with all IDs in the given radius

    // Outputs
    Output<int>                  hoveredID;   ///< ID of currently hovered node
    Output<globjects::Texture *> selectedIDs; ///< Texture containing the selected IDs in a given radius
    Output<int>                  numIDs;      ///< Number of selected IDs in selectedIDs


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] environment
    *    Environment to which the stage belongs (must NOT be null!)
    *  @param[in] name
    *    Stage name
    */
    PickingStage(gloperate::Environment * environment, const std::string & name = "PickingStage");

    /**
    * @brief
    *   Destructor
    */
    ~PickingStage();

    /**
    *  @brief
    *    Get ID at position
    *
    *  @param[in] x
    *    X-position
    *  @param[in] y
    *    Y-position
    *
    *  @return
    *    ID of selected object, or -1
    */
    int pickAt(int x, int y);


protected:
    // Virtual Stage interface
    virtual void onContextInit(gloperate::AbstractGLContext *) override;
    virtual void onContextDeinit(gloperate::AbstractGLContext *) override;
    virtual void onProcess() override;


protected:
    std::vector<float>                  m_idBuffer;         ///< Current selection buffer
    int                                 m_width;            ///< Selection buffer width
    int                                 m_height;           ///< Selection buffer height
    std::unique_ptr<globjects::Texture> m_selectionTexture; ///< Texture containing all selected IDs
    std::unique_ptr<globjects::Buffer>  m_selectionBuffer;  ///< Buffer containing all selected IDs
};


} // namespace gtx
