
#pragma once


#include <cppexpose/plugin/plugin_api.h>

#include <string>
#include <memory>
#include <vector>
#include <set>

#include <gloperate/pipeline/Stage.h>
#include <gloperate/pipeline/Input.h>
#include <gloperate/pipeline/Output.h>

#include <gtx/gtx-version.h>

#include <gtxlib-stages/gtxlib-stages_api.h>


namespace globjects
{
    class Framebuffer;
    class Texture;
    class Program;
    class Shader;
}

namespace gloperate
{
    class AbstractDrawable;
}


namespace gtx
{


class DataSet;


/**
*  @brief
*    Stage that computes the selection for each node
*/
class GTXLIB_STAGES_API SelectionStage : public gloperate::Stage
{
public:
    CPPEXPOSE_DECLARE_COMPONENT(
        SelectionStage, gloperate::Stage
      , ""   // Tags
      , ""   // Icon
      , ""   // Annotations
      , "Stage that computes the selection for each node"
      , GTX_AUTHOR_ORGANIZATION
      , "v1.0.0"
    )


public:
    // Inputs
    Input<gtx::DataSet *>                dataset;                 ///< Dataset object
    Input<std::vector<std::string>>      attributes;              ///< List of attributes
    Input<gloperate::AbstractDrawable *> edgeGeometry;            ///< Geometry containing the edges of the data set
    Input< std::set<int> >               selectedIds;             ///< List of selected nodes
    Input<globjects::Texture *>          staticAttributesTexture; ///< Texture buffer containing the static attributes
    Input<globjects::Texture *>          nodeAttributesTexture;   ///< Texture buffer containing the node attributes
    Input<int>                           numStaticAttributes;     ///< Number of static attributes
    Input<int>                           numNodeAttributes;       ///< Number of node attributes

    // Outputs
    Output<bool>                         selectionActive;  ///< 'true' if at least one node has been selected, else 'false'
    Output<globjects::Texture *>         selectionTexture; ///< Texture that contains the current selection of nodes


public:
    /**
    *  @brief
    *    Constructor
    *
    *  @param[in] environment
    *    Environment to which the stage belongs (must NOT be null!)
    *  @param[in] name
    *    Stage name
    */
    SelectionStage(gloperate::Environment * environment, const std::string & name = "SelectionStage");

    /**
    *  @brief
    *    Destructor
    */
    virtual ~SelectionStage();


protected:
    // Virtual Stage interface
    virtual void onContextInit(gloperate::AbstractGLContext *) override;
    virtual void onContextDeinit(gloperate::AbstractGLContext *) override;
    virtual void onProcess() override;


protected:
    // OpenGL objects
    std::unique_ptr<globjects::Texture>     m_texture1;         ///< Selection texture #1
    std::unique_ptr<globjects::Texture>     m_texture2;         ///< Selection texture #2
    std::unique_ptr<globjects::Framebuffer> m_fbo1;             ///< FBO for rendering into texture #1
    std::unique_ptr<globjects::Framebuffer> m_fbo2;             ///< FBO for rendering into texture #2
    std::unique_ptr<globjects::Program>     m_program;          ///< Shader program
    std::unique_ptr<globjects::Shader>      m_vertexShader;     ///< Vertex shader
    std::unique_ptr<globjects::Shader>      m_geometryShader;   ///< Geometry shader
    std::unique_ptr<globjects::Shader>      m_fragmentShader;   ///< Fragment shader
    std::unique_ptr<globjects::Shader>      m_attributesShader; ///< Attributes shader (geometry)
};


} // namespace gtx
