
#include <gtxlib-stages/ConfigurationStage.h>

#include <iostream>

#include <cppexpose/variant/Variant.h>


using namespace cppexpose;


namespace gtx
{


CPPEXPOSE_COMPONENT(ConfigurationStage, gloperate::Stage)


ConfigurationStage::ConfigurationStage(gloperate::Environment * environment, const std::string & name)
: gloperate::Stage(environment, "ConfigurationStage", name)
, attributeNames("attributeNames", this)
, textureNames("textureNames", this)
, colorMapNames("colorMapNames", this)
, geometryType("geometryType", this, "Tubes")
, tessellation("tessellation", this, 16.0f)
, texture("texture", this, "tube_texture_arrow.png")
, colorMap("colorMap", this, "color_gradient")
, mapHeight("mapHeight", this, "Node - height")
, height("height", this, 0.0f)
, mapRadius("mapRadius", this, "None")
, radius("radius", this, 0.0f)
, mapColor("mapColor", this, "None")
, color("color", this, 0.5f)
, mapAlpha("mapAlpha", this, "None")
, alpha("alpha", this, 1.0f)
, mapTextureX("mapTextureX", this, "Node - length")
, textureX("textureX", this, 0.0f)
, mapTextureY("mapTextureY", this, "None")
, textureY("textureY", this, 0.0f)
, minRadius("minRadius", this, 0.00002f)
, maxRadius("maxRadius", this, 0.00008f)
, sizeOfSingleStretch("sizeOfSingleStretch", this, 1.0f)
, sizeOfSingleTorsion("sizeOfSingleTorsion", this, 1.0f)
, animationSpeed("animationSpeed", this, 1.0f)
, configuration("configuration", this)
{
    // Set options
    attributeNames.setOption("hidden", true);
    textureNames.setOption("hidden", true);
    colorMapNames.setOption("hidden", true);

    m_geometries.push_back("None");
    m_geometries.push_back("Tubes");
    m_geometries.push_back("Spheres");
    geometryType.setOption("choices", toVariant(m_geometries));

    tessellation.setOption("minimumValue", 2.0f);
    tessellation.setOption("maximumValue", 128.0f);

    texture.setOption("choices", toVariant(m_textures));

    colorMap.setOption("choices", toVariant(m_colorMaps));

    height.setOption("minimumValue", 0.0f);
    height.setOption("maximumValue", 20000.0f);

    radius.setOption("minimumValue", 0.0f);
    radius.setOption("maximumValue", 1.0f);

    mapHeight.setOption("choices", std::vector<std::string>());
    mapRadius.setOption("choices", std::vector<std::string>());
    mapColor.setOption("choices", std::vector<std::string>());
    mapAlpha.setOption("choices", std::vector<std::string>());
    mapTextureX.setOption("choices", std::vector<std::string>());
    mapTextureY.setOption("choices", std::vector<std::string>());

    color.setOption("minimumValue", 0.0f);
    color.setOption("maximumValue", 1.0f);

    alpha.setOption("minimumValue", 0.0f);
    alpha.setOption("maximumValue", 1.0f);

    textureX.setOption("minimumValue", 0.0f);
    textureX.setOption("maximumValue", 1.0f);

    textureY.setOption("minimumValue", 0.0f);
    textureY.setOption("maximumValue", 1.0f);

    minRadius.setOption("minimumValue", 0.00001f);
    minRadius.setOption("maximumValue", 0.002f);
    minRadius.setOption("updateOnDrag", true);

    maxRadius.setOption("minimumValue", 0.00001f);
    maxRadius.setOption("maximumValue", 0.002f);
    maxRadius.setOption("updateOnDrag", true);

    sizeOfSingleStretch.setOption("minimumValue", 0.0f);
    sizeOfSingleStretch.setOption("maximumValue", 1.0f);

    sizeOfSingleTorsion.setOption("minimumValue", 0.0f);
    sizeOfSingleTorsion.setOption("maximumValue", 1.0f);

    animationSpeed.setOption("minimumValue", 0.0f);
    animationSpeed.setOption("maximumValue", 1.0f);

    // Updates
    attributeNames.valueChanged.connect([this] (const std::vector<std::string> & attributes)
    {
        updateAttributes(attributes);
    });

    textureNames.valueChanged.connect([this] (const std::vector<std::string> & filenames)
    {
        updateTextures(filenames);
    });

    colorMapNames.valueChanged.connect([this] (const std::vector<std::string> & filenames)
    {
        updateColorMaps(filenames);
    });
}

ConfigurationStage::~ConfigurationStage()
{
}

void ConfigurationStage::onProcess()
{
    // Update configuration
    m_configuration.geometryType        = geometryType.value() == "Spheres" ? 2.0f : geometryType.value() == "Tubes" ? 1.0f : 0.0f;
    m_configuration.tessellation        = tessellation.value();
    m_configuration.textureID           = indexOf(m_textures, texture.value());
    m_configuration.colorMapID          = indexOf(m_colorMaps, colorMap.value());
    m_configuration.height              = height.value();
    m_configuration.radius              = radius.value();
    m_configuration.color               = color.value();
    m_configuration.alpha               = alpha.value();
    m_configuration.textureX            = textureX.value();
    m_configuration.textureY            = textureY.value();
    m_configuration.minRadius           = minRadius.value();
    m_configuration.maxRadius           = maxRadius.value();
    m_configuration.sizeOfSingleStretch = sizeOfSingleStretch.value();
    m_configuration.sizeOfSingleTorsion = sizeOfSingleTorsion.value();
    m_configuration.animationSpeed      = animationSpeed.value();

    // Update mappings
    m_configuration.height   = mappingValue(mapHeight.value(), height.value());
    m_configuration.radius   = mappingValue(mapRadius.value(), radius.value());
    m_configuration.color    = mappingValue(mapColor.value(), color.value());
    m_configuration.alpha    = mappingValue(mapAlpha.value(), alpha.value());
    m_configuration.textureX = mappingValue(mapTextureX.value(), textureX.value());
    m_configuration.textureY = mappingValue(mapTextureY.value(), textureY.value());

    // Set outputs
    configuration.setValue(&m_configuration);
}

void ConfigurationStage::updateAttributes(const std::vector<std::string> & attributes)
{
    // Update attributes
    m_attributes.clear();
    m_attributes.push_back("None");
    for (std::string attr : attributes)
    {
        m_attributes.push_back(attr);
    }

    // Update choices on mapping attributes
    mapHeight  .setOption("choices", toVariant(m_attributes));
    mapRadius  .setOption("choices", toVariant(m_attributes));
    mapColor   .setOption("choices", toVariant(m_attributes));
    mapAlpha   .setOption("choices", toVariant(m_attributes));
    mapTextureX.setOption("choices", toVariant(m_attributes));
    mapTextureY.setOption("choices", toVariant(m_attributes));
}

void ConfigurationStage::updateTextures(const std::vector<std::string> & filenames)
{
    m_textures = filenames;
    texture.setOption("choices", toVariant(m_textures));
}

void ConfigurationStage::updateColorMaps(const std::vector<std::string> & filenames)
{
    m_colorMaps = filenames;
    colorMap.setOption("choices", toVariant(m_colorMaps));
}

cppexpose::Variant ConfigurationStage::toVariant(const std::vector<std::string> & values) const
{
    Variant choices = Variant::array();

    for (std::string value : values)
    {
        choices.asArray()->push_back(value);
    }

    return choices;
}

float ConfigurationStage::mappingValue(const std::string & mapping, float staticValue) const
{
    // Check if value is static or shall be mapped
    if (mapping != "None")
    {
        // It shall be mapped. Get index of attribute it is mapped to
        auto index = indexOf(m_attributes, mapping);

        // Encode attribute
        int value = -(int)index;

        // Return value
        return (float)(value);
    }

    // Not mapped, so return static value
    return staticValue;
}

int ConfigurationStage::indexOf(const std::vector<std::string> & values, const std::string & value) const
{
    auto index = find(values.begin(), values.end(), value) - values.begin();
    return index;
}


} // namespace gtx
