
#include <gtxlib-stages/DemoPipeline.h>

#include <iostream>

#include <cppassist/memory/make_unique.h>

#include <cppexpose/variant/Variant.h>

#include <glbinding/gl/enum.h>

#include <gloperate/stages/base/BasicFramebufferStage.h>
#include <gloperate/stages/base/ClearStage.h>
#include <gloperate/stages/base/RasterizationStage.h>
#include <gloperate/stages/base/TimerStage.h>
#include <gloperate/stages/base/BlitStage.h>

#include <gtxlib-core/gtxlib-core.h>

#include <gtxlib-stages/TrackballStage.h>
#include <gtxlib-stages/WorldStage.h>
#include <gtxlib-stages/TrajectoryStage.h>
#include <gtxlib-stages/DataLoaderStage.h>
#include <gtxlib-stages/TemporalFilterStage.h>
#include <gtxlib-stages/GpuDataStage.h>
#include <gtxlib-stages/MappingStage.h>
#include <gtxlib-stages/ColorMapStage.h>
#include <gtxlib-stages/TextureArrayStage.h>


using namespace cppassist;
using namespace cppexpose;
using namespace gloperate;


namespace gtx
{


CPPEXPOSE_COMPONENT(DemoPipeline, gloperate::Pipeline)


void DemoPipeline::initialize()
{
}

DemoPipeline::DemoPipeline(Environment * environment, const std::string & name)
: Pipeline(environment, "DemoPipeline", name)
, canvasInterface(this)
, colorTarget("colorTarget", this)
, depthTarget("depthTarget", this)
, colorTargetOut("colorTargetOut", this)
, m_colorMap(cppassist::make_unique<ColorMapStage>(environment, "ColorMap"))
, m_textures(cppassist::make_unique<TextureArrayStage>(environment, "Textures"))
, m_trackball(cppassist::make_unique<TrackballStage>(environment, "Trackball"))
, m_timer(cppassist::make_unique<TimerStage>(environment, "Timer"))
, m_dataLoader(cppassist::make_unique<DataLoaderStage>(environment, "DataLoader"))
, m_gpuData(cppassist::make_unique<GpuDataStage>(environment, "GpuData"))
, m_mapping(cppassist::make_unique<MappingStage>(environment, "Mapping"))
, m_temporalFilter(cppassist::make_unique<TemporalFilterStage>(environment, "TemporalFilter"))
, m_world(cppassist::make_unique<WorldStage>(environment, "World"))
, m_trajectories(cppassist::make_unique<TrajectoryStage>(environment, "Trajectories"))
, m_framebuffer(cppassist::make_unique<BasicFramebufferStage>(environment, "Framebuffer"))
, m_clear(cppassist::make_unique<ClearStage>(environment, "Clear"))
, m_rasterization(cppassist::make_unique<RasterizationStage>(environment, "Rasterization"))
, m_blit(cppassist::make_unique<gloperate::BlitStage>(environment, "Blit"))
{
    // [DEBUG]
//  cppassist::setVerbosityLevel(cppassist::LogMessage::Level::Debug + 2);

    // Get data path
    std::string dataPath = gtx::dataPath();

    // Get dataset path
    std::string datasetPath = gtx::datasetPath();

    // Color map stage
    addStage(m_colorMap.get());

    // Textures stage
    addStage(m_textures.get());
    m_textures->path = dataPath + "/textures";
    *m_textures->createInput<std::string>("tex0") = "tube_texture_none.png";
    *m_textures->createInput<std::string>("tex1") = "tube_texture_arrow.png";
    *m_textures->createInput<std::string>("tex2") = "tube_texture_grid.png";
    *m_textures->createInput<std::string>("tex3") = "tube_texture_daytime.png";

    // Trackball stage
    addStage(m_trackball.get());
    m_trackball->viewport << canvasInterface.viewport;

    // Timer stage
    addStage(m_timer.get());
    m_timer->timeDelta << canvasInterface.timeDelta;
    m_timer->factor = 0.1f;

    // Data loader stage
    addStage(m_dataLoader.get());

    // Gpu data representation stage
    addStage(m_gpuData.get());
    m_gpuData->dataset << m_dataLoader->dataset;

    // Create update trigger for UI when a new dataset has been loaded
    auto * updateTrigger = this->createInput("updateTrigger") << m_gpuData->attributeNames;
    updateTrigger->setOption("hidden", true);

    // Mapping stage
    addStage(m_mapping.get());
    m_mapping->attributes << m_gpuData->attributeNames;
    m_mapping->primaryAttribute = "Zero";
    m_mapping->nodeColor = "Node - velocity";

    // Temporal filter stage
    addStage(m_temporalFilter.get());
    m_temporalFilter->dataset << m_dataLoader->dataset;

    // World map stage
    addStage(m_world.get());
    m_world->projection = "Geocentric";
    m_world->camera << m_trackball->camera;
    m_world->mapLayersTexture << m_gpuData->mapLayersTexture;

    // Trajectory stage
    addStage(m_trajectories.get());
    m_trajectories->drawMode = 5;
    m_trajectories->projectionVertexShader << m_world->projectionVertexShader;
    m_trajectories->projectionGeometryShader << m_world->projectionGeometryShader;
    m_trajectories->dataset << m_dataLoader->dataset;
    m_trajectories->nodeGeometry << m_gpuData->nodeGeometry;
    m_trajectories->patchGeometry << m_gpuData->patchGeometry;
    m_trajectories->nodeGeometrySize << m_gpuData->nodeGeometrySize;
    m_trajectories->camera << m_trackball->camera;
    m_trajectories->staticAttributesTexture << m_gpuData->staticAttributesTexture;
    m_trajectories->nodeAttributesTexture << m_gpuData->nodeAttributesTexture;
    m_trajectories->lineAttributesTexture << m_gpuData->lineAttributesTexture;
    m_trajectories->timeSpan1 << m_temporalFilter->timeSpan1;
    m_trajectories->colorMaps << m_colorMap->colorMapsTexture;
    m_trajectories->textures << m_textures->texture;
    m_trajectories->primaryAttribute = "Zero";
    m_trajectories->primaryAttributeIndex << m_mapping->primaryAttributeIndex;
    m_trajectories->numStaticAttributes << m_gpuData->numStaticAttributes;
    m_trajectories->numNodeAttributes << m_gpuData->numNodeAttributes;
    m_trajectories->numEdgeAttributes << m_gpuData->numEdgeAttributes;
    m_trajectories->numLineAttributes << m_gpuData->numLineAttributes;
    m_trajectories->viewport << canvasInterface.viewport;
    m_trajectories->attributeNames << m_gpuData->attributeNames;
    m_trajectories->textureNames << m_textures->filenames;
    m_trajectories->colorMapNames << m_colorMap->colorMapNames;
    m_trajectories->time << m_timer->virtualTime;
    m_trajectories->hoveredID = -1;

    // Overlay framebuffer stage
    addStage(m_framebuffer.get());
    m_framebuffer->viewport << canvasInterface.viewport;

    // Overlay clear stage
    addStage(m_clear.get());
    m_clear->renderInterface.viewport << canvasInterface.viewport;
    m_clear->createInput("colorTarget") << m_framebuffer->colorBuffer;
    m_clear->createInput("colorValue") = gloperate::Color(0.0f, 0.0f, 0.0f, 1.0f);
    m_clear->createInput("depthTarget") << m_framebuffer->depthBuffer;
    m_clear->createInput("depthValue") = 1.0f;
    m_clear->createInput("camera") << m_trackball->camera;
    m_clear->createInput("world") << m_world->renderPass;
    m_clear->createInput("trajectories") << m_trajectories->renderPass;
    m_clear->createOutput<gloperate::ColorRenderTarget *>("colorTargetOut");
    m_clear->createOutput<gloperate::DepthRenderTarget *>("depthTargetOut");

    // Rasterization stage
    addStage(m_rasterization.get());
    m_rasterization->createInput<gloperate::ColorRenderTarget *>("colorTarget")->connect(m_clear->output("colorTargetOut"));
    m_rasterization->createOutput<gloperate::ColorRenderTarget *>("colorTargetOut");
    m_rasterization->createInput<gloperate::DepthRenderTarget *>("depthTarget")->connect(m_clear->output("depthTargetOut"));
    m_rasterization->createOutput<gloperate::DepthRenderTarget *>("depthTargetOut");
    m_rasterization->renderInterface.viewport << canvasInterface.viewport;
    m_rasterization->drawable << m_trajectories->renderPass;

    // Blit stage
    addStage(m_blit.get());
    m_blit->source.connect(m_rasterization->output("colorTargetOut"));
    m_blit->sourceViewport << canvasInterface.viewport;
    m_blit->target << this->colorTarget;
    m_blit->targetViewport << canvasInterface.viewport;

    // Outputs
    colorTargetOut << m_blit->targetOut;

    // Load default dataset
    m_dataLoader->filePath = datasetPath + "/airtraffic/dataset.demo";
}

DemoPipeline::~DemoPipeline()
{
}


} // namespace gtx
