
#include <gtxlib-stages/FilterStage.h>

#include <cppassist/memory/make_unique.h>

#include <glbinding/gl/enum.h>

#include <globjects/Texture.h>
#include <globjects/Buffer.h>

#include <gloperate/base/Range.h>
#include <gloperate/base/ExtendedProperties.h>

#include <gtxlib-core/gtxlib-core.h>
#include <gtxlib-core/data/DataSet.h>


using namespace gloperate;


namespace gtx
{


CPPEXPOSE_COMPONENT(FilterStage, gloperate::Stage)


FilterStage::FilterStage(Environment * environment, const std::string & name)
: Stage(environment, "FilterStage", name)
, dataset("dataset", this)
, filtersTexture("filtersTexture", this)
, m_updateFilters(true)
{
    // Set options
    dataset.setOption("hidden", true);
}

FilterStage::~FilterStage()
{
}

void FilterStage::onContextDeinit(AbstractGLContext *)
{
    m_filtersTexture = nullptr;
}

void FilterStage::onInputValueChanged(AbstractSlot * slot)
{
    // If dataset has changed, update filter options
    if (slot == &dataset)
    {
        m_updateFilters = true;
    }

    // Invalidate outputs
    filtersTexture.invalidate();
}

void FilterStage::onInputValueInvalidated(AbstractSlot * slot)
{
    // If dataset has changed, update filter options
    if (slot == &dataset)
    {
        m_updateFilters = true;
    }

    // Invalidate outputs
    filtersTexture.invalidate();
}

void FilterStage::onProcess()
{
    // Get data path
    std::string dataPath = gtx::dataPath();

    // Update filter slots
    if (m_updateFilters)
    {
        updateFilters();
    }

    // Update filter texture
    if (!filtersTexture.isValid())
    {
        updateTexture();
    }
}

void FilterStage::updateFilters()
{
    // Remove old filters
    std::vector<AbstractSlot *> inputs = this->inputs();

    for (auto * input : inputs)
    {
        if (input->name() != "dataset")
        {
            removeInput(input);
        }
    }

    // Get data set
    DataSet * dataset = *this->dataset;
    if (dataset)
    {
        // Iterate over nodes and edges
        for (int i : {0, 1})
        {
            // Get item group
            auto & group = (i == 0) ? static_cast<AbstractItemGroup &>(dataset->nodes()) : static_cast<AbstractItemGroup &>(dataset->edges());
            std::string type = (i == 0) ? "node" : "edge";

            // Get attributes
            for (auto attr : group.attributes())
            {
                // Get attribute info
                auto * attributeInfo = group.attribute(attr);

                // Check if property for that attribute already exists
                std::string attrName = type + "_" + attr;
                if (input(attrName) != nullptr) continue;

                // Create property
                auto property = createInput<Range>(attrName);
                property->setOption("minimumValue", attributeInfo->minimumValue());
                property->setOption("maximumValue", attributeInfo->maximumValue());
                property->setValue(Range(attributeInfo->minimumValue(), attributeInfo->maximumValue()));
            }
        }
    }

    // Finished
    m_updateFilters = false;
}

void FilterStage::updateTexture()
{
    static const int maxAttrs  = 32; // Maximum number of attributes per item
    static const int numValues = 4;  // Number of values per attribute (min, max, reserved, reserved)

    // Determine texture size
    int numFloats = numValues * maxAttrs;
    numFloats = ((numFloats / 1024) + 1) * 1024;

    // Create texture data
    float * data = (float*)malloc(numFloats * sizeof(float));
    for (int i=0; i<numFloats; i++) {
        data[i] = 0.0f;
    }

    // Get data set
    DataSet * dataset = *this->dataset;
    if (dataset)
    {
        // Iterate over nodes and edges
        int attrIndex = dataset->statics().attributes().size();
        for (int i : {0, 1})
        {
            // Get item group
            auto & group = (i == 0) ? static_cast<AbstractItemGroup &>(dataset->nodes()) : static_cast<AbstractItemGroup &>(dataset->edges());
            std::string type = (i == 0) ? "node" : "edge";

            // Get attributes
            for (auto attr : group.attributes())
            {
                // Get property for attribute
                std::string attrName = type + "_" + attr;
                auto * prop = static_cast<Input<Range> *>(input(attrName));
                if (!prop) continue;

                // Update filter
                Range range = prop->value();
                data[attrIndex * numValues + 0] = range.minimumValue();
                data[attrIndex * numValues + 1] = range.maximumValue();
                attrIndex++;
            }
        }
    }

    // Create texture
    m_filtersTexture = cppassist::make_unique<globjects::Texture>(gl::GL_TEXTURE_BUFFER);
    globjects::Buffer * buffer = new globjects::Buffer();
    buffer->setData(numFloats * sizeof(float), data, gl::GL_STATIC_DRAW);
    m_filtersTexture->texBuffer(gl::GL_R32F, buffer);

    // Release texture data
    free(data);

    // Update output
    filtersTexture.setValue(m_filtersTexture.get());
}


} // namespace gtx
