
#include <gtxlib-stages/GpuDataStage.h>

#include <cppassist/memory/make_unique.h>

#include <cppfs/FilePath.h>

#include <glbinding/gl/enum.h>

#include <gloperate/base/Environment.h>
#include <gloperate/base/ResourceManager.h>

#include <gtxlib-core/base/Area.h>
#include <gtxlib-core/data/DataSet.h>
#include <gtxlib-core/geometry/NodeGeometry.h>
#include <gtxlib-core/geometry/PatchGeometry.h>
#include <gtxlib-core/geometry/EdgeGeometry.h>
#include <gtxlib-core/geometry/LineGeometry.h>
#include <gtxlib-core/geometry/AttributeStorage.h>

#include <gtxlib-stages/Tools.h>


using namespace cppfs;


namespace gtx
{


CPPEXPOSE_COMPONENT(GpuDataStage, gloperate::Stage)


GpuDataStage::GpuDataStage(gloperate::Environment * environment, const std::string & name)
: gloperate::Stage(environment, "GpuDataStage", name)
, dataset("dataset", this)
, maxNodes("maxNodes", this, 0)
, nodeGeometry("nodeGeometry", this)
, patchGeometry("patchGeometry", this)
, edgeGeometry("edgeGeometry", this)
, lineGeometry("lineGeometry", this)
, nodeGeometrySize("nodeGeometrySize", this, 0)
, mapLayersTexture("mapLayersTexture", this)
, staticAttributesTexture("staticAttributesTexture", this)
, nodeAttributesTexture("nodeAttributesTexture", this)
, edgeAttributesTexture("edgeAttributesTexture", this)
, lineAttributesTexture("lineAttributesTexture", this)
, attributeNames("attributeNames", this)
, numStaticAttributes("numStaticAttributes", this)
, numNodeAttributes("numNodeAttributes", this)
, numEdgeAttributes("numEdgeAttributes", this)
, numLineAttributes("numLineAttributes", this)
, area("area", this)
{
}

GpuDataStage::~GpuDataStage()
{
}

void GpuDataStage::onContextDeinit(gloperate::AbstractGLContext *)
{
    // Release OpenGL objects
    m_nodeGeometry     = nullptr;
    m_patchGeometry    = nullptr;
    m_edgeGeometry     = nullptr;
    m_lineGeometry     = nullptr;
    m_attributeStorage = nullptr;
    m_mapLayersTexture = nullptr;
}

void GpuDataStage::onProcess()
{
    // Create geometries
    m_nodeGeometry  = cppassist::make_unique<gtx::NodeGeometry>();
    m_patchGeometry = cppassist::make_unique<gtx::PatchGeometry>();
    m_edgeGeometry  = cppassist::make_unique<gtx::EdgeGeometry>();
    m_lineGeometry  = cppassist::make_unique<gtx::LineGeometry>();

    // Create attribute storage on GPU
    m_attributeStorage = cppassist::make_unique<gtx::AttributeStorage>();

    // Create GPU data
    if (dataset.value())
    {
        m_nodeGeometry->setData(*dataset.value(), *maxNodes);
        m_patchGeometry->setNodeGeometry(m_nodeGeometry.get());
        m_edgeGeometry->setData(*dataset.value());
        m_lineGeometry->setData(*dataset.value());
        m_attributeStorage->setData(*dataset.value());
        createMapLayersTexture();
    }
    else
    {
        m_attributeStorage->clearData();
    }

    // Update outputs
    nodeGeometry.setValue(m_nodeGeometry.get());
    patchGeometry.setValue(m_patchGeometry.get());
    edgeGeometry.setValue(m_edgeGeometry.get());
    lineGeometry.setValue(m_lineGeometry.get());
    nodeGeometrySize.setValue(m_nodeGeometry->size());
    mapLayersTexture.setValue(m_mapLayersTexture.get());
    staticAttributesTexture.setValue(m_attributeStorage->staticAttributesTexture());
    nodeAttributesTexture.setValue(m_attributeStorage->nodeAttributesTexture());
    edgeAttributesTexture.setValue(m_attributeStorage->edgeAttributesTexture());
    lineAttributesTexture.setValue(m_attributeStorage->lineAttributesTexture());
    attributeNames.setValue(m_attributeStorage->attributes());
    numStaticAttributes.setValue(m_attributeStorage->numStaticAttributes());
    numNodeAttributes.setValue(m_attributeStorage->numNodeAttributes());
    numEdgeAttributes.setValue(m_attributeStorage->numEdgeAttributes());
    numLineAttributes.setValue(m_attributeStorage->numLineAttributes());
    if (*dataset) {
        area.setValue(glm::vec4(
            glm::radians(dataset->area().minPos().getLat()),
            glm::radians(dataset->area().minPos().getLong()),
            glm::radians(dataset->area().maxPos().getLat()),
            glm::radians(dataset->area().maxPos().getLong())
        ));
    }
}

void GpuDataStage::createMapLayersTexture()
{
    // Get directory that contains the dataset
    std::string dir = FilePath(FilePath(dataset->filename()).directoryPath()).path();

    // Get texture filenames
    std::vector<std::string> textures;
    for (auto layer : dataset->mapLayers()) {
        std::string filename = dir + "/" + layer;
        textures.push_back(filename);
    }

    // Add default map layer (transparent)
    if (textures.size() < 1) {
        textures.push_back(gtx::dataPath() + "/textures/transparent.png");
    }

    // Create 2D texture array
    m_mapLayersTexture = Tools::createTextureArray(textures);
}


} // namespace gtx
