
#include <gtxlib-stages/PickingStage.h>

#include <vector>
#include <set>

#include <glbinding/gl/enum.h>

#include <globjects/Texture.h>
#include <globjects/Buffer.h>

#include <gloperate/rendering/ColorRenderTarget.h>


using namespace cppexpose;
using namespace globjects;
using namespace gloperate;


namespace gtx
{


CPPEXPOSE_COMPONENT(PickingStage, gloperate::Stage)


PickingStage::PickingStage(Environment * environment, const std::string & name)
: Stage(environment, "PickingStage", name)
, idTarget("idTarget", this)
, mousePos("mousePos", this)
, fixId("fixId", this, false)
, radius("radius", this, 1)
, hoveredID("hoveredID", this, -1)
, selectedIDs("selectedIDs", this)
, numIDs("numIDs", this, 0)
, m_width(0)
, m_height(0)
{
}

PickingStage::~PickingStage()
{
}

int PickingStage::pickAt(int x, int y)
{
    int id = -1;

    if (x >= 0 && x < m_width && y >= 0 && y < m_height) {
        id = (int)m_idBuffer[y * m_width + x];
    }

    // The ID buffer is cleared with -1, and through blending, IDs are added to that.
    // So, if id == -1, nothing is selected, otherwise, the value is ID-1
    // [TODO] Disable blending for ID layer
    if (fixId.value() && id != -1) id = id + 1;

    return id;
}

void PickingStage::onContextInit(AbstractGLContext *)
{
    // Create selection texture and buffer
    m_selectionTexture = cppassist::make_unique<globjects::Texture>(gl::GL_TEXTURE_BUFFER);
    m_selectionBuffer  = cppassist::make_unique<globjects::Buffer>();
}

void PickingStage::onContextDeinit(AbstractGLContext *)
{
    // Release OpenGL objects
    m_selectionTexture = nullptr;
    m_selectionBuffer  = nullptr;
}

void PickingStage::onProcess()
{
    // Texture if ID buffer is valid
    if (!idTarget.value()) return;

    // Get ID buffer
    auto * texture = idTarget->textureAttachment();
    if (!texture) return;

    // Get viewport size
    m_width  = texture->getLevelParameter(0, gl::GL_TEXTURE_WIDTH);
    m_height = texture->getLevelParameter(0, gl::GL_TEXTURE_HEIGHT);

    // Read ID buffer from GPU
    m_idBuffer.resize(m_width * m_height);
    texture->getImage(0, gl::GL_RED, gl::GL_FLOAT, m_idBuffer.data());

    // Pick ID at current mouse position
    int id = pickAt(mousePos->x, mousePos->y);

    // Update hovered ID
    if (id != *hoveredID) {
        hoveredID.setValue(id);
    }

    // Pick IDs in region
    if (*radius > 1) {
        // Set of selected IDs
        std::set<int> ids;

        // Determine all selected IDs in a given area
        int r = *radius;
        for (int y = mousePos->y - 20; y <= mousePos->y + r; y++) {
            for (int x = mousePos->x - 20; x <= mousePos->x + r; x++) {
                // Pick ID
                int id = pickAt(x, y);

                // Add selected ID to set
                if (id > -1) {
                    ids.insert(id);
                }
            }
        }

        // Convert set into array
        std::vector<float> data;
        for (int id : ids) {
            data.push_back((float)id);
        }

        // Update selection texture
        m_selectionBuffer->setData(data, gl::GL_STATIC_DRAW);
        m_selectionTexture->texBuffer(gl::GL_R32F, m_selectionBuffer.get());

        // Update output
        selectedIDs.setValue(m_selectionTexture.get());
        numIDs.setValue(data.size());
    }
}


} // namespace gtx
