
#include <gtxlib-stages/SelectionStage.h>

#include <iostream>

#include <glbinding/gl/gl.h>

#include <globjects/Framebuffer.h>
#include <globjects/FramebufferAttachment.h>
#include <globjects/Texture.h>
#include <globjects/Program.h>
#include <globjects/Shader.h>

#include <gloperate/base/Environment.h>
#include <gloperate/rendering/AbstractDrawable.h>

#include <gtxlib-core/data/DataSet.h>


using namespace globjects;
using namespace gloperate;


namespace gtx
{


CPPEXPOSE_COMPONENT(SelectionStage, gloperate::Stage)


SelectionStage::SelectionStage(Environment * environment, const std::string & name)
: Stage(environment, "SelectionStage", name)
, dataset("dataset", this)
, attributes("attributes", this)
, edgeGeometry("edgeGeometry", this)
, selectedIds("selectedIds", this)
, staticAttributesTexture("staticAttributesTexture", this)
, nodeAttributesTexture("nodeAttributesTexture", this)
, numStaticAttributes("numStaticAttributes", this)
, numNodeAttributes("numNodeAttributes", this)
, selectionActive("selectionActive", this)
, selectionTexture("selectionTexture", this)
{
}

SelectionStage::~SelectionStage()
{
}

void SelectionStage::onContextInit(AbstractGLContext *)
{
    // Get data path
    std::string dataPath = gtx::dataPath();

    // Create textures
    m_texture1 = Texture::createDefault(gl::GL_TEXTURE_2D);
    m_texture2 = Texture::createDefault(gl::GL_TEXTURE_2D);

    // Create program
    m_program = cppassist::make_unique<globjects::Program>();

    m_vertexShader.reset(environment()->resourceManager()->load<globjects::Shader>(dataPath + "/shaders/selection/selection.vert"));
    m_program->attach(m_vertexShader.get());

    m_geometryShader.reset(environment()->resourceManager()->load<globjects::Shader>(dataPath + "/shaders/selection/selection.geom"));
    m_program->attach(m_geometryShader.get());

    m_fragmentShader.reset(environment()->resourceManager()->load<globjects::Shader>(dataPath + "/shaders/selection/selection.frag"));
    m_program->attach(m_fragmentShader.get());

    m_attributesShader.reset(environment()->resourceManager()->load<globjects::Shader>(dataPath + "/shaders/base/attributes.geom"));
    m_program->attach(m_attributesShader.get());
}

void SelectionStage::onContextDeinit(AbstractGLContext *)
{
    // Release OpenGL objects
    m_texture1       = nullptr;
    m_texture2       = nullptr;
    m_fbo1           = nullptr;
    m_fbo2           = nullptr;
    m_program        = nullptr;
    m_vertexShader   = nullptr;
    m_geometryShader = nullptr;
    m_fragmentShader = nullptr;
}

void SelectionStage::onProcess()
{
    int idAttributeIndex = -1;
    bool nodesSelected = false;

    // Determine if the dataset has a node-attribute 'id'
    const std::vector<std::string> & values = *attributes;
    auto it = std::find(values.begin(), values.end(), "Node - id");
    if (it != values.end()) {
        idAttributeIndex = std::distance(values.begin(), it);
    }

    // Initialize list of selected node IDs
    std::array<int, 50> ids;
    for (size_t i=0; i<50; i++) {
        ids[i] = -1;
    }

    // Get list of selected node IDs
    size_t i = 0;
    for (int id : *selectedIds) {
        ids[i] = id;

        nodesSelected = true;

        i++;
        if (i >= 50) break;
    }

    // Determine texture size
    // unsigned int numNodes = dataset->nodes().size();
    unsigned int width  = 1024;
    unsigned int height = 1024; //(numNodes / 1024) + 1;

    // Get viewport size
    if (!m_texture1.get() || !m_texture2.get() || dataset.hasChanged())
    {
        // Create textures
        m_texture1->image2D(0, gl::GL_R32F, width, height, 0, gl::GL_RED, gl::GL_FLOAT, nullptr);
        m_texture2->image2D(0, gl::GL_R32F, width, height, 0, gl::GL_RED, gl::GL_FLOAT, nullptr);

        // Setup FBOs
        m_fbo1 = cppassist::make_unique<globjects::Framebuffer>();
        m_fbo1->attachTexture(gl::GL_COLOR_ATTACHMENT0, m_texture1.get());
        m_fbo2 = cppassist::make_unique<globjects::Framebuffer>();
        m_fbo2->attachTexture(gl::GL_COLOR_ATTACHMENT0, m_texture2.get());
    }

    // Clear input texture for first rendering step
    m_fbo1->bind(gl::GL_FRAMEBUFFER);
    gl::glViewport(0, 0, width, height);
    m_fbo1->clearBuffer(gl::GL_COLOR, 0, 100.0f);

    // Calculate k-neighborhood step by step (alternating textures)
    for (int i=0; i<10; i++)
    {
        // Alternate source and destination texture
        Texture     * texture = (i%2 == 0) ? m_texture1.get() : m_texture2.get();
        Framebuffer * fbo     = (i%2 == 0) ? m_fbo2.get()     : m_fbo1.get();

        // Bind texture
        texture->bindActive(0);
        staticAttributesTexture->bindActive(1);
        nodeAttributesTexture->bindActive(2);

        // Update program
        m_program->setUniform<int>("nodeSelection",    0);
        m_program->setUniform<int>("staticAttributes", 1);
        m_program->setUniform<int>("nodeAttributes",   2);
        m_program->setUniform<int>("numStaticAttributes", numStaticAttributes.value());
        m_program->setUniform<int>("numNodeAttributes", numNodeAttributes.value());
        m_program->setUniform<int>("numEdgeAttributes", 0);
        m_program->setUniform<int>("numLineAttributes", 0);
        m_program->setUniform<int>("idAttributeIndex", idAttributeIndex);
        m_program->setUniform< std::array<int, 50> >("selectedIds", ids);

        // Bind FBO
        fbo->bind(gl::GL_FRAMEBUFFER);

        // Set viewport
        gl::glViewport(0, 0, width, height);

        // Clear texture
        fbo->clearBuffer(gl::GL_COLOR, 0, 100.0f);

        // Setup rendering state
        gl::glEnable(gl::GL_BLEND);
        gl::glBlendEquation(gl::GL_MIN);

        // Process edges
        m_program->use();
        edgeGeometry->draw();

        // Unbind FBO
        m_fbo1->unbind();

        // Reset rendering state
        gl::glDisable(gl::GL_BLEND);
        gl::glBlendEquation(gl::GL_FUNC_ADD);
    }

    // Update outputs
    selectionActive.setValue(nodesSelected);
    selectionTexture.setValue(m_texture1.get());
}


} // namespace gtx
