
#include <gtxlib-stages//TrackballStage.h>

#include <iostream>

#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/euler_angles.hpp>

#include <cppassist/memory/make_unique.h>

#include <gloperate/base/Environment.h>
#include <gloperate/input/MouseEvent.h>


using namespace gloperate;


namespace gtx
{


CPPEXPOSE_COMPONENT(TrackballStage, gloperate::Stage)


TrackballStage::TrackballStage(Environment * environment, const std::string & name)
: Stage(environment, name), AbstractEventConsumer(environment->inputManager())
, viewport("viewport", this)
, pitch("pitch", this, 0.0f)
, yaw("yaw", this, 0.0f)
, zoom("zoom", this, 2.5f)
, rotateX("rotateX", this, 0.0f)
, rotateY("rotateY", this, 0.0f)
, camera("camera", this)
, m_camera(cppassist::make_unique<Camera>())
, m_lastMousePosition(0, 0)
, m_dragging(false)
, m_rotating(false)
{
}

TrackballStage::~TrackballStage()
{
}

void TrackballStage::onEvent(InputEvent * event)
{
    float pitch   = *this->pitch;
    float yaw     = *this->yaw;
    float zoom    = *this->zoom;
    float rotateX = *this->rotateX;
    float rotateY = *this->rotateY;

    switch(event->type())
    {
        case InputEvent::Type::MouseButtonPress:
            if (static_cast<MouseEvent*>(event)->button() == MouseButtonLeft)
            {
                // Start dragging
                assert(!m_dragging);
                m_lastMousePosition = static_cast<MouseEvent*>(event)->pos();
                m_dragging = true;
            }

            if (static_cast<MouseEvent*>(event)->button() == MouseButtonRight)
            {
                // Start rotating
                assert(!m_rotating);
                m_lastMousePosition = static_cast<MouseEvent*>(event)->pos();
                m_rotating = true;
            }

            break;

        case InputEvent::Type::MouseButtonRelease:
            if (static_cast<MouseEvent*>(event)->button() == MouseButtonLeft)
            {
                // Stop dragging
                assert(m_dragging);
                m_dragging = false;
            }

            if (static_cast<MouseEvent*>(event)->button() == MouseButtonRight)
            {
                // Stop m_rotating
                assert(m_rotating);
                m_rotating = false;
            }
            break;

        case InputEvent::Type::MouseMove:
            // Dragging
            if (m_dragging)
            {
                auto currentMousePosition = static_cast<MouseEvent*>(event)->pos();
                float scaling = 0.2f * zoom;

                pitch += (m_lastMousePosition.y - currentMousePosition.y) * scaling;
                pitch = glm::clamp(pitch, -89.0f, 89.0f);

                yaw += (m_lastMousePosition.x - currentMousePosition.x) *scaling;
                yaw = yaw -180 % 360 + 180;

                m_lastMousePosition = currentMousePosition;
                camera.invalidate();
            }

            // Rotation
            if (m_rotating)
            {
                auto currentMousePosition = static_cast<MouseEvent*>(event)->pos();
                float scaling = 0.1f;

                rotateX += (m_lastMousePosition.x - currentMousePosition.x) * scaling * 2.0f;
                while (rotateX < -180.0f) rotateX += 360.0f;
                while (rotateX >  180.0f) rotateX -= 360.0f;

                rotateY += (currentMousePosition.y - m_lastMousePosition.y) * scaling;
                rotateY = glm::clamp(rotateY, -90.0f, 0.0f);

                m_lastMousePosition = currentMousePosition;

                camera.invalidate();
            }

            break;

        case InputEvent::Type::MouseWheelScroll:
            // Zooming
            if (!m_dragging)
            {
                float factor = (zoom < 1.0f) ? glm::pow(zoom, 1.2f) : zoom;
                zoom -= static_cast<MouseEvent*>(event)->wheelDelta().y * factor / 3000;
                zoom = glm::clamp(zoom, 0.005f, 10.0f);

                camera.invalidate();
            }
            break;

        default:
            break;
    }

    this->pitch.setValue(pitch);
    this->yaw.setValue(yaw);
    this->zoom.setValue(zoom);
    this->rotateX.setValue(rotateX);
    this->rotateY.setValue(rotateY);
}

void TrackballStage::onProcess()
{
    // [DEBUG]
    /*
    std::cout << "========================================" << std::endl;
    std::cout << "pitch:   " << *pitch << std::endl;
    std::cout << "yaw:     " << *yaw << std::endl;
    std::cout << "zoom:    " << *zoom << std::endl;
    std::cout << "rotateX: " << *rotateX << std::endl;
    std::cout << "rotateY: " << *rotateY << std::endl;
    std::cout << "========================================" << std::endl;
    */

    // Calculate position on the surface of the earth
    auto mat = glm::yawPitchRoll(glm::radians(*yaw), glm::radians(*pitch), 0.0f);
    auto center = glm::vec3(mat * glm::vec4(0.0f, 0.0f, 1.0f, 1.0f));

    // Calculate local coordinate system
    auto d = center;
    auto r = glm::cross(d, glm::vec3(0.0f, 1.0f, 0.0f));
    auto u = -glm::cross(d, r);

    // Calculate rotation matrix
    glm::mat4 rot = glm::rotate(glm::mat4(), glm::radians(*rotateX), d) * glm::rotate(glm::mat4(), glm::radians(*rotateY), r);

    // Rotate coordinate system
    u = glm::vec3(rot * glm::vec4(u, 1.0f));
    r = glm::vec3(rot * glm::vec4(d, 1.0f));

    // Calculate new camera position
    auto eye = center + r * (*zoom);

    // Update camera
    m_camera->lookAt(eye / 1.0f, center, u);
    m_camera->perspective(glm::radians(40.0f), viewport->z / viewport->w, 0.0001f, 64.0f);
    camera.setValue(m_camera.get());
}


} // namespace gtx
