
#include <gtxlib-stages/WorldGeometryStage.h>

#include <glm/trigonometric.hpp>
#include <glm/gtc/constants.hpp>

#include <cppassist/memory/make_unique.h>

#include <cppexpose/variant/Variant.h>

#include <glbinding/gl/enum.h>

#include <globjects/base/File.h>
#include <globjects/base/AbstractStringSource.h>
#include <globjects/Shader.h>

#include <gloperate/rendering/Quad.h>
#include <gloperate/rendering/Sphere.h>

#include <gtxlib-core/gtxlib-core.h>
#include <gtxlib-core/geometry/Icosahedron.h>


using namespace cppexpose;
using namespace gloperate;


namespace gtx
{


CPPEXPOSE_COMPONENT(WorldGeometryStage, gloperate::Stage)


WorldGeometryStage::WorldGeometryStage(Environment * environment, const std::string & name)
: Stage(environment, name)
, projection("projection", this, "Geocentric")
, drawable("drawable", this)
, vertexShader("vertexShader", this)
, fragmentShader("fragmentShader", this)
, projectionVertexShader("projectionVertexShader", this)
, projectionGeometryShader("projectionGeometryShader", this)
{
    // Set options
    Variant projections = Variant::array();
    projections.asArray()->push_back("Geocentric");
    projections.asArray()->push_back("Mercator");
    projections.asArray()->push_back("Transverse Mercator");
    projections.asArray()->push_back("360 Degrees");
    projection.setOption("choices", projections);
}

WorldGeometryStage::~WorldGeometryStage()
{
}

void WorldGeometryStage::onContextInit(AbstractGLContext *)
{
}

void WorldGeometryStage::onContextDeinit(AbstractGLContext *)
{
    // Release OpenGL objects
    m_geometry                 = nullptr;
    m_vertexShader             = nullptr;
    m_fragmentShader           = nullptr;
    m_vertexShaderSource       = nullptr;
    m_fragmentShaderSource     = nullptr;
    m_projectionShaderSource   = nullptr;
    m_projectionVertexShader   = nullptr;
    m_projectionGeometryShader = nullptr;
}

void WorldGeometryStage::onProcess()
{
    std::string projection = *this->projection;

    // Get data path
    std::string dataPath = gtx::dataPath();

    // Geocentric
    if (projection == "Geocentric")
    {
        // Create geometry
        m_geometry = cppassist::make_unique<Sphere>(1.0f);

        // Create shaders for rendering the geometry
        m_vertexShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/world.vert");
        m_fragmentShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/world.frag");

        // Create projection shader
        m_projectionShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/proj_geocentric.glsl");
    }

    // Mercator projection
    else if (projection == "Mercator")
    {
        // Create geometry
        m_geometry = cppassist::make_unique<Quad>(2.0f, 1.0f);

        // Create shaders for rendering the geometry
        m_vertexShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/map.vert");
        m_fragmentShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/map_mercator.frag");

        // Create projection shader
        m_projectionShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/proj_mercator.glsl");
    }

    // Transverse mercator projection
    else if (projection == "Transverse Mercator")
    {
        // Create geometry
        m_geometry = cppassist::make_unique<Quad>(2.0f, 1.0f);

        // Create shaders for rendering the geometry
        m_vertexShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/map.vert");
        m_fragmentShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/map_transverse_mercator.frag");

        // Create projection shader
        m_projectionShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/proj_transverse_mercator.glsl");
    }

    // 360 degrees projection
    else if (projection == "360 Degrees")
    {
        // Create geometry
        m_geometry = cppassist::make_unique<Quad>(1.0f, 1.0f);

        // Create shaders for rendering the geometry
        m_vertexShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/map.vert");
        m_fragmentShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/map_360degrees.frag");

        // Create projection shader
        m_projectionShaderSource = globjects::Shader::sourceFromFile(dataPath + "/shaders/world/proj_360degrees.glsl");
    }

    // Create shader objects
    m_vertexShader             = cppassist::make_unique<globjects::Shader>(gl::GL_VERTEX_SHADER,   m_vertexShaderSource.get());
    m_fragmentShader           = cppassist::make_unique<globjects::Shader>(gl::GL_FRAGMENT_SHADER, m_fragmentShaderSource.get());
    m_projectionVertexShader   = cppassist::make_unique<globjects::Shader>(gl::GL_VERTEX_SHADER,   m_projectionShaderSource.get());
    m_projectionGeometryShader = cppassist::make_unique<globjects::Shader>(gl::GL_GEOMETRY_SHADER, m_projectionShaderSource.get());

    // Set output values
    drawable.setValue(m_geometry.get());
    vertexShader.setValue(m_vertexShader.get());
    fragmentShader.setValue(m_fragmentShader.get());
    projectionVertexShader.setValue(m_projectionVertexShader.get());
    projectionGeometryShader.setValue(m_projectionGeometryShader.get());
}


} // namespace gtx
